/*  The Ultimate POCSAG decoder v1.03d
 *
 *  Author: BHES <upocsag@s-lab.se>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#define _GNU_SOURCE

#ifndef SOUNDCARD
#ifdef HAVE_LINUX_SOUNDCARD_H
#include <linux/soundcard.h>
#define SOUNDCARD
#endif
#endif

#ifndef SOUNDCARD
#ifdef HAVE_SYS_SOUNDCARD_H
#include <sys/soundcard.h>
#define SOUNDCARD
#endif
#endif

#ifdef USE_MYSQL
  #include <mysql/mysql.h>
  #include "mysql_config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>
#include <pthread.h>
#include <glob.h>
#include <sys/time.h>
#include "pocsag.h"

#ifndef SHA1_H
#define SHA1_H
#include "sha1.h"
#endif

int main(int argc, char **argv)
{
  int32_t fd,ch,fmt,sr,i,j,k,l,m,dsp_init_ok,pcm_byte_count,bitrate,bitrate_index,found,channel;
  int32_t raw_batches = 0;
  int32_t raw_parity_errors = 0;
  int32_t raw_files_read = 0;
  int32_t raw_ecc_errors_correctable = 0;
  int32_t raw_ecc_errors_uncorrectable = 0;
  int32_t raw_timestamp = 0;
  int32_t raw_timestamp_found = 0;
  uint32_t raw_cw_temp, raw_cw_ecc_status,*ts;
  time_t tss;

  list.item = NULL;
  bitrate_index = 0;
  uint32_t checksum;
  struct sigaction sv;
  struct recipients_w rece;
  struct tcp_struct pcm,raw_s;
  struct string_list *file_list;
  file_list = NULL;
  char *z,*zz;
  int32_t queue_full;
  statsfile = NULL;
  queue_full = 0;
  cw_temp = nmalloc(sizeof(int32_t)*16);
  memset(&pcm,0,sizeof(struct tcp_struct));
  memset(&raw_s,0,sizeof(struct tcp_struct));
  memset(&rece,0,sizeof(struct recipients_w));
  sv.sa_handler = (void*)pipe_handler;
  sv.sa_flags   = SA_RESTART;
  sigaction(SIGPIPE,&sv,NULL);
  dsp_init_ok = 0;
  fd = 0;
  pcm_byte_count = 0;
  main_config_index = 0;
  struct pocraw *raw;
  char *tmp;
  FILE *file;
  struct tm *now;
  char *current_ts,*old_ts;
  setbuf (stdout , 0);
  pcm_server_new_connections = NULL;
  raw_server_new_connections = NULL;
  memset(&pcm_server_parameters,0,sizeof(struct server_parameters));
  memset(&raw_server_parameters,0,sizeof(struct server_parameters));
  pcm_server_header_string = NULL;
  raw_server_header_string = NULL;
  raw_chunk = NULL;
  config = nmalloc(sizeof(struct configuration_new));
  memset(config,0,sizeof(struct configuration_new));
  #ifdef USE_MYSQL
  register_log_function(config, &log_mysql,"mysql");
  #endif
  register_log_function(config, &log_console,"console");
  register_log_function(config, &log_winmsg,"winmsg");
  register_log_function(config, &log_irc,"irc");
  register_log_function(config, &log_file,"file");
  register_log_function(config, &log_mail,"mail");
  register_log_function(config, &log_telnet,"telnet");
  register_log_function(config, &log_external, "external");
  register_log_function(config, NULL,"main");

  config->log_config = nmalloc(config->count*sizeof(struct log_functions_config *));
  for (i=0;i<config->count;i++)
  {
    config->log_config[i] = nmalloc(sizeof(struct log_functions_config));
    memset(config->log_config[i],0,sizeof(struct log_functions_config));
  }

  for (i=0;i<config->count;i++)
  {
    config->log_config[i]->str_num[discard_errored_messages] = no;
    config->log_config[i]->str_num_is_set[discard_errored_messages] = 1;
  }

  main_config_index = get_log_function_index(config,"main");
  if (main_config_index >= 0)
  {
    /* Initialize default values */
    config->log_config[main_config_index]->num[dsp_sample_rate]               = CONFIG_DSP_SAMPLE_RATE;
    config->log_config[main_config_index]->num[dsp_sample_bits]               = CONFIG_DSP_SAMPLE_BITS;
    config->log_config[main_config_index]->str_num[dsp_input_mode]            = CONFIG_DSP_INPUT_MODE;
    config->log_config[main_config_index]->str_num_is_set[dsp_input_mode]     = 1;
    config->log_config[main_config_index]->num[dsp_input_channels]            = CONFIG_DSP_INPUT_CHANNELS;;
    config->log_config[main_config_index]->num[dsp_decode_channels]           = CONFIG_DSP_DECODE_CHANNELS;
    config->log_config[main_config_index]->num[dsp_left_weight]               = CONFIG_DSP_LEFT_WEIGHT;
    config->log_config[main_config_index]->num[dsp_right_weight]              = CONFIG_DSP_RIGHT_WEIGHT;
    config->log_config[main_config_index]->num[noise_smoothing_threshold]     = CONFIG_THRESHOLD;
    config->log_config[main_config_index]->num[raw_logging_error_threshold]   = CONFIG_ERROR_THRESHOLD;

    config->log_config[main_config_index]->str_num[noise_smoothing]           = yes;

    config->log_config[main_config_index]->str[input_source]                  = "/dev/dsp";
    config->log_config[main_config_index]->str_num[input_mode]                = dsp_pcm;
    config->log_config[main_config_index]->str_num_is_set[input_mode]         = 1;

    config->log_config[main_config_index]->str[stats_file]                    = "stats.dat";
    config->log_config[main_config_index]->str_num[stats]                     = yes;
    config->log_config[main_config_index]->str_num_is_set[stats]              = 1;

    config->log_config[main_config_index]->num[pcm_server_port]               = 5000;
    config->log_config[main_config_index]->str_num[pcm_server]                = no;
    config->log_config[main_config_index]->str_num_is_set[pcm_server]         = 1;

    config->log_config[main_config_index]->num[raw_server_port]               = 5001;
    config->log_config[main_config_index]->str_num[raw_server]                = no;
    config->log_config[main_config_index]->str_num_is_set[raw_server]         = 1;

    config->log_config[main_config_index]->str[program_logfile]               = "pocsag.log";
    config->log_config[main_config_index]->str_num[program_log]               = yes;
    config->log_config[main_config_index]->str_num_is_set[program_log]        = 1;

    config->log_config[main_config_index]->str_num[print_codewords]           = no;
    config->log_config[main_config_index]->str_num_is_set[print_codewords]    = 1;

    config->log_config[main_config_index]->str_num[print_codewords_eh]        = yes;
    config->log_config[main_config_index]->str_num_is_set[print_codewords_eh] = 1;

    config->log_config[main_config_index]->str_num[raw_logging]               = no;
    config->log_config[main_config_index]->str_num_is_set[raw_logging]        = 1;

    config->log_config[main_config_index]->str[charset]                       = charset_file;

    config->log_config[get_log_function_index(config,"console")]->rec         = parse_rcpt_list_new(&config->log_config[get_log_function_index(config,"console")]->rec, "a", -1,to ); /* dummy "recipient" for the console output function */
    config->log_config[get_log_function_index(config,"file")]->rec            = parse_rcpt_list_new(&config->log_config[get_log_function_index(config,"file")]->rec, logfile_date_template, -1,logfile); /* dummy "recipient" for the file output function */

    for (i=0;i<config_int_array_enum_count;i++)
    {
      config->log_config[main_config_index]->num_list[i] = nmalloc(sizeof(struct integer_list));
      memset(config->log_config[main_config_index]->num_list[i],0,sizeof(struct integer_list));
      config->log_config[main_config_index]->num_list[i]->count = config_bitrates_num_alpha_size[i];
      config->log_config[main_config_index]->num_list[i]->list  = nmalloc(sizeof(int32_t)*config_bitrates_num_alpha_size[i]);
      memcpy(config->log_config[main_config_index]->num_list[i]->list,config_bitrates_num_alpha[i],sizeof(int32_t)*config_bitrates_num_alpha_size[i]);
    }

    i = get_log_function_index(config,"mail");
    if (i >= 0)
    {
      config->log_config[i]->str[mta] = mta_cmd;
    }

    i = get_log_function_index(config,"winmsg");
    if (i >= 0)
    {
      config->log_config[i]->str[smbclient] = smbclient_cmd;
      config->log_config[i]->str[nmblookup] = nmblookup_cmd;
    }

    i = get_log_function_index(config,"console");
    if (i >= 0)
    {

      config->log_config[i]->str_num[enabled] = yes;
      config->log_config[i]->str_num_is_set[enabled] = 1;
      config->log_config[i]->str_num[eh_enable] = yes;
      config->log_config[i]->str_num_is_set[eh_enable] = 1;
    }

    i = get_log_function_index(config,"file");
    if (i >= 0)
    {

      config->log_config[i]->str_num[enabled] = yes;
      config->log_config[i]->str_num_is_set[enabled] = 1;
    }

    i = get_log_function_index(config,"irc");
    if (i >= 0)
    {
      config->log_config[i]->str_num[eh_enable] = yes;
      config->log_config[i]->str_num_is_set[eh_enable] = 1;
    }

    i = get_log_function_index(config,"external");
    if (i >= 0)
    {
      config->log_config[i]->str_num[enabled] = no;
      config->log_config[i]->str_num_is_set[enabled] = 1;
    }

    for (i=0;i<sizeof(eh_ptr)/sizeof(char*);i++)
    {
      j = get_log_function_index(config,eh_lookup[i]);

      if (j >= 0)
      {
        for (k=0;k<config_error_highlight_enum_count;k++)
        {
          config->log_config[j]->error_highlight[k] = nmalloc(strlen(eh_ptr[i][k]) + 1);
          memcpy(config->log_config[j]->error_highlight[k],eh_ptr[i][k],strlen(eh_ptr[i][k]) + 1);
          config->log_config[j]->error_highlight[k][strlen(eh_ptr[i][k])] = 0;
        }
      }
    }

    config->log_config[main_config_index]->num[dsp_buffer_length]    = CONFIG_PCM_RAW_CHUNK_LENGTH_SECONDS;
    config->log_config[main_config_index]->num[dsp_max_queue_length] = CONFIG_AUDIO_MAX_QUEUE_SECONDS;

    config_parse(config,argc,argv);
    if (conf_ok == 0)
    {
      printf("%s\n\n",program_string);
      printf("Use \"%s -h\" to see the help.\n",argv[0]);
      exit(0);
    }

    build_character_tables(config->log_config[main_config_index]->str[charset]);

    pthread_mutex_init(&time_mutex, NULL);

    i = get_log_function_index(config,"telnet");
    if (i >= 0)
    {
      if ((config->log_config[i]->str_num[enabled] == yes) && (config->log_config[i]->str_num_is_set[enabled]))
      {
        pthread_mutex_init(&telnet_server_mutex, NULL);
        pthread_cond_init (&telnet_server_cond, NULL);
        pthread_attr_init(&telnet_server_attr);
        pthread_attr_setdetachstate(&telnet_server_attr, PTHREAD_CREATE_DETACHED);
        pthread_create(&telnet_server_thread, &telnet_server_attr, (void *)telnet_listen, (void *)config->log_config[i]);
      }
    }

    #ifdef DEBUG
    dump_config(config);
    pthread_mutex_init(&memory_mutex, NULL);
    pthread_cond_init (&memory_cond, NULL);
    #endif

    pthread_mutex_init(&pcm_server_mutex, NULL);
    pthread_mutex_init(&pcm_server_mutex2, NULL);
    pthread_mutex_init(&raw_server_mutex, NULL);
    pthread_mutex_init(&raw_server_mutex2, NULL);

    pthread_mutex_init(&pcm_server_auth, NULL);
    pthread_mutex_init(&raw_server_mutex, NULL);
    pthread_mutex_init(&stats_mutex, NULL);
    pthread_mutex_init(&memory_mutex2, NULL);
    pthread_mutex_init(&print_log_mutex,NULL);
    pthread_mutex_init(&int_add_remove, NULL);
    pthread_mutex_init(&exit_mutex, NULL);
    pthread_cond_init(&exit_cond, NULL);

    statsfile = config->log_config[main_config_index]->str[stats_file];

    switch(config->log_config[main_config_index]->str_num[dsp_input_mode])
    {
      case stereo:config->log_config[main_config_index]->num[dsp_input_channels] = 2;
                  config->log_config[main_config_index]->num[dsp_decode_channels] = 2;

                  break;

      case left:config->log_config[main_config_index]->num[dsp_input_channels] = 1;
                config->log_config[main_config_index]->num[dsp_decode_channels] = 1;
                config->log_config[main_config_index]->num[dsp_left_weight] = 1;
                config->log_config[main_config_index]->num[dsp_right_weight] = 0;
                break;

      case right:config->log_config[main_config_index]->num[dsp_input_channels] = 2;
                 config->log_config[main_config_index]->num[dsp_decode_channels] = 1;
                 config->log_config[main_config_index]->num[dsp_left_weight] = 0;
                 config->log_config[main_config_index]->num[dsp_right_weight] = 1;
                 break;
    }

    config->log_config[main_config_index]->num[dsp_buffer_length_bytes] =
    round_up_nearest_multiple(
    (config->log_config[main_config_index]->num[dsp_buffer_length]*
    config->log_config[main_config_index]->num[dsp_sample_rate]*
    config->log_config[main_config_index]->num[dsp_input_channels]*
    (config->log_config[main_config_index]->num[dsp_sample_bits]/8)/1000),
    config->log_config[main_config_index]->num[dsp_input_channels]*
    (config->log_config[main_config_index]->num[dsp_sample_bits]/8));

    config->log_config[main_config_index]->num[dsp_max_queue_length_bytes] =
    round_up_nearest_multiple(
    (config->log_config[main_config_index]->num[dsp_max_queue_length]*
    config->log_config[main_config_index]->num[dsp_sample_rate]*
    config->log_config[main_config_index]->num[dsp_input_channels]*
    (config->log_config[main_config_index]->num[dsp_sample_bits]/8)/1000),
    config->log_config[main_config_index]->num[dsp_input_channels]*
    (config->log_config[main_config_index]->num[dsp_sample_bits]/8));

    bitcount      = nmalloc(config->log_config[main_config_index]->num[dsp_decode_channels]*sizeof(uint32_t *));
    buffer        = nmalloc(config->log_config[main_config_index]->num[dsp_decode_channels]*sizeof(uint32_t *));
    decode_buffer = nmalloc(config->log_config[main_config_index]->num[dsp_decode_channels]*sizeof(uint32_t *));

    for (i=0;i<config->log_config[main_config_index]->num[dsp_decode_channels];i++)
    {
      bitcount[i]      = nmalloc(config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(uint32_t  ));
      memset(bitcount[i],0,config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(uint32_t  ));

      buffer[i]        = nmalloc(config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(uint32_t *));
      memset(buffer[i],0,config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(uint32_t *));

      decode_buffer[i] = nmalloc(config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(struct message_block));
      memset(decode_buffer[i],0,config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(struct message_block));

      for (j=0;j<config->log_config[main_config_index]->num_list[bitrates]->count;j++)
      {
        buffer[i][j] = nmalloc(CONFIG_LISTSIZE*sizeof(uint32_t));
        memset(buffer[i][j],0,CONFIG_LISTSIZE*sizeof(uint32_t));
      }
    }

    audio_queue = nmalloc(sizeof(struct pointer_queue));
    memset(audio_queue,0,sizeof(struct pointer_queue));

    log_queue = nmalloc(sizeof(struct pointer_queue));
    memset(log_queue,0,sizeof(struct pointer_queue));

    pcm_server_audio_queue = nmalloc(sizeof(struct pointer_queue));
    memset(pcm_server_audio_queue,0,sizeof(struct pointer_queue));

    raw_server_queue = nmalloc(sizeof(struct pointer_queue));
    memset(raw_server_queue,0,sizeof(struct pointer_queue));

    if ((config->log_config[main_config_index]->str_num[input_mode] == dsp_pcm) &&  (config->log_config[main_config_index]->str_num_is_set[input_mode]))
    {

      fd = open(config->log_config[main_config_index]->str[input_source],O_RDONLY);
      if (fd == -1)
      {
        asprintf(&z,"open: %s (%s)\n",strerror(errno),config->log_config[main_config_index]->str[input_source]);
        print_log(MSG_ERROR,z);
        nfree(z);
        exit(0);
      }

      fmt = ((config->log_config[main_config_index]->num[dsp_sample_bits] == 16) ? AFMT_S16_LE : AFMT_U8);
      if (ioctl(fd, SNDCTL_DSP_SETFMT, &fmt) == -1)
      {
        asprintf(&z,"ioctl: %s\n",strerror(errno));
        print_log(MSG_ERROR,z);
        nfree(z);
        exit(0);
      }

      if (fmt != ((config->log_config[main_config_index]->num[dsp_sample_bits] == 16) ? AFMT_S16_LE : AFMT_U8))
      {
        asprintf(&z,"Device %s does not support the selected mode.\n",config->log_config[main_config_index]->str[input_source]);
        print_log(MSG_ERROR,z);
        nfree(z);
        exit(0);
      }

      ch = config->log_config[main_config_index]->num[dsp_input_channels];
      if (ioctl(fd, SNDCTL_DSP_CHANNELS, &ch) == -1)
      {
        asprintf(&z,"ioctl: %s\n",strerror(errno));
        print_log(MSG_ERROR,z);
        nfree(z);
        exit(0);
      }

      if (ch != config->log_config[main_config_index]->num[dsp_input_channels])
      {
        asprintf(&z,"Opening DSP device: Requested %d channels, but got %d\n",
                       config->log_config[main_config_index]->num[dsp_input_channels],
                       ch);
        print_log(MSG_ERROR,z);
        nfree(z);
        exit(0);
      }

      sr = config->log_config[main_config_index]->num[dsp_sample_rate];
      if (ioctl(fd, SNDCTL_DSP_SPEED, &sr) == -1)
      {
        asprintf(&z,"ioctl: %s\n",strerror(errno));
        print_log(MSG_ERROR,z);
        nfree(z);
        exit(0);
      }
      dsp_init_ok = 1;
    }

    if ((config->log_config[main_config_index]->str_num[stats] == yes) && (config->log_config[main_config_index]->str_num_is_set[stats]))
    {
      memset(&s_stats,0,sizeof(struct poc_stats));
      memset(&t_stats,0,sizeof(struct poc_stats));

      s_stats.s.first_run     = time(NULL);
      t_stats.s.first_run     = time(NULL);

      s_stats.s.started       = time(NULL);
      t_stats.s.started       = time(NULL);
      update_stats_structs();
    }

    if (((config->log_config[main_config_index]->str_num[input_mode] == tcp_pcm) && (config->log_config[main_config_index]->str_num_is_set[input_mode]))
       ||
       ((config->log_config[main_config_index]->str_num[input_mode] == dsp_pcm) && (config->log_config[main_config_index]->str_num_is_set[input_mode])))
    {
      pa_param = nmalloc(sizeof(struct pa_parameters));
      pa_param->init = 1;
      pa_param->sample_rate = config->log_config[main_config_index]->num[dsp_sample_rate];
      pa_param->bits        = config->log_config[main_config_index]->num[dsp_sample_bits];
      pa_param->channels    = config->log_config[main_config_index]->num[dsp_input_channels];
      pa_param->length = 0;
      pa_param->num_chan    = config->log_config[main_config_index]->num[dsp_decode_channels];
      pa_param->lw          = config->log_config[main_config_index]->num[dsp_left_weight];
      pa_param->rw          = config->log_config[main_config_index]->num[dsp_right_weight];
      pa_param->str         = 0;
      pa_param->threshold   = config->log_config[main_config_index]->num[noise_smoothing_threshold];
      pa_param->smoothing   = config->log_config[main_config_index]->str_num[noise_smoothing];
      pa_param->highest_bitrate = config->log_config[main_config_index]->num_list[bitrates]->list[config->log_config[main_config_index]->num_list[bitrates]->count - 1];
      pa_param->num_list = config->log_config[main_config_index]->num_list;
      process_audio((void *)pa_param);
      pa_param->init = 0;

      pthread_mutex_init(&audio_mutex, NULL);
      pthread_cond_init (&audio_cond, NULL);
      pthread_attr_init(&audio_attr);
      pthread_attr_setdetachstate(&audio_attr, PTHREAD_CREATE_DETACHED);
      pthread_create(&audio_thread, &audio_attr, (void *)process_audio, (void *)pa_param);

      if ((config->log_config[main_config_index]->str_num[pcm_server] == yes) && (config->log_config[main_config_index]->str_num_is_set[pcm_server]))
      {
        pthread_cond_init (&pcm_server_cond,  NULL);
        pcm_server_parameters.id_string              = "PCM";
        pcm_server_parameters.password               = config->log_config[main_config_index]->str[pcm_server_password];
        pcm_server_parameters.server_port            = config->log_config[main_config_index]->num[pcm_server_port];
        pcm_server_parameters.build_header_string    = &pcm_build_header_string;
        pcm_server_parameters.header_string          = &pcm_server_header_string;
        pcm_server_parameters.server_limit           = PCM_SERVER_LIMIT;
        pcm_server_parameters.server_timeout         = PCM_SERVER_TIMEOUT;
        pcm_server_parameters.connections            = &pcm_connections;
        pcm_server_parameters.server_writable        = &pcm_server_writable;
        pcm_server_parameters.server_readable        = &pcm_server_readable;
        pcm_server_parameters.server_writable_new    = &pcm_server_writable_new;
        pcm_server_parameters.server_readable_new    = &pcm_server_readable_new;
        pcm_server_parameters.server_new_connections = &pcm_server_new_connections;
        pcm_server_parameters.server_mutex           = &pcm_server_mutex;
        pcm_server_parameters.server_mutex2          = &pcm_server_mutex2;
        pcm_server_parameters.server_auth            = &pcm_server_auth;
        pcm_server_parameters.server_cond            = &pcm_server_cond;
        pcm_server_parameters.queue                  = pcm_server_audio_queue;

        pthread_attr_init(&pcm_server_attr);
        pthread_attr_setdetachstate(&pcm_server_attr, PTHREAD_CREATE_DETACHED);

        pthread_attr_init(&pcm_server_listen_attr);
        pthread_attr_setdetachstate(&pcm_server_listen_attr, PTHREAD_CREATE_DETACHED);

        pthread_create(&pcm_server_listen_thread, &pcm_server_listen_attr, (void *)server_listen, (void *)&pcm_server_parameters);
        pthread_create(&pcm_server_thread,        &pcm_server_attr,        (void *)server, (void *)&pcm_server_parameters);
      }
    }

    if ((config->log_config[main_config_index]->str_num[raw_server] == yes) && (config->log_config[main_config_index]->str_num_is_set[raw_server]))
    {
      pthread_cond_init (&raw_server_cond,  NULL);
      raw_server_parameters.id_string              = "RAW";
      raw_server_parameters.password               = config->log_config[main_config_index]->str[raw_server_password];
      raw_server_parameters.server_port            = config->log_config[main_config_index]->num[raw_server_port];
      raw_server_parameters.build_header_string    = &raw_build_header_string;
      raw_server_parameters.header_string          = &raw_server_header_string;
      raw_server_parameters.server_limit           = RAW_SERVER_LIMIT;
      raw_server_parameters.server_timeout         = RAW_SERVER_TIMEOUT;
      raw_server_parameters.connections            = &raw_connections;
      raw_server_parameters.server_writable        = &raw_server_writable;
      raw_server_parameters.server_readable        = &raw_server_readable;
      raw_server_parameters.server_writable_new    = &raw_server_writable_new;
      raw_server_parameters.server_readable_new    = &raw_server_readable_new;
      raw_server_parameters.server_new_connections = &raw_server_new_connections;
      raw_server_parameters.server_mutex           = &raw_server_mutex;
      raw_server_parameters.server_mutex2          = &raw_server_mutex2;
      raw_server_parameters.server_auth            = &raw_server_auth;
      raw_server_parameters.server_cond            = &raw_server_cond;
      raw_server_parameters.queue                  = raw_server_queue;

      pthread_attr_init(&raw_server_attr);
      pthread_attr_setdetachstate(&raw_server_attr, PTHREAD_CREATE_DETACHED);

      pthread_attr_init(&raw_server_listen_attr);
      pthread_attr_setdetachstate(&raw_server_listen_attr, PTHREAD_CREATE_DETACHED);

      pthread_create(&raw_server_listen_thread, &raw_server_listen_attr, (void *)server_listen, (void *)&raw_server_parameters);
      pthread_create(&raw_server_thread,        &raw_server_attr,        (void *)server, (void *)&raw_server_parameters);
    }

    pthread_mutex_init(&log_mutex, NULL);
    pthread_cond_init (&log_cond, NULL);
    pthread_attr_init(&log_attr);
    pthread_attr_setdetachstate(&log_attr, PTHREAD_CREATE_DETACHED);
    pthread_create(&log_thread, &log_attr, (void *)message_log, (void *)0);

    asprintf(&z,"%s successfully launched. Input mode: %s Input source: %s\n",program_string,config_str_keyword_strings[config->log_config[main_config_index]->str_num[input_mode]],config->log_config[main_config_index]->str[input_source]);
    print_log(MSG_NOTICE,z);
    nfree(z);

    if ((dsp_init_ok) && ((config->log_config[main_config_index]->str_num[input_mode] == dsp_pcm) && (config->log_config[main_config_index]->str_num_is_set[input_mode])))
    {
      while (1)
      {
        if (config->log_config[main_config_index]->num[dsp_buffer_length_bytes]*audio_queue->count < config->log_config[main_config_index]->num[dsp_max_queue_length_bytes])
        {
          queue_full = 0;

          audio = nmalloc(sizeof(struct pcm_raw_chunk));
          audio->ptr = nmalloc(config->log_config[main_config_index]->num[dsp_buffer_length_bytes]);
          audio->length       = read(fd,audio->ptr,config->log_config[main_config_index]->num[dsp_buffer_length_bytes]);
          audio->samplerate   = config->log_config[main_config_index]->num[dsp_sample_rate];
          audio->bits         = config->log_config[main_config_index]->num[dsp_sample_bits];
          audio->channels     = config->log_config[main_config_index]->num[dsp_input_channels];
          audio->bitrate_list = config->log_config[main_config_index]->num_list[bitrates];

          pthread_mutex_lock(&pcm_server_mutex2);
          if (pcm_server_header_string)
          {
            nfree(pcm_server_header_string);
            pcm_server_header_string = NULL;
          }
          pcm_server_header_string = pcm_build_header_string(audio);
          pthread_mutex_unlock(&pcm_server_mutex2);

          pthread_mutex_lock(&raw_server_mutex2);
          if (raw_server_header_string)
          {
            nfree(raw_server_header_string);
            raw_server_header_string = NULL;
          }

          raw_server_header_string = raw_build_header_string(audio);
          pthread_mutex_unlock(&raw_server_mutex2);

          if ((config->log_config[main_config_index]->str_num[pcm_server] == yes) && (config->log_config[main_config_index]->str_num_is_set[pcm_server]))
          {
            if (pcm_server_audio_queue->count < 20)
            {
              queue_full = 0;
              pcm_server_chunk             = nmalloc(sizeof(struct pcm_raw_chunk));
              pcm_server_chunk->length     = audio->length;
              pcm_server_chunk->samplerate = config->log_config[main_config_index]->num[dsp_sample_rate];
              pcm_server_chunk->bits       = config->log_config[main_config_index]->num[dsp_sample_bits];
              pcm_server_chunk->channels   = config->log_config[main_config_index]->num[dsp_input_channels];
              pcm_server_chunk->ptr        = nmalloc(audio->length);
              memcpy(pcm_server_chunk->ptr,audio->ptr,audio->length);
              pthread_mutex_lock(&pcm_server_mutex);
              nenqueue(pcm_server_audio_queue, pcm_server_chunk);
              pthread_cond_signal(&pcm_server_cond);
              pthread_mutex_unlock(&pcm_server_mutex);
            } else
            {
              queue_full++;
              if (queue_full<10)
              {
                print_log(MSG_WARNING,"PCM server output queue is full. Possible reasons: Server thread died, or computer is too slow.\n");
                print_log(MSG_WARNING,"Sleeping for 2 seconds to let it catch up.\n");
                sleep(2);
              } else
              {
                print_log(MSG_ERROR,"PCM server output queue is not emptying quickly enough. Something is wrong. Exiting.\n");
                exit(0);
              }
            }
          }

          if ((config->log_config[main_config_index]->str_num[stats] == yes) && (config->log_config[main_config_index]->str_num_is_set[stats]))
          {
            pthread_mutex_lock(&stats_mutex);
            pcm_byte_count += audio->length;
            while (pcm_byte_count >= (1024*1024))
            {
              s_stats.s.pcm_mbyte += 1;
              t_stats.s.pcm_mbyte += 1;
              pcm_byte_count      -= (1024*1024);
            }
            if (t_stats.s.pcm_mbyte>0)
            {
              stats_add_to_file(statsfile,&t_stats);
            }
            pthread_mutex_unlock(&stats_mutex);
          }
          pthread_mutex_lock(&audio_mutex);
          nenqueue(audio_queue, audio);
          pthread_cond_signal(&audio_cond);
          pthread_mutex_unlock(&audio_mutex);
        } else
        {
          queue_full++;
          if (queue_full<10)
          {
            print_log(MSG_WARNING,"Audio output queue is full. Possible reasons: Decoding thread died, or computer is too slow.\n");
            print_log(MSG_WARNING,"Sleeping for 2 seconds to let it catch up.\n");
            sleep(2);
          } else
          {
            print_log(MSG_ERROR,"Audio output queue is not emptying quickly enough. Something is wrong. Exiting.\n");
            exit(0);
          }
        }
      }
    }

    if ((config->log_config[main_config_index]->str_num[input_mode] == tcp_pcm) && (config->log_config[main_config_index]->str_num_is_set[input_mode]))
    {
      pcm.host = config->log_config[main_config_index]->str[input_source];
      pcm.password = config->log_config[main_config_index]->str[input_password];
      if (pcm.host)
      {
        while (1)
        {
          if (config->log_config[main_config_index]->num[dsp_buffer_length_bytes]*audio_queue->count < config->log_config[main_config_index]->num[dsp_max_queue_length_bytes])
          {
            tcp_input(&pcm,(pcm.connected ? 102400:0),&pcm_header_string_match_parse);
            if (pcm.connected)
            {
              if (pcm.msg_connected == 0)
              {
                asprintf(&z,"Connection to %s successful. Remote host stream information: %dhz %d bits %d channels\n",pcm.host,pcm.samplerate,pcm.bits,pcm.channels);
                print_log(MSG_NOTICE,z);
                nfree(z);
                pcm.msg_connected = 1;
              }
              if (pcm.length>0)
              {
                audio               = nmalloc(sizeof(struct pcm_raw_chunk));
                memset(audio,0,sizeof(struct pcm_raw_chunk));
                audio->bitrate_list = config->log_config[main_config_index]->num_list[bitrates];
                audio->length       = pcm.length;
                audio->samplerate   = pcm.samplerate;
                audio->bits         = pcm.bits;
                audio->channels     = pcm.channels;
                audio->ptr          = (unsigned char*)pcm.buf;

                pthread_mutex_lock(&pcm_server_mutex2);
                if (pcm_server_header_string)
                {
                  nfree(pcm_server_header_string);
                  pcm_server_header_string = NULL;
                }

                pcm_server_header_string = pcm_build_header_string(audio);
                pthread_mutex_unlock(&pcm_server_mutex2);

                pthread_mutex_lock(&raw_server_mutex2);
                if (raw_server_header_string)
                {
                  nfree(raw_server_header_string);
                  raw_server_header_string = NULL;
                }

                raw_server_header_string = raw_build_header_string(audio);
                pthread_mutex_unlock(&raw_server_mutex2);

                if ((config->log_config[main_config_index]->str_num[pcm_server] == yes) && (config->log_config[main_config_index]->str_num_is_set[pcm_server]))
                {
                  if (pcm_server_audio_queue->count < 20)
                  {
                    queue_full = 0;
                    pcm_server_chunk             = nmalloc(sizeof(struct pcm_raw_chunk));
                    memset(pcm_server_chunk,0,sizeof(struct pcm_raw_chunk));
                    pcm_server_chunk->length     = pcm.length;
                    pcm_server_chunk->samplerate = pcm.samplerate;
                    pcm_server_chunk->bits       = pcm.bits;
                    pcm_server_chunk->channels   = pcm.channels;
                    pcm_server_chunk->ptr        = nmalloc(audio->length);
                    memcpy(pcm_server_chunk->ptr,audio->ptr,audio->length);
                    pthread_mutex_lock(&pcm_server_mutex);
                    nenqueue(pcm_server_audio_queue, pcm_server_chunk);
                    pthread_cond_signal(&pcm_server_cond);
                    pthread_mutex_unlock(&pcm_server_mutex);
                  } else
                  {
                    queue_full++;
                    if (queue_full<10)
                    {
                      print_log(MSG_WARNING,"PCM server output queue is full. Possible reasons: Server thread died, or computer is too slow.\n");
                      print_log(MSG_WARNING,"Sleeping for 2 seconds to let it catch up.\n");
                      sleep(2);
                    } else
                    {
                      print_log(MSG_ERROR,"PCM server output queue is not emptying quickly enough. Something is wrong. Exiting.\n");
                      exit(0);
                    }
                  }
                }

                if ((config->log_config[main_config_index]->str_num[stats] == yes) && (config->log_config[main_config_index]->str_num_is_set[stats]))
                {
                  pthread_mutex_lock(&stats_mutex);
                  pcm_byte_count += audio->length;

                  while (pcm_byte_count >= (1024*1024))
                  {
                    s_stats.s.pcm_mbyte += 1;
                    t_stats.s.pcm_mbyte += 1;
                    pcm_byte_count      -= (1024*1024);
                  }

                  if (t_stats.s.pcm_mbyte>0)
                  {
                    stats_add_to_file(statsfile,&t_stats);
                  }
                  pthread_mutex_unlock(&stats_mutex);
                }
                pthread_mutex_lock(&audio_mutex);
                nenqueue(audio_queue, audio);
                pthread_cond_signal(&audio_cond);
                pthread_mutex_unlock(&audio_mutex);
              }
            } else
            {
              pcm.msg_connected = 0;
              asprintf(&z,"Connection to %s failed. Sleeping...\n",pcm.host);
              print_log(MSG_WARNING,z);
              nfree(z);
              sleep(5);
            }
          } else
          {
            queue_full++;
            if (queue_full<10)
            {
              print_log(MSG_WARNING,"Audio output queue is full. Possible reasons: Decoding thread died, or computer is too slow.\n");
              print_log(MSG_WARNING,"Sleeping for 2 seconds to let it catch up.\n");
              sleep(2);
            } else
            {
              print_log(MSG_ERROR,"Audio output queue is not emptying quickly enough. Something is wrong. Exiting.\n");
              exit(0);
            }
            pthread_mutex_lock(&audio_mutex);
            pthread_cond_signal(&audio_cond);
            pthread_mutex_unlock(&audio_mutex);
          }
        }
      } else
      {
        print_log(MSG_ERROR,"TCP-PCM input mode selected, but no hostname has been configured.\n");
        exit(0);
      }
    }

    if ((config->log_config[main_config_index]->str_num[input_mode] == tcp_raw) && (config->log_config[main_config_index]->str_num_is_set[input_mode]))
    {
      raw_s.host = config->log_config[main_config_index]->str[input_source];
      raw_s.password = config->log_config[main_config_index]->str[input_password];
      if (raw_s.host)
      {
        while (1)
        {
          tcp_input(&raw_s,(raw_s.connected ? sizeof(struct raw_server_chunk):0),&raw_header_string_match_parse);
          if (raw_s.connected)
          {
            if (raw_s.msg_connected == 0)
            {
              zz = nmalloc(5000);
              memset(zz,0,5000);

              for (i=0;i<raw_s.bitrate_list->count;i++)
              {
                asprintf(&z,"%d ",raw_s.bitrate_list->list[i]);
                strcat(zz,z);
                nfree(z);
              }

              asprintf(&z,"Connection to %s successful. Remote host stream information: Channels: %d  Bitrates: %sbps\n",raw_s.host,raw_s.channels,zz);
              print_log(MSG_NOTICE,z);
              nfree(z);
              nfree(zz);
              raw_s.msg_connected = 1;

              pthread_mutex_lock(&raw_server_mutex2);
              if (raw_server_header_string)
              {
                nfree(raw_server_header_string);
                raw_server_header_string = NULL;
              }

              raw_server_header_string = raw_s.header_string;
              raw_s.header_string = NULL;
              pthread_mutex_unlock(&raw_server_mutex2);
            }

            if (raw_s.length>0)
            {
              pthread_mutex_lock(&raw_server_mutex2);
              if (raw_s.bitrate_list->count == config->log_config[main_config_index]->num_list[bitrates]->count)
              {
                memcpy(config->log_config[main_config_index]->num_list[bitrates]->list,raw_s.bitrate_list->list,sizeof(uint32_t)*raw_s.bitrate_list->count);
              }
              pthread_mutex_unlock(&raw_server_mutex2);

              if ((raw_s.channels != config->log_config[main_config_index]->num[dsp_decode_channels])
                 ||
                 (raw_s.bitrate_list->count != config->log_config[main_config_index]->num_list[bitrates]->count))
              {

                pthread_mutex_lock(&raw_server_mutex2);

                for (i=0;i<config->log_config[main_config_index]->num[dsp_decode_channels];i++)
                {
                  nfree(decode_buffer[i]);
                }

                nfree(decode_buffer);
                decode_buffer = nmalloc(raw_s.channels*sizeof(uint32_t *));
                for (i=0;i<raw_s.channels;i++)
                {
                  decode_buffer[i] = nmalloc(raw_s.bitrate_list->count*sizeof(struct message_block));
                  memset(decode_buffer[i],0,raw_s.bitrate_list->count*sizeof(struct message_block));
                }

                if (config->log_config[main_config_index]->num_list[bitrates]->list)
                {
                  nfree(config->log_config[main_config_index]->num_list[bitrates]->list);
                }

                config->log_config[main_config_index]->num_list[bitrates]->list = nmalloc(sizeof(uint32_t)*raw_s.bitrate_list->count);
                memcpy(config->log_config[main_config_index]->num_list[bitrates]->list,raw_s.bitrate_list->list,sizeof(uint32_t)*raw_s.bitrate_list->count);

                config->log_config[main_config_index]->num_list[bitrates]->count = raw_s.bitrate_list->count;
                config->log_config[main_config_index]->num[dsp_decode_channels]  = raw_s.channels;
                update_stats_structs();
                pthread_mutex_unlock(&raw_server_mutex2);
              }

              if (raw_chunk == NULL)
              {
                raw_chunk = nmalloc(sizeof(struct raw_server_chunk));
              }

              memcpy(raw_chunk,raw_s.buf,sizeof(struct raw_server_chunk));

              for (i=0;i<(sizeof(struct raw_server_chunk)/(sizeof(uint32_t)));i++)
              {
                ntohl_ptr(((uint32_t*)raw_chunk + i));
              }

              decode_batch((uint32_t*)&raw_chunk->codewords, raw_chunk->channel, raw_chunk->bitrate, raw_chunk->decode_now, config->log_config[main_config_index]->num_list,raw_chunk->unixtime);

              if ((config->log_config[main_config_index]->str_num[raw_server] == yes) && (config->log_config[main_config_index]->str_num_is_set[raw_server]))
              {
                if (raw_server_queue->count < 20)
                {
                  queue_full = 0;
                  raw_server_chunk = nmalloc(sizeof(struct pcm_raw_chunk));
                  memset(raw_server_chunk,0,sizeof(struct pcm_raw_chunk));
                  raw_server_chunk->length = sizeof(struct raw_server_chunk);
                  raw_server_chunk->channels     = config->log_config[main_config_index]->num[dsp_decode_channels];
                  raw_server_chunk->bitrate_list = config->log_config[main_config_index]->num_list[bitrates];
                  raw_server_chunk->ptr          = (unsigned char*)raw_s.buf;

                  pthread_mutex_lock(&raw_server_mutex);
                  nenqueue(raw_server_queue, raw_server_chunk);
                  pthread_cond_signal(&raw_server_cond);
                  pthread_mutex_unlock(&raw_server_mutex);
                } else
                {
                  queue_full++;
                  if (queue_full<10)
                  {
                    print_log(MSG_WARNING,"RAW server output queue is full. Possible reasons: Server thread died, or computer is too slow.\n");
                    print_log(MSG_WARNING,"Sleeping for 2 seconds to let it catch up.\n");
                    sleep(2);
                  } else
                  {
                    print_log(MSG_ERROR,"RAW server output queue is not emptying quickly enough. Something is wrong. Exiting.\n");
                    exit(0);
                  }
                }
              } else
              {
                nfree(raw_s.buf);
              }
            }
          } else
          {
            raw_s.msg_connected = 0;
            asprintf(&z,"Connection to %s failed. Sleeping...\n",raw_s.host);
            print_log(MSG_WARNING,z);
            nfree(z);
            sleep(5);
           }
        }
      } else
      {
        print_log(MSG_ERROR,"TCP-RAW input mode selected, but no hostname has been configured.\n");
        exit(0);
      }
    }

    if ((config->log_config[main_config_index]->str_num[input_mode] == file_raw) &&  (config->log_config[main_config_index]->str_num_is_set[input_mode]))
    {
      config->log_config[main_config_index]->str_num[raw_logging] = no;
      use_exit_flag = 1;
      file_list = glob_files(config->log_config[main_config_index]->str[input_source]);
      for (i=0;i<file_list->count;i++)
      {
        if (!(file = fopen(file_list->str[i],"r")) == (long int)NULL)
        {
          asprintf(&z,"Opened file %s\n",file_list->str[i]);
          print_log(MSG_NOTICE,z);
          nfree(z);
          tmp = malloc(strlen(raw_header) + 1);
          memset(tmp,0,strlen(raw_header) + 1);
          j = fread(tmp,strlen(raw_header), 1, file);
          if (j)
          {
            raw = NULL;
            if (strcmp(tmp,(char *)&raw_header) == 0)
            {
              nfree(tmp);
              raw_files_read++;
              raw = malloc(sizeof(struct pocraw)*BATCHES);
              memset(raw,0,sizeof(struct pocraw)*BATCHES);
              m = 0;
              while ((j = fread(raw,sizeof(struct pocraw),BATCHES,file))>0)
              { m++;
                for (k=0;k<(j*sizeof(struct pocraw))/sizeof(uint32_t);k++)
                {
                  ntohl_ptr(((uint32_t *)raw) + k);
                }
                checksum = 0;
                for (k=0;k<j;k++)
                {
                  checksum = (*(raw + k)).unixtime + (*(raw + k)).infoblock;
                  for (l=0;l<16;l++)
                  {
                    checksum += (*(raw + k)).cw[l];
                  }

                  found = 0;
                  bitrate =  ((*(raw + k)).infoblock & 0xFFFF);
                  channel = (((*(raw + k)).infoblock & 0xF00000) >> 20);

                  if (checksum == (*(raw + k)).checksum)
                  {
                    for (l=0;l<config->log_config[main_config_index]->num_list[bitrates]->count;l++)
                    {
                      if (config->log_config[main_config_index]->num_list[bitrates]->list[l] == bitrate)
                      {
                        bitrate_index = l;
                        found = 1;
                      }
                    }

                    if (found == 0)
                    {
                      if (config->log_config[main_config_index]->num_list[bitrates]->list == NULL)
                      {
                        config->log_config[main_config_index]->num_list[bitrates]->list = nmalloc(sizeof(uint32_t));
                        config->log_config[main_config_index]->num_list[bitrates]->count = 1;
                        config->log_config[main_config_index]->num_list[bitrates]->list[config->log_config[main_config_index]->num_list[bitrates]->count - 1] = bitrate;
                      } else
                      {
                        config->log_config[main_config_index]->num_list[bitrates]->list = nrealloc(config->log_config[main_config_index]->num_list[bitrates]->list,sizeof(uint32_t)*++config->log_config[main_config_index]->num_list[bitrates]->count);
                        config->log_config[main_config_index]->num_list[bitrates]->list[config->log_config[main_config_index]->num_list[bitrates]->count - 1] = bitrate;
                      }

                      for (l=0;l<config->log_config[main_config_index]->num[dsp_decode_channels];l++)
                      {
                        if (decode_buffer[l] == NULL)
                        {
                          decode_buffer[l] = nmalloc(sizeof(struct message_block));
                          memset(decode_buffer[i],0,sizeof(struct message_block));
                        } else
                        {
                          decode_buffer[l] = nrealloc(decode_buffer[i],sizeof(struct message_block)*config->log_config[main_config_index]->num_list[bitrates]->count);
                          memset(&decode_buffer[i][config->log_config[main_config_index]->num_list[bitrates]->count - 1],0,sizeof(struct message_block));
                        }
                      }
                    }

                    while ((channel+1)>config->log_config[main_config_index]->num[dsp_decode_channels])
                    {
                      if (decode_buffer == NULL)
                      {
                        decode_buffer = nmalloc(++config->log_config[main_config_index]->num[dsp_decode_channels]*sizeof(struct message_block *));
                      } else
                      {
                        decode_buffer = nrealloc(decode_buffer,++config->log_config[main_config_index]->num[dsp_decode_channels]*sizeof(uint32_t *));
                      }
                      decode_buffer[config->log_config[main_config_index]->num[dsp_decode_channels] - 1] = nmalloc(config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(struct message_block));
                      memset(decode_buffer[config->log_config[main_config_index]->num[dsp_decode_channels] - 1],0,config->log_config[main_config_index]->num_list[bitrates]->count*sizeof(struct message_block));
                    }
                    bitrate_index = found ? bitrate_index: (config->log_config[main_config_index]->num_list[bitrates]->count - 1);
                    raw_batches++;

                    for (m=0;m<16;m++)
                    {
                      raw_cw_temp = (*(raw + k)).cw[m];

                      if (parity_ok(raw_cw_temp) == 0)
                      {
                        raw_parity_errors++;
                      }
                      ecc(&raw_cw_temp,&raw_cw_ecc_status);

                      if ((raw_cw_ecc_status) == -1)
                      {
                        raw_ecc_errors_uncorrectable++;
                      } else if (raw_cw_ecc_status>0)
                      {
                        raw_ecc_errors_correctable++;
                      }
                    }

                    if ((raw_timestamp > 0) && (raw_timestamp_found))
                    {
                      if  (raw_timestamp > (*(raw + k)).unixtime)
                      {
                        pthread_mutex_lock(&time_mutex);
                        current_ts = nmalloc(50);
                        old_ts     = nmalloc(50);
                        memset(current_ts,0,50);
                        memset(old_ts,0,50);
                        ts = &(*(raw + k)).unixtime;
                        tss = (time_t)(*ts & 0xFFFFFFFF);
                        now = localtime((time_t *)&tss);
                        strftime(current_ts,49,date_format_long,now);
                        ts = (uint32_t*)&raw_timestamp;
                        tss = (time_t)(*ts & 0xFFFFFFFF);
                        now = localtime((time_t *)&tss);
                        strftime(old_ts,49,date_format_long,now);
                        pthread_mutex_unlock(&time_mutex);
                        asprintf(&z,"Current batch timestamp (%s) is older than the last batch that was processed (%s). This should only happen when you load RAW files in the wrong order.\n",current_ts,old_ts);
                        nfree(current_ts);
                        nfree(old_ts);
                        print_log(MSG_ERROR,z);
                        nfree(z);
                        goto endit;
                      }
                    } else
                    {
                      raw_timestamp = (*(raw + k)).unixtime;
                      raw_timestamp_found = 1;
                    }
                    decode_batch((uint32_t*)&(*(raw + k)).cw,channel,bitrate_index,0,config->log_config[main_config_index]->num_list,(*(raw + k)).unixtime);
                  } else
                  {
                    asprintf(&z,"Checksum mismatch (%08X, expected %08X) in chunk #%d in file %s\n",checksum,(*(raw + k)).checksum,(m-1)*BATCHES + k+1,file_list->str[i]);
                    print_log(MSG_ERROR,z);
                    nfree(z);
                  }
                }
              }
              if (raw)
              {
                nfree(raw);
                raw = NULL;
              }
            } else
            {
              asprintf(&z,"Bad file header in %s\n",file_list->str[i]);
              print_log(MSG_ERROR,z);
              nfree(z);
            }
          }
        } else
        {
          asprintf(&z,"Couldn't open %s\n",file_list->str[i]);
          print_log(MSG_ERROR,z);
          nfree(z);
        }
        nfree(file_list->str[i]);
      }

      if (file_list->count == 0)
      {
        asprintf(&z,"No file matches: %s\n",config->log_config[main_config_index]->str[input_source]);
        print_log(MSG_ERROR,z);
        nfree(z);
        goto endit;
      }

      pthread_mutex_lock(&exit_mutex);
      exit_now = 1;
      pthread_cond_wait(&exit_cond,&exit_mutex);
      pthread_mutex_unlock(&exit_mutex);
      printf("RAW file input finished. Files read: %d  Batches: %d (%d codewords). %d codewords had ECC errors (%d correctable, %d uncorrectable), %d had parity errors.\n",raw_files_read,raw_batches,raw_batches*16,(raw_ecc_errors_correctable+raw_ecc_errors_uncorrectable),raw_ecc_errors_correctable,raw_ecc_errors_uncorrectable,raw_parity_errors);
      endit:;
    }
    nfree(file_list);
    file_list = NULL;
  } else
  {
    print_log(MSG_ERROR,"Error: Couldn't look up the \"main\" configuration index. This should never happen. Exiting.\n");
    exit(0);
  }
  return 1;
}

void decode(int32_t channel, int32_t samplerate, int32_t count, int32_t *list, struct integer_list **num_list)
{
  int32_t queue_full;
  int32_t preamble_cnt;
  int32_t i,j,k,l,x,tmp,z;
  struct raw_server_chunk *rs;
  struct pcm_raw_chunk *audio;

  rs = NULL;
  queue_full = 0;

  for (i=0;i<num_list[bitrates]->count;i++)
  {
    for (j=0;j<count;j++)
    {
      x = ((list[j] < 0 ? nearest(samplerate/num_list[bitrates]->list[i],0-list[j]) : nearest(samplerate/num_list[bitrates]->list[i],list[j]))/(samplerate/num_list[bitrates]->list[i]));

      if (x==0)
      {
        x++;
      }

      for (k=0;k<x;k++)
      {
        for (l=0;l<(CONFIG_LISTSIZE-1);l++)
        {
          buffer[channel][i][l] <<= 1;
          buffer[channel][i][l] += (buffer[channel][i][l+1] >> 31);
        }
        buffer[channel][i][CONFIG_LISTSIZE-1] <<= 1;
        buffer[channel][i][CONFIG_LISTSIZE-1] += (list[j] < 0 ? 0:1);
        bitcount[channel][i] = (bitcount[channel][i] > 0 ? --bitcount[channel][i] : 0);

        preamble_cnt = 0;

        if ((buffer[channel][i][0] == ~CONFIG_SYNC) || ((buffer[channel][i][0] == CONFIG_SYNC) && (bitcount[channel][i] == 0)))
        {
          bitcount[channel][i] = CONFIG_BATCH_LEN;
          for (tmp=0;tmp<16;tmp++)
          {
            cw_temp[tmp] = ((buffer[channel][i][0] == CONFIG_SYNC) ? buffer[channel][i][tmp+1]:~buffer[channel][i][tmp+1]);
          }

          if ((config->log_config[main_config_index]->str_num[stats] == yes) && (config->log_config[main_config_index]->str_num_is_set[stats]))
          {
            pthread_mutex_lock(&stats_mutex);
            stats_add(&s_stats,num_list[bitrates]->list[i],channel,0,0,0,0,checksum_ecc_ok(cw_temp,16),16-checksum_ecc_ok(cw_temp,16));
            stats_add(&t_stats,num_list[bitrates]->list[i],channel,0,0,0,0,checksum_ecc_ok(cw_temp,16),16-checksum_ecc_ok(cw_temp,16));
            stats_add_to_file(statsfile,&t_stats);
            pthread_mutex_unlock(&stats_mutex);
          }
/********************/
          if ((config->log_config[main_config_index]->str_num[raw_server] == yes) && (config->log_config[main_config_index]->str_num_is_set[raw_server]))
          {
            if (raw_server_queue->count<20)
            {
              queue_full = 0;
              rs = nmalloc(sizeof(struct raw_server_chunk));
              memset(rs,0,sizeof(struct raw_server_chunk));

              rs->channel         = channel;
              rs->unixtime        = time(NULL);
              rs->bitrate         = i;
              rs->decode_now      = (buffer[channel][i][CONFIG_LISTSIZE-1] == ((buffer[channel][i][0] == CONFIG_SYNC) ? CONFIG_SYNC:~CONFIG_SYNC) ? 0:1);

              memcpy(&rs->codewords,cw_temp,sizeof(uint32_t)*16);

              for (z=0;z<(sizeof(struct raw_server_chunk)/(sizeof(uint32_t)));z++)
              {
                htonl_ptr(((uint32_t*)rs + z));
              }

              audio               = nmalloc(sizeof(struct pcm_raw_chunk));
              audio->length       = sizeof(struct raw_server_chunk);
              audio->channels     = config->log_config[main_config_index]->num[dsp_decode_channels];
              audio->bitrate_list = config->log_config[main_config_index]->num_list[bitrates];
              audio->ptr          = (unsigned char*)rs;

              pthread_mutex_lock(&raw_server_mutex2);
              if (raw_server_header_string)
              {
                nfree(raw_server_header_string);
                raw_server_header_string = NULL;
              }

              raw_server_header_string = raw_build_header_string(audio);
              pthread_mutex_unlock(&raw_server_mutex2);

              pthread_mutex_lock(&raw_server_mutex);
              nenqueue(raw_server_queue, audio);
              pthread_cond_signal(&raw_server_cond);
              pthread_mutex_unlock(&raw_server_mutex);
            } else
            {
              queue_full++;
              if (queue_full<10)
              {
                print_log(MSG_WARNING,"RAW server output queue is full. Possible reasons: Server thread died, or computer is too slow.\n");
                print_log(MSG_WARNING,"Sleeping for 2 seconds to let it catch up.\n");
                sleep(2);
              } else
              {
                print_log(MSG_ERROR,"RAW server output queue is not emptying quickly enough. Something is wrong. Exiting.\n");
                exit(0);
              }
            }
          }
/****************************/
          decode_batch(cw_temp,channel,i,(buffer[channel][i][CONFIG_LISTSIZE-1] == ((buffer[channel][i][0] == CONFIG_SYNC) ? CONFIG_SYNC:~CONFIG_SYNC) ? 0:1),num_list,time(NULL));
        }
      }
    }
  }
}

struct msg_str *pocsag_string_decode(uint32_t *cw, int32_t count, int32_t function_bits, struct integer_list **num_list)
{
  /* Takes a list of raw POCSAG codewords and extracts the messages */

  int32_t i,y,z;
  int32_t char_bit_len;
  int32_t offset;
  int32_t msglen;
  char *zz;

  if ((get_int_index(num_list[alphanumeric_bits]->list,
                          num_list[alphanumeric_bits]->count,
                          function_bits,
                          &y,
                          &z) >= 0) ^
       (get_int_index(num_list[numeric_bits]->list,
                          num_list[numeric_bits]->count,
                          function_bits,
                          &y,
                          &z) >= 0))
  {
    ret = nmalloc(sizeof(struct msg_str));

    char_bit_len = (get_int_index(num_list[alphanumeric_bits]->list,num_list[alphanumeric_bits]->count,function_bits,&y,&z) >= 0) ? 7:4;
    offset       = (get_int_index(num_list[alphanumeric_bits]->list,num_list[alphanumeric_bits]->count,function_bits,&y,&z) >= 0) ? 24:27;
    msglen       = (get_int_index(num_list[alphanumeric_bits]->list,num_list[alphanumeric_bits]->count,function_bits,&y,&z) >= 0) ? (int32_t)ceil((double)count*20/7) : count*5;

    ret->msglen  = msglen;
    ret->str     = nmalloc(ret->msglen + 1);
    memset(ret->str,0,ret->msglen + 1);

    for (i=0;i<count*20;i++)
    {
      /* Copy the message from the incoming buffer of
         POCSAG codewords to the output buffer, one bit at a time. */
      ret->str[i/char_bit_len] = (ret->str[i/char_bit_len] >> 1) | (cw[i/20] & (1 << (30-(i % 20)))) >> (offset-(i % 20));
    }
    ret->str[ret->msglen] = 0;

    if (get_int_index(num_list[numeric_bits]->list,num_list[numeric_bits]->count,function_bits,&y,&z) >= 0)
    {
      for (i=0;i<count*5;i++)
      {
        ret->str[i] = num_charset[ret->str[i]];
      }
    }
    ret->ok = 1;
    return ret;
  } else
  {
    asprintf(&zz,"Couldn't determine message type for function %d.\n",function_bits);
    print_log(MSG_ERROR,zz);
    nfree(zz);
    if ((get_int_index(num_list[alphanumeric_bits]->list,num_list[alphanumeric_bits]->count,function_bits,&y,&z) == -1)
       &&
       ((get_int_index(num_list[numeric_bits]->list,num_list[numeric_bits]->count,function_bits,&y,&z) == -1)))
       {
         asprintf(&zz,"Couldn't find %d in either the alphanumeric nor numeric list. Please correct this configuration error.\n",function_bits);
         print_log(MSG_ERROR,zz);
         nfree(zz);
       } else if ((get_int_index(num_list[alphanumeric_bits]->list,num_list[alphanumeric_bits]->count,function_bits,&y,&z) >= 0)
                 &&
                 (get_int_index(num_list[numeric_bits]->list,num_list[numeric_bits]->count,function_bits,&y,&z) >= 0))
       {
         asprintf(&zz,"Reason: Function %d was found in both the alphanumeric and the numeric list. Please correct this configuration error.\n",function_bits);
         print_log(MSG_ERROR,zz);
         nfree(zz);
       }

    ret = nmalloc(sizeof(struct msg_str));
    ret->str = NULL;
    ret->ok = 0;
    return ret;
  }
}

void process_audio(void *parameters) /* Converts incoming PCM stream to a list of the distances (in samples) between
                                        the zero transitions. */
{
  struct pa_parameters *args = (struct pa_parameters *)parameters;
  static struct pa **snd;
  double threshold_val;
  int32_t ch,length,i;
  struct zc ns_zero_crossings_new, zero_crossings_new;
  unsigned char *str  = args->str;
  double threshold    = ((double)args->threshold)/100;

  static struct pointer_queue *pa_queue;
  static void **tmp_v;
  int32_t tmp_i;
  static int32_t smoothing;
  struct pcm_raw_chunk *ab;

  threshold_val = ((double)args->sample_rate/(double)args->highest_bitrate)*threshold;
  smoothing = args->smoothing;

  memset(&ns_zero_crossings_new,0,sizeof(struct zc));
  memset(&zero_crossings_new,0,sizeof(struct zc));

  if (args->init)
  {
    snd = nmalloc(sizeof(struct pa *)*args->channels);
    tmp_v = nmalloc(sizeof(void *));
    for (i=0;i<args->channels;i++)
    {
      snd[i] = nmalloc(sizeof(struct pa));
      memset(snd[i],0,sizeof(struct pa));
    }

    pa_queue = nmalloc(sizeof(struct pointer_queue));
    pa_queue->count = 0;
  } else
  {

    while (1)
    {
      pthread_mutex_lock(&audio_mutex);
      pthread_cond_wait(&audio_cond,&audio_mutex);
      while (audio_queue->count > 0)
      {
        nenqueue(pa_queue,ndequeue(audio_queue,&tmp_i,tmp_v));
      }
      pthread_mutex_unlock(&audio_mutex);
      while (pa_queue->count > 0)
      {
        ab = ndequeue(pa_queue,&tmp_i,tmp_v);
        str = ab->ptr;
        length = ab->length;

        if ((ab->samplerate != args->sample_rate) || (ab->bits != args->bits) || (ab->channels != args->channels))
        {
          for (i=0;i<args->channels;i++)
          {
            nfree(snd[i]);
          }
          nfree(snd);

          snd = nmalloc(sizeof(struct pa *)*args->channels);
          for (i=0;i<args->channels;i++)
          {
            snd[i] = nmalloc(sizeof(struct pa));
            memset(snd[i],0,sizeof(struct pa));
          }

          args->sample_rate = ab->samplerate;
          args->bits        = ab->bits;
          args->channels    = ab->channels;
        }

        for (ch=0;ch<args->num_chan;ch++)
        {
          zero_crossings_new.zc_ptr = nmalloc(sizeof(int32_t)*ZC_MEM_BLOCK_SIZE);
          zero_crossings_new.allocated = ZC_MEM_BLOCK_SIZE;
          zero_crossings_new.index = 0;

          for (i=0;i<((length)/(args->channels*args->bits/8));i++)
          {
            if ((args->bits == 8) && (args->channels == 1))
            {
              snd[ch]->sample = (int32_t)((char)str[i]);
            }
            if ((args->bits == 8) && (args->channels == 2))
            {
              if (args->num_chan == 1)
              {
                snd[ch]->sample = ((int32_t)(((char)str[2*i*(args->bits/8)])*((double)args->lw) + ((char)str[2*i*(args->bits/8) + 1])*((double)args->rw)));
              } else
              {
                snd[ch]->sample = ((int32_t)(((char)str[2*i*(args->bits/8) + ch])));
              }
            }
            if ((args->bits == 16) && (args->channels == 1))
            {
              snd[ch]->sample = (int32_t)*((int16_t *)str + i);
            }
            if ((args->bits == 16) && (args->channels == 2))
            {
              if (args->num_chan == 1)
              {
                snd[ch]->sample = (int32_t)(*((int16_t *)str + 2*i)*((double)args->lw) + *((int16_t *)str + 2*i + 1)*((double)args->rw));
              } else
              {
                snd[ch]->sample = (int32_t)*((int16_t *)str + 2*i + ch);
              }
            }

            snd[ch]->sample_count++;

            if (((snd[ch]->sample > 0) && (snd[ch]->last_sample <= 0)) || ((snd[ch]->sample < 0) && (snd[ch]->last_sample >= 0)))
            {
              if ((snd[ch]->sample > 0) && (snd[ch]->last_sample <= 0))
              {
                zero_crossings_new.zc_ptr[zero_crossings_new.index] = (0 - (int32_t)(snd[ch]->sample_count - snd[ch]->last_crossing));
              }
              else if ((snd[ch]->sample < 0) && (snd[ch]->last_sample >= 0))
              {
                zero_crossings_new.zc_ptr[zero_crossings_new.index] = (int32_t)(snd[ch]->sample_count - snd[ch]->last_crossing);
              }

              zero_crossings_new.index++;

              if (zero_crossings_new.index == zero_crossings_new.allocated)
              {
                zero_crossings_new.allocated += ZC_MEM_BLOCK_SIZE;
                zero_crossings_new.zc_ptr = nrealloc(zero_crossings_new.zc_ptr, (sizeof(int32_t)*zero_crossings_new.allocated));
              }

              snd[ch]->last_crossing = snd[ch]->sample_count;
            }

            if ((snd[ch]->sample_count > 0x80000000) && (snd[ch]->last_crossing > 0x80000000))
            {
              snd[ch]->sample_count = snd[ch]->sample_count - 0x80000000;
              snd[ch]->last_crossing = snd[ch]->last_crossing - 0x80000000;
            }
            snd[ch]->last_sample = snd[ch]->sample;
          }

          zero_crossings_new.count = zero_crossings_new.index;

          if ((threshold>0) && (smoothing == yes))
          {
            ns_zero_crossings_new.zc_ptr = nmalloc(sizeof(int32_t)*ZC_MEM_BLOCK_SIZE);
            ns_zero_crossings_new.allocated = ZC_MEM_BLOCK_SIZE;
            ns_zero_crossings_new.index = 0;

            ns_zero_crossings_new.zc_ptr[0] = snd[ch]->list_last;

            for (i=0;i<zero_crossings_new.count;i++)
            {
              snd[ch]->cont = 1;
              if (snd[ch]->flag)
              {
                if (ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] >= 0)
                {
                  ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] = ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] + (zero_crossings_new.zc_ptr[i] < 0 ? (0-zero_crossings_new.zc_ptr[i]) : zero_crossings_new.zc_ptr[i]);
                } else
                {
                  ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] = ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] - (zero_crossings_new.zc_ptr[i] < 0 ? (0-zero_crossings_new.zc_ptr[i]) : zero_crossings_new.zc_ptr[i]);
                }
                snd[ch]->flag = 0;
                snd[ch]->cont = 0;
              }

              if ((double)((zero_crossings_new.zc_ptr[i] < 0 ? (0-zero_crossings_new.zc_ptr[i]) : zero_crossings_new.zc_ptr[i]) <= threshold_val) && snd[ch]->cont)
              {
                if (ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] >= 0)
                {
                  ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] = ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] + (zero_crossings_new.zc_ptr[i] < 0 ? (0-zero_crossings_new.zc_ptr[i]) : zero_crossings_new.zc_ptr[i]);
                } else
                {
                  ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] = ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] - (zero_crossings_new.zc_ptr[i] < 0 ? (0-zero_crossings_new.zc_ptr[i]) : zero_crossings_new.zc_ptr[i]);
                }
                snd[ch]->flag = 1;
              } else if ((double)( (zero_crossings_new.zc_ptr[i] < 0 ? (0-zero_crossings_new.zc_ptr[i]) : zero_crossings_new.zc_ptr[i]) > threshold_val) && snd[ch]->cont)
              {
                ns_zero_crossings_new.index++;
                if (ns_zero_crossings_new.index == ns_zero_crossings_new.allocated)
                {
                  ns_zero_crossings_new.allocated += ZC_MEM_BLOCK_SIZE;
                  ns_zero_crossings_new.zc_ptr = nrealloc(ns_zero_crossings_new.zc_ptr,(sizeof(int32_t)*ns_zero_crossings_new.allocated));
                }
                ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index] = zero_crossings_new.zc_ptr[i];
              }
            }

            ns_zero_crossings_new.count = ns_zero_crossings_new.index;
            snd[ch]->list_last = ns_zero_crossings_new.zc_ptr[ns_zero_crossings_new.index];
            if(zero_crossings_new.zc_ptr)
            {
              nfree(zero_crossings_new.zc_ptr);
            }
            memcpy(&zero_crossings_new,&ns_zero_crossings_new,sizeof(struct zc));
          }

          if (zero_crossings_new.count == 0)
          {
            if (zero_crossings_new.zc_ptr)
            {
              nfree(zero_crossings_new.zc_ptr);
            }
            zero_crossings_new.zc_ptr = nmalloc(sizeof(int32_t));
            zero_crossings_new.zc_ptr[0] = 0;
            zero_crossings_new.index = 1;
          }

          decode(ch,args->sample_rate,zero_crossings_new.count,zero_crossings_new.zc_ptr,args->num_list);

          if (zero_crossings_new.zc_ptr)
          {
            nfree(zero_crossings_new.zc_ptr);
          }
          memset(&zero_crossings_new,0,sizeof(struct zc));
        }
      nfree(ab->ptr);
      nfree(ab);
      }
    }
  }
}

int32_t nearest(int32_t target, int32_t number)
{
  double half = 0.5;
  int32_t res;
  res = target * (int32_t)((number + half*target) / target);
  return res;
}

void decode_batch(uint32_t *cw, int32_t channel, int32_t bitrate, int32_t decode_now, struct integer_list **num_list, uint32_t unixtime)
{
  int32_t p,i,j,m;
  struct tm *now;
  uint32_t tmp,n,*r;
  struct message_block *msg_tmp;
  uint32_t codeword,cw_ecc;
  char *batch_cw_tmp,*tmp_str,*z;
  time_t tss;

  decode_buffer[channel][bitrate].msg               = NULL;
  decode_buffer[channel][bitrate].msg_fixed         = NULL;
  decode_buffer[channel][bitrate].msg_fixed_no_crlf = NULL;
  decode_buffer[channel][bitrate].msg_uncorrected   = NULL;

  if (config->log_config[main_config_index]->str_num_is_set[raw_logging] && (config->log_config[main_config_index]->str_num[raw_logging] == yes))
  {
    if (checksum_ecc_ok(cw,16) >= config->log_config[main_config_index]->num[raw_logging_error_threshold])
    {
      log_raw_data((char *)&pocsag_raw_log_ok,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],1,channel,cw,16,unixtime);
    } else
    {
      log_raw_data((char *)&pocsag_raw_log_err,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],1,channel,cw,16,unixtime);
    }
  }

  if (config->log_config[main_config_index]->str_num_is_set[print_codewords] && (config->log_config[main_config_index]->str_num[print_codewords] == yes))
  {
    batch_cw_tmp = nmalloc(1000);
    memset(batch_cw_tmp,0,1000);
    asprintf(&tmp_str,"POCSAG Codewords (CH %u) (Bitrate %u bps) ",channel,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate]);
    strcat(batch_cw_tmp,tmp_str);
    nfree(tmp_str);
    for (m=0;m<16;m++)
    {
      if (config->log_config[main_config_index]->str_num_is_set[print_codewords_eh] && (config->log_config[main_config_index]->str_num[print_codewords_eh] == yes))
      {
        codeword = cw[m];
        cw_ecc = 0;
        ecc(&codeword,&cw_ecc);
        switch (cw_ecc)
        {
          case -1:asprintf(&tmp_str,"%c%s%08X%c%s ",0x1B,"[1;31m",cw[m],0x1B,"[0m");
                  break;
          case 0: asprintf(&tmp_str,"%08X ",cw[m]);
                  break;
          case 1: asprintf(&tmp_str,"%c%s%08X%c%s ",0x1B,"[0;32m",cw[m],0x1B,"[0m");
                  break;
          case 2: asprintf(&tmp_str,"%c%s%08X%c%s ",0x1B,"[0;32m",cw[m],0x1B,"[0m");
                  break;
        }
        strcat(batch_cw_tmp,tmp_str);
        nfree(tmp_str);
      } else
      {
        asprintf(&tmp_str,"%08X ",cw[m]);
        strcat(batch_cw_tmp,tmp_str);
        nfree(tmp_str);
      }
    }

    if (batch_cw_tmp[strlen(batch_cw_tmp)-1] == 0x20)
    {
      batch_cw_tmp[strlen(batch_cw_tmp)-1] = 0;
    }

    asprintf(&z,"%s\n",batch_cw_tmp);
    print_log(MSG_NOTICE,z);
    nfree(batch_cw_tmp);
    nfree(z);
  }

  for (p=0;p<16;p++)
  {
    if ((cw[p] & 0x80000000) && decode_buffer[channel][bitrate].ric_found)
    {
      decode_buffer[channel][bitrate].cw_len++;
      decode_buffer[channel][bitrate].codewords = (decode_buffer[channel][bitrate].cw_len == 1) ? nmalloc(sizeof(uint32_t)):nrealloc(decode_buffer[channel][bitrate].codewords,sizeof(int32_t)*decode_buffer[channel][bitrate].cw_len);
      decode_buffer[channel][bitrate].codewords[decode_buffer[channel][bitrate].cw_len - 1] = cw[p];
    }

    if ((((cw[p] & 0x80000000) == 0) || ((p == 15) && (((int32_t)cw[p] & 0x80000000) != 0) && decode_now)) && (decode_buffer[channel][bitrate].ric_found))
    {
      decode_buffer[channel][bitrate].codewords_uncorrected = decode_buffer[channel][bitrate].cw_len ? nmalloc(sizeof(uint32_t)*decode_buffer[channel][bitrate].cw_len) : NULL;
      decode_buffer[channel][bitrate].codewords_ecc_status  = decode_buffer[channel][bitrate].cw_len ? nmalloc(sizeof(uint32_t)*decode_buffer[channel][bitrate].cw_len) : NULL;
      decode_buffer[channel][bitrate].codewords_ecc_flags   = decode_buffer[channel][bitrate].cw_len ? nmalloc(sizeof(uint32_t)*decode_buffer[channel][bitrate].cw_len) : NULL;

      if (decode_buffer[channel][bitrate].cw_len)
      {
        memcpy(decode_buffer[channel][bitrate].codewords_uncorrected,decode_buffer[channel][bitrate].codewords,sizeof(int32_t)*decode_buffer[channel][bitrate].cw_len);
      }

      for (i=0;i<decode_buffer[channel][bitrate].cw_len;i++)
      {
        ecc(&decode_buffer[channel][bitrate].codewords[i], &decode_buffer[channel][bitrate].codewords_ecc_status[i]);
      }

      decode_buffer[channel][bitrate].codewords_ecc_flags_str = nmalloc(decode_buffer[channel][bitrate].cw_len + 1);

      for (i=0;i<decode_buffer[channel][bitrate].cw_len;i++)
      {
        switch(decode_buffer[channel][bitrate].codewords_ecc_status[i])
        {
          case -1:decode_buffer[channel][bitrate].codewords_ecc_flags[i]     =  2;
                  decode_buffer[channel][bitrate].codewords_ecc_flags_str[i] = '2';
                  break;
          case 0: decode_buffer[channel][bitrate].codewords_ecc_flags[i]     =  0;
                  decode_buffer[channel][bitrate].codewords_ecc_flags_str[i] = '0';
                  break;
          case 1: decode_buffer[channel][bitrate].codewords_ecc_flags[i]     =  1;
                  decode_buffer[channel][bitrate].codewords_ecc_flags_str[i] = '1';
                  break;
          case 2: decode_buffer[channel][bitrate].codewords_ecc_flags[i] =      1;
                  decode_buffer[channel][bitrate].codewords_ecc_flags_str[i] = '1';
                  break;
        }
      }

      decode_buffer[channel][bitrate].codewords_ecc_flags_str[decode_buffer[channel][bitrate].cw_len] = 0;

      j = 0;

      for (i=0;i<decode_buffer[channel][bitrate].cw_len;i++)
      {
        j += decode_buffer[channel][bitrate].codewords_ecc_flags[i];
      }

      j += (decode_buffer[channel][bitrate].ric != decode_buffer[channel][bitrate].ric_uncorrected);

      decode_buffer[channel][bitrate].errored = (j > 0) ? 1:0;

      rv = pocsag_string_decode(decode_buffer[channel][bitrate].codewords,decode_buffer[channel][bitrate].cw_len,decode_buffer[channel][bitrate].function_bits,num_list);
      decode_buffer[channel][bitrate].msg_len = rv->msglen;
      decode_buffer[channel][bitrate].msg = rv->str;
      nfree(rv);
      rv = pocsag_string_decode(decode_buffer[channel][bitrate].codewords_uncorrected,decode_buffer[channel][bitrate].cw_len,decode_buffer[channel][bitrate].function_bits,num_list);
      decode_buffer[channel][bitrate].msg_uncorrected = rv->str;
      decode_buffer[channel][bitrate].ok = rv->ok;
      nfree(rv);

      if (decode_buffer[channel][bitrate].ok)
      {
        if (func(decode_buffer[channel][bitrate].function_bits,config->log_config[main_config_index]->num_list[alphanumeric_bits]->count,config->log_config[main_config_index]->num_list[alphanumeric_bits]->list))
        {
          if ((config->log_config[main_config_index]->str_num[stats] == yes) && (config->log_config[main_config_index]->str_num_is_set[stats]))
          {
            pthread_mutex_lock(&stats_mutex);
            if (decode_buffer[channel][bitrate].errored)
            {
              stats_add(&s_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,0,1,0,0,0,0);
              stats_add(&t_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,0,1,0,0,0,0);
              stats_add_to_file(statsfile,&t_stats);
            } else
            {
              stats_add(&s_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,1,0,0,0,0,0);
              stats_add(&t_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,1,0,0,0,0,0);

              stats_add_to_file(statsfile,&t_stats);
            }
            pthread_mutex_unlock(&stats_mutex);
          }
          decode_buffer[channel][bitrate].msg_fixed_no_crlf = translate(decode_buffer[channel][bitrate].msg,decode_buffer[channel][bitrate].msg_len,char_table_no_crlf);
          decode_buffer[channel][bitrate].msg_fixed = translate(decode_buffer[channel][bitrate].msg,decode_buffer[channel][bitrate].msg_len,char_table);
        } else
        {
          if ((config->log_config[main_config_index]->str_num[stats] == yes) && (config->log_config[main_config_index]->str_num_is_set[stats]))
          {
            pthread_mutex_lock(&stats_mutex);
            if (decode_buffer[channel][bitrate].errored)
            {
              stats_add(&s_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,0,0,0,1,0,0);
              stats_add(&t_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,0,0,0,1,0,0);
              stats_add_to_file(statsfile,&t_stats);
            } else
            {
              stats_add(&s_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,0,0,1,0,0,0);
              stats_add(&t_stats,config->log_config[main_config_index]->num_list[bitrates]->list[bitrate],channel,0,0,1,0,0,0);
              stats_add_to_file(statsfile,&t_stats);
            }
            pthread_mutex_unlock(&stats_mutex);
          }

          decode_buffer[channel][bitrate].msg_fixed_no_crlf = nmalloc(decode_buffer[channel][bitrate].msg_len + 1);
          decode_buffer[channel][bitrate].msg_fixed = nmalloc(decode_buffer[channel][bitrate].msg_len + 1);
          memcpy(decode_buffer[channel][bitrate].msg_fixed_no_crlf,decode_buffer[channel][bitrate].msg,decode_buffer[channel][bitrate].msg_len + 1);
          memcpy(decode_buffer[channel][bitrate].msg_fixed,decode_buffer[channel][bitrate].msg,decode_buffer[channel][bitrate].msg_len + 1);
        }

        decode_buffer[channel][bitrate].idle_prefixed_codewords_len             = decode_buffer[channel][bitrate].cw_len + (decode_buffer[channel][bitrate].ric & 7)*2 + 1;
        decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected_len = decode_buffer[channel][bitrate].cw_len + (decode_buffer[channel][bitrate].ric_uncorrected & 7)*2 + 1;
        decode_buffer[channel][bitrate].idle_prefixed_codewords                 = decode_buffer[channel][bitrate].idle_prefixed_codewords_len ? nmalloc(sizeof(uint32_t)*decode_buffer[channel][bitrate].idle_prefixed_codewords_len) : NULL;
        decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected     = decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected_len ? nmalloc(sizeof(uint32_t)*decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected_len) : NULL;
        if (decode_buffer[channel][bitrate].idle_prefixed_codewords_len)
        {
          memset(decode_buffer[channel][bitrate].idle_prefixed_codewords,0,sizeof(uint32_t)*decode_buffer[channel][bitrate].idle_prefixed_codewords_len);
        }
        if (decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected_len)
        {
          memset(decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected,0,sizeof(uint32_t)*decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected_len);
        }

        for(i=0;i<2*(decode_buffer[channel][bitrate].ric & 7);i++)
        {
          decode_buffer[channel][bitrate].idle_prefixed_codewords[i] = CONFIG_IDLE;
        }

        for(i=0;i<2*(decode_buffer[channel][bitrate].ric_uncorrected & 7);i++)
        {
          decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected[i] = CONFIG_IDLE;
        }

        memcpy(&decode_buffer[channel][bitrate].idle_prefixed_codewords[2*(decode_buffer[channel][bitrate].ric & 7)],
               &decode_buffer[channel][bitrate].ric_cw,sizeof(uint32_t));
        memcpy(&decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected[2*(decode_buffer[channel][bitrate].ric_uncorrected & 7)],
               &decode_buffer[channel][bitrate].ric_cw_uncorrected,sizeof(uint32_t));
        memcpy(&decode_buffer[channel][bitrate].idle_prefixed_codewords[2*(decode_buffer[channel][bitrate].ric & 7) + 1], decode_buffer[channel][bitrate].codewords,sizeof(uint32_t)*decode_buffer[channel][bitrate].cw_len);
        memcpy(&decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected[2*(decode_buffer[channel][bitrate].ric_uncorrected & 7) + 1], decode_buffer[channel][bitrate].codewords_uncorrected,sizeof(uint32_t)*decode_buffer[channel][bitrate].cw_len);

        for(i=0;i<decode_buffer[channel][bitrate].cw_len;i++)
        {
          htonl_ptr(&decode_buffer[channel][bitrate].codewords[i]);
          htonl_ptr(&decode_buffer[channel][bitrate].codewords_uncorrected[i]);
        }
        for(i=0;i<decode_buffer[channel][bitrate].idle_prefixed_codewords_len;i++)
        {
          htonl_ptr(&decode_buffer[channel][bitrate].idle_prefixed_codewords[i]);
        }
        for(i=0;i<decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected_len;i++)
        {
          htonl_ptr(&decode_buffer[channel][bitrate].idle_prefixed_codewords_uncorrected[i]);
        }

        decode_buffer[channel][bitrate].channel = channel;
        decode_buffer[channel][bitrate].bitrate = config->log_config[main_config_index]->num_list[bitrates]->list[bitrate];
        decode_buffer[channel][bitrate].date    = nmalloc(50);
        memset(decode_buffer[channel][bitrate].date,0,50);
        r = (uint32_t *)&unixtime;
        tss = (time_t)(*r & 0xFFFFFFFF);
        pthread_mutex_lock(&time_mutex);
        now = localtime((time_t *)&tss);
        strftime(decode_buffer[channel][bitrate].date,49,date_format_long,now);
        pthread_mutex_unlock(&time_mutex);
        decode_buffer[channel][bitrate].unixtime = unixtime;
        msg_tmp = nmalloc(sizeof(struct message_block));
        memcpy(msg_tmp,&decode_buffer[channel][bitrate],sizeof(struct message_block));
        pthread_mutex_lock(&log_mutex);
        nenqueue(log_queue, msg_tmp);
        pthread_cond_signal(&log_cond);
        pthread_mutex_unlock(&log_mutex);
      }
      decode_buffer[channel][bitrate].cw_len = 0;
      decode_buffer[channel][bitrate].ric_found = 0;
    }

    if ((cw[p] & 0x80000000) == 0)
    {
      tmp = cw[p];
      ecc(&tmp,&n);
      if (n==0)
      {
        if (!(parity_ok(tmp)))
        {
          tmp = parity(tmp);
        }
      }
      if ((tmp != cw[p]) && (tmp == CONFIG_IDLE))
      {
        cw[p] = tmp;
      }
      if (cw[p] != CONFIG_IDLE)
      {
        decode_buffer[channel][bitrate].ric_found = 1;
        decode_buffer[channel][bitrate].codewords = NULL;
        decode_buffer[channel][bitrate].function_bits = ((cw[p] & 0x00001800) >> 11);
        decode_buffer[channel][bitrate].ric_cw = cw[p];
        decode_buffer[channel][bitrate].ric_cw_uncorrected = cw[p];
        ecc(&decode_buffer[channel][bitrate].ric_cw, &decode_buffer[channel][bitrate].ric_cw_ecc_status);

        switch(decode_buffer[channel][bitrate].ric_cw_ecc_status)
        {
          case -1:decode_buffer[channel][bitrate].ric_status = 2;
                  break;
          case 0: decode_buffer[channel][bitrate].ric_status = 0;
                  break;
          case 1: decode_buffer[channel][bitrate].ric_status = 1;
                  break;
          case 2: decode_buffer[channel][bitrate].ric_status = 1;
                  break;
        }
        decode_buffer[channel][bitrate].ric             = (((decode_buffer[channel][bitrate].ric_cw & 0x7FFFE000) >> 10) | (p/2));
        decode_buffer[channel][bitrate].ric_uncorrected = (((decode_buffer[channel][bitrate].ric_cw_uncorrected & 0x7FFFE000) >> 10) | (p/2));
      } else
      {
        decode_buffer[channel][bitrate].ric_found = 0;
        decode_buffer[channel][bitrate].codewords = NULL;
      }
    }
  }
}


void ecc(uint32_t *cw, uint32_t *cnt)
{ /* Adapted from bch3121.c, written by Robert Morelos-Zaragoza */
  int32_t i, j, q, aux, elp[3], s[5], s3, loc[3],  reg[3];
  int32_t syn_error = 0;
  int32_t n = 31;
  int32_t length = 31;
  int32_t alpha_to[] = {1,2,4,8,16,5,10,20,13,26,17,7,14,28,29,31,27,19,3,6,12,24,21,15,30,25,23,11,22,9,18,0,0};
  int32_t index_of[] = {-1,0,1,18,2,5,19,11,3,29,6,27,20,8,12,23,4,10,30,17,7,22,28,26,21,25,9,16,13,14,24,15,1};

  *cnt = 0;

  for (i = 1; i <= 4; i++)
  {
    s[i] = 0;
    for (j = 0; j < length; j++)
    {
      if (*cw & (1 << (j + 1)))
      {
        s[i] ^= alpha_to[(i * j) % n];
      }
    }
    if (s[i] != 0)
    {
      syn_error = 1;
    }
    s[i] = index_of[s[i]];
  }
  if (syn_error)
  {
    if (s[1] != -1)
    {
      if (s[1] != -1)
      {
        s3 = (s[1] * 3) % n;
        if (s[3] == s3)
        {
          *cnt = 1;
          *cw ^= (1 << (s[1] + 1));
        }
        else
        {
          if (s[3] != -1)
          {
           aux = alpha_to[s3] ^ alpha_to[s[3]];
          }
          else
          {
            aux = alpha_to[s3];
          }
          elp[0] = 0;
          elp[1] = (s[2] - index_of[aux] + n) % n;
          elp[2] = (s[1] - index_of[aux] + n) % n;
          for (i = 1; i <= 2; i++)
          {
            reg[i] = elp[i];
          }
          *cnt = 0;
          for (i = 1; i <= n; i++)
          {
            q = 1;
            for (j = 1; j <= 2; j++)
            {
              if (reg[j] != -1)
              {
                reg[j] = (reg[j] + j) % n;
                q ^= alpha_to[reg[j]];
              }
            }
            if (!q)
            {
              loc[*cnt] = i % n;
              (*cnt)++;
            }
          }

          if (*cnt == 2)
          {
            for (i = 0; i < 2; i++)
            {
               *cw ^= (1 << (loc[i] + 1));
            }
          }
          else
          {
             *cnt = -1;
          }
        }
      }
      else if (s[2] != -1)
      {
        *cnt = -1;
      }
    }
  }
}

void build_character_tables(char *fn)
{
  char *buf,*strtemp,*z;
  FILE *fp;
  char *rc;
  uint32_t i,j,chr;

  buf = nmalloc(1000);
  memset(buf,0,1000);

  char_table              = nmalloc(256*sizeof(char *));
  char_table_no_crlf      = nmalloc(256*sizeof(char *));
  char_table_crlf_space   = nmalloc(256*sizeof(char *));
  char_table_mysql_escape = nmalloc(256*sizeof(char *));

  const char mysql_escape_from[]  = {0x00,0x0A,0x0D,0x1A,0x22,0x27,0x5C};
  const char mysql_escape_to[]    = {0x30,0x6E,0x72,0x5A,0x22,0x27,0x5C};

  for (i=0;i<256;i++)
  {
    char_table[i]                 = nmalloc(2);
    char_table[i][0]              = i;
    char_table[i][1]              = 0;
    char_table_no_crlf[i]         = nmalloc(2);
    char_table_no_crlf[i][0]      = i;
    char_table_no_crlf[i][1]      = 0;
    char_table_crlf_space[i]      = nmalloc(2);
    char_table_crlf_space[i][0]   = i;
    char_table_crlf_space[i][1]   = 0;
    char_table_mysql_escape[i]    = nmalloc(2);
    char_table_mysql_escape[i][0] = i;
    char_table_mysql_escape[i][1] = 0;
  }

  char_table[0][0] = 0x20;
  char_table[0][1] = 0;

  char_table_no_crlf[0][0] = 0x20;
  char_table_no_crlf[0][1] = 0;

  char_table_crlf_space[10][0] = 0x20;
  char_table_crlf_space[13][0] = 0x20;

  for (i=0;i<256;i++)
  {
    for (j=0;j<sizeof(mysql_escape_from)/sizeof(char);j++)
    {
      if (mysql_escape_from[j] == i)
      {
        char_table_mysql_escape[i % 256] = nrealloc(char_table_mysql_escape[i % 256],4);
        memset(char_table_mysql_escape[i % 256],0,4);
        strcat(char_table_mysql_escape[i % 256], "\\");
        strncat(char_table_mysql_escape[i % 256], &mysql_escape_to[j], 1);
      }
    }
  }

  if (!(fp = fopen(fn,"r")) == 0)
  {
    strtemp = nmalloc(50);
    memset(strtemp,0,50);

    while (!(rc = fgets(buf,500,fp)) == 0)
    {
      if ((buf[strlen(buf)-1] == 0x0A) || (buf[strlen(buf)-1] == 0x0D))
      {
        buf[strlen(buf)-1] = 0;
      }
      for (i=0;i<strlen(buf);i++)
      {
        if ((buf[strlen(buf)-i-1] == 0x0A) || (buf[strlen(buf)-i-1] == 0x0D))
        {
          buf[strlen(buf)-i-1] = 0;
        }
      }
      if (sscanf(buf,"0x%02X \"%[^\n\a\"]\"",&chr,strtemp) == 2)
      {
        char_table[chr % 256] = nrealloc(char_table[chr % 256],strlen(strtemp)+1);
        memset(char_table[chr % 256],0,strlen(strtemp)+1);
        strncat(char_table[chr % 256],strtemp,strlen(strtemp));

        if ((chr != 10) && (chr != 13))
        {
          char_table_no_crlf[chr % 256] = nrealloc(char_table_no_crlf[chr % 256],strlen(strtemp)+1);
          memset(char_table_no_crlf[chr % 256],0,strlen(strtemp)+1);
          strncat(char_table_no_crlf[chr % 256],strtemp,strlen(strtemp));

          char_table_crlf_space[chr % 256] = nrealloc(char_table_crlf_space[chr % 256],strlen(strtemp)+1);
          memset(char_table_crlf_space[chr % 256],0,strlen(strtemp)+1);
          strncat(char_table_crlf_space[chr % 256],strtemp,strlen(strtemp));
        }
      }
    }
    fclose(fp);
    nfree(strtemp);
  } else
  {
    asprintf(&z,"Couldn't open character table file: %s\n",fn);
    print_log(MSG_WARNING,z);
    nfree(z);
  }
  nfree(buf);
}

unsigned char *translate(unsigned char *str, int32_t str_len,char **table)
{
  int32_t a,l;
  unsigned char *tmpstr;

  tmpstr = nmalloc(sizeof(char));
  tmpstr[0] = 0;
  for (a=0;a<str_len;a++)
  {
    l = strlen((char *)tmpstr);
    tmpstr = nrealloc(tmpstr,strlen((char *)tmpstr)+strlen(table[str[a]])+1);
    strcat((char *)tmpstr,table[str[a]]);
    tmpstr[l+strlen(table[str[a]])] = 0;
  }
  return tmpstr;
}

int32_t func(int32_t fb, int32_t len, int32_t *bitlist)
{
  int32_t i;
  int32_t found = 0;
  for (i=0;i<len;i++)
  {
    if (bitlist[i] == fb)
    {
      found = 1;
    }
  }
  return found;
}

int32_t bch(uint32_t cwo)
{ /* This function was taken from the POCSAG encoder
     written by Stefan Petersen in 1995. */
  int32_t lcw,parity,bit;
  lcw = cwo;
  parity = 0;
  for (bit=0;bit<21;bit++)
  {
    if ((cwo & 0x80000000) != 0)
    {
      cwo = cwo ^ 0xED200000;
    }
    cwo = cwo << 1;
  }
  lcw = lcw | (cwo >> 21);
  cwo = lcw;
  for(bit=0; bit<32; bit++)
  {
    if ((cwo & 0x80000000) != 0)
    {
      parity++;
    }
    cwo = cwo << 1;
  }
  return (parity % 2) ? (lcw):(lcw+1);
}


int32_t parity(uint32_t cw)
{ /* Adapted from Stefan Petersens POCSAG encoder */
  uint32_t lcw,bit,par = 0;

  lcw = cw;
  for(bit=0; bit<32; bit++)
  {
    if ((cw & 0x80000000) != 0)
    {
      par++;
    }
    cw = cw << 1;
  }
  return (par % 2) ? (lcw+1):(lcw);
}

int32_t parity_ok(uint32_t cw)
{
  uint32_t lcw,bit,par = 0;
  int32_t pb;
  pb = (cw & 1);
  lcw = cw;
  for(bit=0; bit<31; bit++)
  {
    if ((cw & 0x80000000) != 0)
    {
      par++;
    }
    cw = cw << 1;
  }
  return (pb == (par & 1));
}

int32_t checksum_ecc_ok(uint32_t *cw, int32_t count)
{
  uint32_t p,cwtmp,res;
  int32_t ok = 0;

  for (p=0;p<count;p++)
  {
    if ((bch(cw[p] & 0xFFFFF800) & 0xFFFFFFFE) == (cw[p] & 0xFFFFFFFE))
    {
      ok++;
    } else
    {
      cwtmp = cw[p];
      ecc(&cwtmp,&res);
      if ((res > 0) && (res <= 2))
      {
        ok++;
      }
    }
  }
  return ok;
}

void log_raw_data(char *fn, int32_t bitrate, int32_t source, int32_t channel, uint32_t *cw, int32_t cnt, uint32_t u_time)
{
  uint32_t infoblock,*p;
  uint32_t a,b,unixtime = 0;
  struct pocraw *output;
  char *filename,*z;
  struct tm *now;
  FILE *file;
  time_t tss;

  p = (uint32_t *)&u_time;
  filename = nmalloc(51);
  memset(filename,0,51);

  pthread_mutex_lock(&time_mutex);
  tss = (time_t)(*p & 0xFFFFFFFF);
  now = localtime((time_t *)&tss);
  strftime(filename,50,fn,now);
  pthread_mutex_unlock(&time_mutex);
  if ((file = fopen(filename,"r")) == NULL)
  {
    if ((file = fopen(filename,"w+")) == NULL)
    {
      asprintf(&z,"Couldn't open %s for writing.\n",filename);
      print_log(MSG_ERROR,z);
      nfree(z);
    } else
    {
      fwrite((char *)&raw_header, sizeof(raw_header)-1, 1, file);
      fclose(file);
    }
  } else
  {
    fclose(file);
  }

  if ((file = fopen(filename,"ab+")) != NULL)
  {
    infoblock = (uint32_t)((bitrate) | (source << 16) | (channel << 20));
    unixtime = u_time;
    for (a=0;a<(cnt/16);a++)
    {
      output = nmalloc(sizeof(struct pocraw));
      output->infoblock = infoblock;
      output->unixtime  = unixtime;
      memcpy(&output->cw[0],&cw[a*16],sizeof(uint32_t)*16);
      output->checksum     = output->infoblock;
      output->checksum    += output->unixtime;

      for (b=0;b<16;b++)
      {
        output->checksum += output->cw[b];
      }

      for (b=0;b<(sizeof(struct pocraw))/sizeof(uint32_t);b++)
      {
        htonl_ptr(((uint32_t *)output + b));
      }
      fwrite(output,sizeof(struct pocraw),1,file);
      nfree(output);
    }
    fclose(file);
  }
  else
  {
    asprintf(&z,"Couldn't open %s for writing.\n",filename);
    print_log(MSG_ERROR,z);
    nfree(z);
  }
  nfree(filename);
}

void endian_flip(uint32_t *cw)
{
  uint32_t tmpz;
  tmpz = *cw >> 24;
  tmpz = tmpz | ((*cw & 0x00FF0000) >> 8);
  tmpz = tmpz | ((*cw & 0x0000FF00) << 8);
  tmpz = tmpz | ((*cw & 0x000000FF) << 24);
  *cw = tmpz;
}

int32_t round_up_nearest_multiple(int32_t base, int32_t multiplier)
{
  return ((base % multiplier) == 0) ? base:((base/multiplier)*multiplier + multiplier);
}


unsigned char *error_highlight_new(unsigned char *mystr, uint32_t *errored_words, int32_t words, int32_t fbits,
                                   unsigned char **errstr)
{

  /* Because this piece of code may appear to be a bit cryptic,
     I will attempt to provide a verbose explanation.

     It takes a decoded message, and builds a table with the ECC
     status of each codeword.

     The possible values for the ECC status are:
     0: No error
     1: One or two correctable errors
     2: Uncorrectable error.

     The table (t) it builds has as many bytes as the POCSAG codewords has message bits.

     If our message has 4 codewords, with the ECC status "1210", the table will look like this:

     11111111111111111111222222222222222222221111111111111111111100000000000000000000

     This is equal to the ECC status of 80 message bits, which is equal to
     20 numeric character or 11 7-bit ascii characters.

     The next step in the process depends on whether the message is alphanumeric
     or numeric.

     If its alphanumeric, the following thing happens:

     It will build the errpos table by using the boundary of each character in
     the ECC status table to build an index to access the char_codes table:

     *     **     **     **     **     **     **     **     **     **     **     *
     AAAAAAABBBBBBBCCCCCCCDDDDDDDEEEEEEEFFFFFFFGGGGGGGHHHHHHHIIIIIIIJJJJJJJKKKKKKK
     11111111111111111111222222222222222222221111111111111111111100000000000000000000

     By using the beginning and end of each character, and looking at the corresponding
     ECC status, we get the following indexes: 11, 11, 12, 22, 22, 21, 11, 11, 10, 00, 00

     Accessing the char_codes table using these indexes, errpos now contains: 22344322100

     It then iterates over the input string, one character at a time, and uses the errpos
     table to determine where to insert the error highlighting strings.

     Input string: ABCDEFGHIJK

     Iteration 1: Character = 2, previous character = 0

                  We are transitioning from an error-free character to a character
                  with correctable error(s). We need to insert the prefix to indicate
                  this.

                  At string_lookup[2][0] we find the value 5.
                  We divide it by 10 to get the index of the string_ptr_table.
                  We get 0. Since index 0 in that table refers to a zero length
                  string, nothing gets appended to the output string.

                  Next, we take the value found at string_lookup[2][5] mod 10,
                  which is 5. 5 in string_ptr_table refer to the prefix string
                  for a fully corrected character. This gets appended to the output.

                  Output: <FCP>A

      Iteration 2: Character = 2, previous character = 2

                  Output = string_ptr_table[string_lookup[2][2]/10) +
                           string_ptr_table[string_lookup[2][2]%10)

                  Output: <FCP>AB

      Iteration 2: Character = 3, previous character = 2

                  Output = string_ptr_table[string_lookup[3][2]/10) +
                           string_ptr_table[string_lookup[3][2]%10)

                  Output = <FCP>AB<FCS><PUP>C

      Iteration 3: Character = 4, previous character = 3

                  Output = string_ptr_table[string_lookup[4][3]/10) +
                           string_ptr_table[string_lookup[4][3]%10)

                  Output = <FCP>AB<FCS><PUP>C<PUS><UCP>D

                  etc...
*/

  unsigned char *t,*errpos,*newstr,*tmpstr;
  int32_t i,j,bitcnt,last_err,errpos_len = 0;
  int32_t len = 1;
  unsigned char zero = 0;
  errpos = NULL;

  unsigned char *string_ptr_table[]  = {&zero,
                                        errstr[e_partially_corrected_prefix],
                                        errstr[e_partially_corrected_suffix],
                                        errstr[e_partially_uncorrected_prefix],
                                        errstr[e_partially_uncorrected_suffix],
                                        errstr[e_fully_corrected_prefix],
                                        errstr[e_fully_corrected_suffix],
                                        errstr[e_uncorrectable_prefix],
                                        errstr[e_uncorrectable_suffix]};

  const unsigned char char_codes[]         = {0,1,3,0,0,0,0,0,0,0,1,2,3,0,0,0,0,0,0,0,3,3,4};


  const unsigned char string_lookup[5][5]  = {{ 0, 2, 6, 4, 8},
                                              { 1, 0,61,41,81},
                                              { 5,25, 0,45,85},
                                              { 3,23,63, 0,83},
                                              { 7,27,67,47, 0}};

  const unsigned char string_lookup_end[5] = {0,2,6,4,8};

  errpos = NULL;
  if (func(fbits,config->log_config[main_config_index]->num_list[alphanumeric_bits]->count,config->log_config[main_config_index]->num_list[alphanumeric_bits]->list))
  {
    bitcnt = (round_up_nearest_multiple(20*words,7));
    t = nmalloc(bitcnt*sizeof(uint32_t) + 10);
    memset(t,0,bitcnt);
    for (i=0;i<words;i++)
    {
      for (j=0;j<20;j++)
      {
        t[i*20 + j] = errored_words[i];
      }
    }
    errpos_len = (bitcnt/7 + 1);
    errpos = nmalloc(errpos_len + 1);

    for (i=0;i<((bitcnt/7));i++)
    {
      errpos[i] = char_codes[t[i*7]*10 + t[i*7 + 6]];
    }
    errpos[bitcnt/7 + 1] = 0;
    nfree(t);

  } else if (func(fbits,config->log_config[main_config_index]->num_list[numeric_bits]->count,config->log_config[main_config_index]->num_list[numeric_bits]->list))
  {
    errpos_len = (words*5 + 1);
    errpos = nmalloc(errpos_len + 1);
    for (i=0;i<(words*5);i++)
    {
      errpos[i] = char_codes[errored_words[i/5]*10 + errored_words[i/5]];
    }
    errpos[words*5 + 1] = 0;
  }

  last_err = 0;
  newstr = nmalloc(1*sizeof(char));
  newstr[0] = 0;

  if (errpos_len >= strlen((char *)mystr))
  {
    for (i=0;i<strlen((char *)mystr);i++)
    {
      for (j=0;j<3;j++)
      {
         tmpstr = string_ptr_table[j?(string_lookup[errpos[i]][last_err]%10):(string_lookup[errpos[i]][last_err]/10)];
         len = (j < 2 ? len+strlen((char *)tmpstr) : len + 1);
         newstr = nrealloc(newstr, len + 1);
         newstr[len] = 0;
         strncat((char *)newstr, (char *)(j < 2 ? tmpstr : &mystr[i]),((j == 2) ? 1 : strlen((char *)tmpstr)));
      }
      last_err = errpos[i];
    }
    tmpstr = string_ptr_table[string_lookup_end[last_err]];
    len += strlen((char *)tmpstr);
    newstr = nrealloc(newstr,len);
    strcat((char *)newstr,(char *)tmpstr);
  } else
  {
    newstr = nrealloc((char *)newstr,strlen((char *)mystr)+1);
    memcpy(newstr,mystr,strlen((char *)mystr)+1);
  }
  if (errpos)
  {
    nfree(errpos);
    errpos = NULL;
  }
  return newstr;
}

unsigned char *string_replace(unsigned char *mystr, unsigned char *from_str, unsigned char *to_str)
{
  unsigned char *str,*out;
  int32_t new_len,cnt = 0;
  if (strlen((char *)from_str)>0)
  {
    str = mystr;
    while (str != NULL)
    {
      str = (unsigned char *)strstr((char *)str,(char *)from_str);
      if (str != NULL)
      {
        str++;
        cnt++;
      }
    }
    str = mystr;
    new_len = strlen((char *)mystr)-(cnt*strlen((char *)from_str))+(cnt*strlen((char *)to_str));
    out = nmalloc(sizeof(char)*(new_len + 1));
    memset(out,0,new_len + 1);
    while (strlen((char *)out) < new_len)
    {
      if (strstr((char *)str,(char *)from_str) != NULL)
      {
        strncat((char *)out,(char *)str,(strstr((char *)str,(char *)from_str) - (char *)str));
        strcat ((char *)out,(char *)to_str);
        str = (unsigned char *)strstr((char *)str,(char *)from_str) + strlen((char *)from_str);
      } else
      {
        strcat((char *)out,(char *)str);
      }
    }
  } else
  {
    out = nmalloc(sizeof(char)*strlen((char *)mystr)+1);
    memcpy(out,mystr,strlen((char *)mystr)+1);
  }
  return out;
}


void n_enqueue(struct pointer_queue *queue, void *to_enqueue, int32_t line)
{
  if (queue->count == 0)
  {
    queue->ptr = nmalloc((++queue->count)*sizeof(void *));
  } else
  {
    queue->ptr = nrealloc(queue->ptr,(++queue->count)*sizeof(void *));
  }
  queue->ptr[queue->count - 1] = to_enqueue;
}

void *n_dequeue(struct pointer_queue *queue, int32_t *i, void **tmp, int32_t line)
{
  if (queue->count == 1)
  {
    *tmp = queue->ptr[0];
    nfree(queue->ptr);
    queue->count--;
    queue->ptr = NULL;
    return *tmp;
  } else if (queue->count > 1)
  {
    *tmp = queue->ptr[0];
    for (*i=0;((*i)<queue->count - 1);(*i)++)
    {
      queue->ptr[*i] = queue->ptr[(*i)+1];
    }
    queue->ptr = nrealloc(queue->ptr,(--queue->count)*sizeof(void*));
    return *tmp;
  } else
  {
    return NULL;
  }
}

void message_log(void)
{
  int32_t   i,j,n,y,z,mi;
  int32_t tmp_i;
  static void **tmp_v;
  struct pointer_queue *log_queue_tmp;
  static struct message_block *msg;
  log_queue_tmp = nmalloc(sizeof(struct pointer_queue));
  log_queue_tmp->count = 0;
  tmp_v = nmalloc(sizeof(void *));
  struct timespec ts;
  struct timeval  tp;
  int32_t rc;

  while (1)
  {
    pthread_mutex_lock(&log_mutex);
    pthread_mutex_lock(&exit_mutex);
    if ((exit_now == 1) && (log_queue->count == 0))
    {
      pthread_cond_signal(&exit_cond);
    }
    pthread_mutex_unlock(&exit_mutex);

    if (use_exit_flag)
    {
      gettimeofday(&tp, NULL);
      ts.tv_sec  = tp.tv_sec;
      ts.tv_nsec = tp.tv_usec * 1000;
      ts.tv_nsec += 100000000; /* make pthread_cond_timedwait timeout 100ms into the future */
      rc = pthread_cond_timedwait(&log_cond,&log_mutex,&ts);
    } else
    {
      pthread_cond_wait(&log_cond,&log_mutex);
    }

    while (log_queue->count > 0)
    {
      nenqueue(log_queue_tmp,ndequeue(log_queue,&tmp_i,tmp_v));
    }
    pthread_mutex_unlock(&log_mutex);

    while (log_queue_tmp->count > 0)
    {
      msg = ndequeue(log_queue_tmp,&tmp_i,tmp_v);
      for (i=0;i<config->count;i++)
      {
        if ( ((config->function[i] != NULL) && ((config->log_config[i]->str_num[enabled] == yes) && (config->log_config[i]->str_num_is_set[enabled]))) &&
             (!((config->log_config[i]->str_num[discard_errored_messages] == yes) && (config->log_config[i]->str_num_is_set[discard_errored_messages]) && (msg->errored)))
           )
        {
          for (j=0;j<config->log_config[i]->rec.count;j++)
          {
            n = get_int_index(config->log_config[i]->ntab[message_policy].num_lookup,config->log_config[i]->ntab[message_policy].num_count,config->log_config[i]->rec.rcpt[j]->num[_id],&y,&z);
            n = (n >= 0 ? config->log_config[i]->ntab[message_policy].num[n] :  allow);
            if  (check_message   (msg,config->log_config[i]->rec.rcpt[j],config->log_config[i]) && (msg->ric > 10))
            {

               mi = get_log_function_index(config,"main");
               if (mi >= 0)
               {
                 (config->function[i])(msg,config->log_config[i]->rec.rcpt[j],config->log_config[i],config->log_config[mi]->ric_strings);
               }
            }
          }
        }
      }

      if (msg->codewords)
      {
        nfree(msg->codewords);
        msg->codewords = NULL;
      }
      if (msg->codewords_uncorrected)
      {
        nfree(msg->codewords_uncorrected);
        msg->codewords_uncorrected = NULL;
      }

      if (msg->idle_prefixed_codewords);
      {
        nfree(msg->idle_prefixed_codewords);
        msg->idle_prefixed_codewords = NULL;
      }

      if (msg->idle_prefixed_codewords_uncorrected);
      {
        nfree(msg->idle_prefixed_codewords_uncorrected);
        msg->idle_prefixed_codewords_uncorrected = NULL;
      }

      if (msg->codewords_ecc_status)
      {
        nfree(msg->codewords_ecc_status);
        msg->codewords_ecc_status = NULL;
      }

      if (msg->codewords_ecc_flags)
      {
        nfree(msg->codewords_ecc_flags);
        msg->codewords_ecc_flags = NULL;
      }

      if (msg->codewords_ecc_flags_str)
      {
        nfree(msg->codewords_ecc_flags_str);
        msg->codewords_ecc_flags_str = NULL;
      }

      if (msg->msg_fixed)
      {
        nfree(msg->msg_fixed);
        msg->msg_fixed = NULL;
      }

      if (msg->msg)
      {
        nfree(msg->msg);
        msg->msg = NULL;
      }

      if (msg->msg_fixed_no_crlf)
      {
        nfree(msg->msg_fixed_no_crlf);
        msg->msg_fixed_no_crlf = NULL;
      }

      if (msg->msg_uncorrected)
      {
        nfree(msg->msg_uncorrected);
        msg->msg_uncorrected = NULL;
      }

      if (msg->date)
      {
        nfree(msg->date);
        msg->date = NULL;
      }
      nfree(msg);
    }
  }
}

void register_log_function(struct configuration_new *lf, void *address, char *str)
{
  char *tmp;

  lf->count++;

  if (lf->function == NULL)
  {
    lf->function       = nmalloc(sizeof(void *));
  }

  if (lf->function_names == NULL)
  {
    lf->function_names = nmalloc(sizeof(char *));
  }

  lf->function = nrealloc(lf->function,lf->count*sizeof(void *));
  lf->function_names = nrealloc(lf->function_names,lf->count*sizeof(char *));
  lf->function_names[lf->count - 1] = nmalloc(sizeof(char)*strlen(str)+1);
  tmp = nmalloc(sizeof(char)*strlen(str)+1);
  memcpy(tmp,str,strlen(str));
  tmp[strlen(str)] = 0;
  str_tolower(tmp);
  memcpy(lf->function_names[lf->count - 1],tmp,strlen(tmp)+1);
  nfree(tmp);
  lf->function_names[lf->count - 1][strlen(str)] = 0;
  lf->function[lf->count - 1] = address;
}

int32_t get_log_function_index(struct configuration_new *lf, char *str)
{
  char *tmp;
  int32_t i,j;
  j = -1;
  tmp = nmalloc(sizeof(char)*strlen(str)+1);
  memcpy(tmp,str,strlen(str));
  tmp[strlen(str)] = 0;
  str_tolower(tmp);

  for (i=0;i<lf->count;i++)
  {
    if (strcmp(tmp,lf->function_names[i]) == 0)
    {
      j = i;
    }
  }

  nfree(tmp);
  return j;
}

void str_tolower(char *str)
{
  int32_t i;
  for (i=0;i<strlen(str);i++)
  {
    str[i] = (str[i] >= 'A' && str[i] <= 'Z') ? (str[i] + 0x20) : str[i];
  }
}

void config_parse(struct configuration_new *config,int32_t argc,char **argv)
{
  int32_t id,i,j,k,m,l,y,z,log_index,cmd_line,cont,index,arg_c,use_config_file;
  char *buf,*buf2,*rc,*tmp,*tmp2,*tmp3,*tmp4,*fn,*str_ptr,*zz,**arg_v;
  FILE *fp;
  fp = NULL;
  arg_c = 0;
  use_config_file = 0;
  arg_v = NULL;

  buf = nmalloc(1000);
  memset(buf,0,1000);
  buf2 = nmalloc(1000);
  memset(buf2,0,1000);
  tmp = nmalloc(1000);
  memset(tmp,0,1000);
  tmp2 = nmalloc(1000);
  memset(tmp2,0,1000);
  tmp3 = nmalloc(1000);
  memset(tmp3,0,1000);
  tmp4 = nmalloc(1000);
  memset(tmp4,0,1000);


  str_ptr = NULL;
  fn = NULL;
  m = 0;
  cmd_line = 0;
  log_index = -1;

  for (i=1;i<argc;i++)
  {
    if (sscanf(argv[i],"-f=%s",buf) == 1)
    {
      asprintf(&fn,"%s",buf);
      conf_ok = 1;
      use_config_file = 1;
    } else if (strcmp(argv[i],"-f") == 0)
    {
      conf_ok = 1;
      use_config_file = 1;
    } else if (strcmp(argv[i],"-c") == 0)
    {
      conf_ok = 1;
      cmd_line = 1;
    } else if (strcmp(argv[i],"-h") == 0)
    {
      printf("%s\n",program_string);
      printf("The following command line parameters are available:\n\n");
      printf("  -h                   Show this help\n");
      printf("  -f=config            Load configuration from file \"config\"\n");
      printf("  -f                   Load configuration from \"%s\"\n",default_config_file_fn);
      printf("  -c                   Load configuration from command line\n\n");
      printf("In combination with -c, the following options can be used\n");
      printf("  --dsp-sample-rate=NUM                Desired sample rate     Default: 48000\n");
      printf("  --dsp-sample-bits=NUM                Desired bit resolution  Default: 16   \n");
      printf("  --dsp-input-mode=(stereo|left|right) Desired channel(s)      Default: stereo\n");
      printf("  --input-mode=MODE                    Input mode              Default: dsp-pcm\n");
      printf("    Possible values: dsp-pcm, tcp-pcm, tcp-raw, file-raw\n\n");
      printf("  --input-source=SOURCE                Input source            Default: /dev/dsp\n");
      printf("  --input-password=PASSWORD\n");
      printf("  --print-codewords=(yes|no)           Print POCSAG codewords  Default: no\n");
      printf("  --print-codewords-eh=(yes|no)        Error highlighting      Default: yes\n");
      printf("  --raw-logging=(yes|no)               Log POCSAG codewords    Default: no\n");
      printf("  --noise-smoothing=(yes|no)           Noise smoothing         Default: yes\n");
      printf("  --noise-smoothing-threshold=NUM                              Default: 60\n");
      printf("  --stats=(yes|no)                     Statistics              Default: yes\n");
      printf("  --stats-file=FILE                    Statistics file         Default: stats.dat\n");
      printf("  --charset=FILE                       Character set           Default: charset.tbl\n");
      printf("  --bitrates=LIST                      Bitrates to decode      Default: 512,1200,2400\n");
      printf("  --alphanumeric-bits=LIST             Alpha bits              Default: 3\n");
      printf("  --numeric-bits=LIST                  Numeric bits            Default: 0,1,2\n");
      printf("  --pcm-server=(yes|no)                PCM stream server       Default: no\n");
      printf("  --pcm-server-port=NUM                PCM stream server port  Default: 5000\n");
      printf("  --pcm-server-password=PASSWORD       Optional password\n");
      printf("  --raw-server=(yes|no)                RAW stream server       Default: no\n");
      printf("  --raw-server-port=NUM                RAW stream server port  Default: 5001\n");
      printf("  --raw-server-password=PASSWORD       Optional password\n");
      printf("  --ric-lookup=STRING                  RIC string lookup\n");
      printf("\n");
      printf("Output function options:\n\n");
      #ifdef USE_MYSQL
      printf("  --mysql,db=STRING\n");
      printf("  --mysql,enabled=(yes|no)                                     Default: no\n");
      printf("  --mysql,message-policy=(allow|drop)                          Default: allow\n");
      printf("  --mysql,ric-whitelist=LIST\n");
      printf("  --mysql,ric-blacklist=LIST\n");
      printf("  --mysql,function-bits-blacklist=LIST\n");
      printf("  --mysql,keyword-whitelist=LIST\n");
      printf("  --mysql,keywork-blacklist=LIST\n");
      printf("  --mysql,discard-errored-messages=(yes|no)                    Default: no\n");
      printf("\n");
      #endif
      printf("  --console,enabled=(yes|no)                                   Default: yes\n");
      printf("  --console,message-policy=(allow|drop)                        Default: allow\n");
      printf("  --console,eh-enable=(yes|no)                                 Default: yes\n");
      printf("  --console,ric-whitelist=LIST\n");
      printf("  --console,ric-blacklist=LIST\n");
      printf("  --console,function-bits-blacklist=LIST\n");
      printf("  --console,keyword-whitelist=LIST\n");
      printf("  --console,keywork-blacklist=LIST\n");
      printf("  --console,discard-errored-messages=(yes|no)                  Default: no\n");
      printf("\n");
      printf("  --winmsg,host=STRING\n");
      printf("  --winmsg,enabled=(yes|no)                                    Default: no\n");
      printf("  --winmsg,message-policy=(allow|drop)                         Default: allow\n");
      printf("  --winmsg,ric-whitelist=LIST\n");
      printf("  --winmsg,ric-blacklist=LIST\n");
      printf("  --winmsg,function-bits-blacklist=LIST\n");
      printf("  --winmsg,keyword-whitelist=LIST\n");
      printf("  --winmsg,keywork-blacklist=LIST\n");
      printf("  --winmsg,discard-errored-messages=(yes|no)                   Default: no\n");
      printf("\n");
      printf("  --irc,host=STRING\n");
      printf("  --irc,enabled=(yes|no)                                       Default: no\n");
      printf("  --irc,eh-enable=(yes|no)                                     Default: yes\n");
      printf("  --irc,message-policy=(allow|drop)                            Default: allow\n");
      printf("  --irc,ric-whitelist=LIST\n");
      printf("  --irc,ric-blacklist=LIST\n");
      printf("  --irc,function-bits-blacklist=LIST\n");
      printf("  --irc,keyword-whitelist=LIST\n");
      printf("  --irc,keywork-blacklist=LIST\n");
      printf("  --irc,discard-errored-messages=(yes|no)                      Default: no\n");
      printf("\n");
      printf("  --file,logfile=STRING\n");
      printf("  --file,enabled=(yes|no)                                      Default: yes\n");
      printf("  --file,message-policy=(allow|drop)                           Default: allow\n");
      printf("  --file,ric-whitelist=LIST\n");
      printf("  --file,ric-blacklist=LIST\n");
      printf("  --file,function-bits-blacklist=LIST\n");
      printf("  --file,keyword-whitelist=LIST\n");
      printf("  --file,keywork-blacklist=LIST\n");
      printf("  --file,discard-errored-messages=(yes|no)                     Default: no\n");
      printf("\n");
      printf("  --mail,to=STRING\n");
      printf("  --mail,from=STRING\n");
      printf("  --mail,mta=STRING                                            Default: /usr/sbin/sendmail\n");
      printf("  --mail,enabled=(yes|no)                                      Default: no\n");
      printf("  --mail,message-policy=(allow|drop)                           Default: allow\n");
      printf("  --mail,ric-whitelist=LIST\n");
      printf("  --mail,ric-blacklist=LIST\n");
      printf("  --mail,function-bits-blacklist=LIST\n");
      printf("  --mail,keyword-whitelist=LIST\n");
      printf("  --mail,keywork-blacklist=LIST\n");
      printf("  --mail,discard-errored-messages=(yes|no)                     Default: no\n");
      printf("\n");
      printf("  --telnet,port=STRING\n");
      printf("  --telnet,enabled=(yes|no)                                    Default: no\n");
      printf("  --telnet,message-policy=(allow|drop)                         Default: allow\n");
      printf("  --telnet,ric-whitelist=LIST\n");
      printf("  --telnet,ric-blacklist=LIST\n");
      printf("  --telnet,function-bits-blacklist=LIST\n");
      printf("  --telnet,keyword-whitelist=LIST\n");
      printf("  --telnet,keywork-blacklist=LIST\n");
      printf("  --telnet,discard-errored-messages=(yes|no)                   Default: no\n");
      printf("\n");
      printf("  --external,external=STRING\n");
      printf("  --external,enabled=(yes|no)                                  Default: no\n");
      printf("  --external,message-policy=(allow|drop)                       Default: allow\n");
      printf("  --external,ric-whitelist=LIST\n");
      printf("  --external,ric-blacklist=LIST\n");
      printf("  --external,function-bits-blacklist=LIST\n");
      printf("  --external,keyword-whitelist=LIST\n");
      printf("  --external,keywork-blacklist=LIST\n");
      printf("  --external,discard-errored-messages=(yes|no)                 Default: no\n");
      printf("\n");
      printf("Please see the documentation for more information about these options.\n");
      exit(0);
    }
  }

  fn = (fn == NULL) ? default_config_file_fn : fn;

  index = 0;
  cont = 1;
  while (cont)
  {
    if (cmd_line)
    {
      index++;
      if (index>=argc)
      {
        cont = 0;
        goto qq;
      }

      char_replace(argv[index],' ',255);
      if (sscanf(argv[index],"--%[^,],%s",buf,tmp3) == 2)
      {
        char_replace(tmp3,255,' ');
        str_tolower(buf);
        log_index = get_log_function_index(config,buf);
        if (log_index == -1)
        {
          asprintf(&zz,"No log handler installed for \"%s\" (%s).\n",buf,argv[index]);
          print_log(MSG_ERROR,zz);
          nfree(zz);
          exit(0);
        } else
        {
          str_ptr = tmp3;
          memcpy(tmp4,buf,strlen(buf));
        }
      } else if (sscanf(argv[index],"--%s",tmp3) == 1)
      {
        char_replace(tmp3,255,' ');
        log_index = get_log_function_index(config,"main");
        if (log_index == -1)
        {
          print_log(MSG_ERROR,"No log handler installed for \"main\". This should never happen. Exiting...\n");
          exit(0);
        } else
        {
          memcpy(tmp4,buf,strlen(buf)+1);
          str_ptr = tmp3;
        }
      }
    } else if (use_config_file)
    {
      if (fp == NULL)
      {
        fp = fopen(fn,"r");
      }
      if (fp)
      {
        if (!(rc = fgets(buf2,990,fp)) == 0)
        {
          if ((buf2[strlen(buf2)-1] == 0x0A) || (buf2[strlen(buf2)-1] == 0x0D))
          {
            buf2[strlen(buf2)-1] = 0;
          }

          if (buf2[0] == '[')
          {
            sscanf(buf2,"[%[a-zA-Z]]",tmp);
            str_tolower(tmp);
            log_index = get_log_function_index(config,tmp);
            if (log_index == -1)
            {
              asprintf(&zz,"No log handler installed for \"%s\"\n",tmp);
              print_log(MSG_ERROR,zz);
              nfree(zz);
              exit(0);
            }
          } else
          {
            str_ptr = buf2;
            memcpy(tmp4,tmp,strlen(tmp)+1);
          }
        } else
        {
          cont = 0;
          goto qq;
        }
      } else
      {
        asprintf(&zz,"Couldn't open config file %s\n",fn);
        print_log(MSG_ERROR,zz);
        nfree(zz);
        exit(0);
      }
    } else
    {
      cont = 0;
      goto qq;
    }

    if (str_ptr)
    {
      /* Configuration input category 1 */
      for (k=0;k<config_wb_enum_count;k++)
      {
        if (sscanf(str_ptr,config_wb_fmt_strings_with_id[k],&id,tmp2) == 2)
        {
          i = get_int_index(config->log_config[log_index]->wb_wrap[k].id_lookup,config->log_config[log_index]->wb_wrap[k].id_count,id,&y,&z);
          if (i == -1)
          {
            config->log_config[log_index]->wb_wrap[k].id_count++;
            if (config->log_config[log_index]->wb_wrap[k].id_lookup == NULL)
            {
              config->log_config[log_index]->wb_wrap[k].id_lookup = nmalloc(sizeof(int32_t));
            } else
            {
              config->log_config[log_index]->wb_wrap[k].id_lookup = nrealloc(config->log_config[log_index]->wb_wrap[k].id_lookup,sizeof(int32_t)*config->log_config[log_index]->wb_wrap[k].id_count);
            }

            if (config->log_config[log_index]->wb_wrap[k].wb == NULL)
            {
              config->log_config[log_index]->wb_wrap[k].wb = nmalloc(sizeof(struct whitelist_blacklist *));
              memset(config->log_config[log_index]->wb_wrap[k].wb,0,sizeof(struct whitelist_blacklist *));
            } else
            {
              config->log_config[log_index]->wb_wrap[k].wb = nrealloc(config->log_config[log_index]->wb_wrap[k].wb,sizeof(struct whitelist_blacklist *)*config->log_config[log_index]->wb_wrap[k].id_count);
              config->log_config[log_index]->wb_wrap[k].wb[config->log_config[log_index]->wb_wrap[k].id_count - 1] = NULL;
            }
            config->log_config[log_index]->wb_wrap[k].id_lookup[config->log_config[log_index]->wb_wrap[k].id_count - 1] = id;
          }
          config->log_config[log_index]->wb_wrap[k].wb[config->log_config[log_index]->wb_wrap[k].id_count - 1] = parse_list(tmp2,config->log_config[log_index]->wb_wrap[k].wb[config->log_config[log_index]->wb_wrap[k].id_count - 1],k,id);

        } else if (sscanf(str_ptr,config_wb_fmt_strings_no_id[k],tmp2) == 1)
        {
          i = get_int_index(config->log_config[log_index]->wb_wrap[k].id_lookup,config->log_config[log_index]->wb_wrap[k].id_count,-1,&y,&z);
          if (i == -1)
          {
            config->log_config[log_index]->wb_wrap[k].id_count++;
            if (config->log_config[log_index]->wb_wrap[k].id_lookup == NULL)
            {
              config->log_config[log_index]->wb_wrap[k].id_lookup = nmalloc(sizeof(int32_t));
            } else
            {
              config->log_config[log_index]->wb_wrap[k].id_lookup = nrealloc(config->log_config[log_index]->wb_wrap[k].id_lookup,sizeof(int32_t)*config->log_config[log_index]->wb_wrap[k].id_count);
            }

            if (config->log_config[log_index]->wb_wrap[k].wb == NULL)
            {
              config->log_config[log_index]->wb_wrap[k].wb = nmalloc(sizeof(struct whitelist_blacklist *));
              memset(config->log_config[log_index]->wb_wrap[k].wb,0,sizeof(struct whitelist_blacklist *));
            } else
            {
              config->log_config[log_index]->wb_wrap[k].wb = nrealloc(config->log_config[log_index]->wb_wrap[k].wb,sizeof(struct whitelist_blacklist *)*config->log_config[log_index]->wb_wrap[k].id_count);
              config->log_config[log_index]->wb_wrap[k].wb[config->log_config[log_index]->wb_wrap[k].id_count - 1] = NULL;
            }
            config->log_config[log_index]->wb_wrap[k].id_lookup[config->log_config[log_index]->wb_wrap[k].id_count - 1] = -1;
          }
          config->log_config[log_index]->wb_wrap[k].wb[config->log_config[log_index]->wb_wrap[k].id_count - 1] = parse_list(tmp2,config->log_config[log_index]->wb_wrap[k].wb[config->log_config[log_index]->wb_wrap[k].id_count - 1],k,-1);
        }
      }

      /* Configuration input category 2 */
      for (k=0;k<config_int_enum_count;k++)
      {

        if (sscanf(str_ptr,config_int_fmt_strings[k],&l) == 1)
        {
          config->log_config[log_index]->num[k] = l;
          config->log_config[log_index]->num_is_set[k] = 1;
        }
      }
      /* Configuration input category 3 */
      for (k=0;k<config_int_array_enum_count;k++)
      {
        if (sscanf(str_ptr,config_int_array_fmt_strings[k],tmp2) == 1)
        {
          config->log_config[log_index]->num_list[k] = parse_int_csv(tmp2,config->log_config[log_index]->num_list[k],1);
        }
      }

      /* Configuration input category 4 */
      for (k=0;k<config_str_enum_count;k++)
      {
        char_replace(str_ptr,' ',255);
        if (sscanf(str_ptr,config_str_fmt_strings[k],tmp2) == 1)
        {
          char_replace(tmp2,255,' ');
          if (config_str_lookup[k])
          {
            j = str_lookup(tmp2,(char **)config_str_keyword_strings,config_str_keyword_enum_count);
            if (j >= 0)
            {
              config->log_config[log_index]->str_num[k] = j;
              config->log_config[log_index]->str_num_is_set[k] = 1;
            } else
            {
              asprintf(&zz,"Invalid option: %s (%s)\n",tmp2,str_ptr);
              print_log(MSG_ERROR,zz);
              nfree(zz);
              exit(0);
            }
          } else
          {
            config->log_config[log_index]->str[k] = nmalloc(strlen(tmp2)+1);
            memcpy(config->log_config[log_index]->str[k],tmp2,strlen(tmp2)+1);
          }
        } else
        {
          char_replace(str_ptr,255,' ');
        }
      }
      /* Configuration input category 5 */
      for (k=0;k<config_rcpt_str_enum_count;k++)
      {
        if (sscanf(str_ptr,config_rcpt_str_fmt_strings_with_id[k],&id,tmp2) == 2)
        {
          if (cmd_line)
          {
            config->log_config[log_index]->str_num[enabled]        = yes;
            config->log_config[log_index]->str_num_is_set[enabled] = 1;
          }
          config->log_config[log_index]->rec = parse_rcpt_list_new(&config->log_config[log_index]->rec, tmp2, id, k);
        } else if (sscanf(str_ptr,config_rcpt_str_fmt_strings_no_id[k],tmp2) == 1)
        {
          if (cmd_line)
          {
            config->log_config[log_index]->str_num[enabled]        = yes;
            config->log_config[log_index]->str_num_is_set[enabled] = 1;
          }
          config->log_config[log_index]->rec = parse_rcpt_list_new(&config->log_config[log_index]->rec, tmp2, -1, k);
        }
      }

     /* Configuration input category 6 */
      for (k=0;k<config_mp_cs_enum_count;k++)
      {
        if (sscanf(str_ptr,mp_cs_fmt_strings_with_id[k],&id,tmp2) == 2)
        {
          i = get_int_index(config->log_config[log_index]->ntab[k].num_lookup,config->log_config[log_index]->ntab[k].num_count,id,&y,&z);
          if (i == -1)
          {
            j = str_lookup(tmp2,(char **)&config_str_keyword_strings,config_str_keyword_enum_count);
            if (j >= 0)
            {
              config->log_config[log_index]->ntab[k].num_count++;
              if (config->log_config[log_index]->ntab[k].num_lookup == NULL)
              {
                config->log_config[log_index]->ntab[k].num_lookup = nmalloc(sizeof(int32_t));
              } else
              {
                config->log_config[log_index]->ntab[k].num_lookup = nrealloc(config->log_config[log_index]->ntab[k].num_lookup,sizeof(int32_t)*config->log_config[log_index]->ntab[k].num_count);
              }

              if (config->log_config[log_index]->ntab[k].num == NULL)
              {
                config->log_config[log_index]->ntab[k].num = nmalloc(sizeof(int32_t));
                memset(config->log_config[log_index]->ntab[k].num,0,sizeof(int32_t));
              } else
              {
                config->log_config[log_index]->ntab[k].num = nrealloc(config->log_config[log_index]->ntab[k].num,sizeof(int32_t)*config->log_config[log_index]->ntab[k].num_count);
                config->log_config[log_index]->ntab[k].num[config->log_config[log_index]->ntab[k].num_count - 1] = 0;
              }
              config->log_config[log_index]->ntab[k].num_lookup[config->log_config[log_index]->ntab[k].num_count - 1] = id;
              config->log_config[log_index]->ntab[k].num[config->log_config[log_index]->ntab[k].num_count - 1] = j;
            } else
            {
              asprintf(&zz,"Section %s, invalid option: %s (%s)\n",tmp4,tmp2,str_ptr);
              print_log(MSG_ERROR,zz);
              nfree(zz);
              exit(0);
            }
          }
        } else if (sscanf(str_ptr,mp_cs_fmt_strings_no_id[k],tmp2) == 1)
        {
          i = get_int_index(config->log_config[log_index]->ntab[k].num_lookup,config->log_config[log_index]->ntab[k].num_count,-1,&y,&z);
          if (i == -1)
          {
            j = str_lookup(tmp2,(char **)&config_str_keyword_strings,config_str_keyword_enum_count);
            if (j >= 0)
            {
              config->log_config[log_index]->ntab[k].num_count++;
              if (config->log_config[log_index]->ntab[k].num_lookup == NULL)
              {
                config->log_config[log_index]->ntab[k].num_lookup = nmalloc(sizeof(int32_t));
              } else
              {
                config->log_config[log_index]->ntab[k].num_lookup = nrealloc(config->log_config[log_index]->ntab[k].num_lookup,sizeof(int32_t)*config->log_config[log_index]->ntab[k].num_count);
              }

              if (config->log_config[log_index]->ntab[k].num == NULL)
              {
                config->log_config[log_index]->ntab[k].num = nmalloc(sizeof(int32_t));
                memset(config->log_config[log_index]->ntab[k].num,0,sizeof(int32_t));
              } else
              {
                config->log_config[log_index]->ntab[k].num = nrealloc(config->log_config[log_index]->ntab[k].num,sizeof(int32_t)*config->log_config[log_index]->ntab[k].num_count);
                config->log_config[log_index]->ntab[k].num[config->log_config[log_index]->ntab[k].num_count - 1] = 0;
              }
              config->log_config[log_index]->ntab[k].num_lookup[config->log_config[log_index]->ntab[k].num_count - 1] = -1;
              config->log_config[log_index]->ntab[k].num[config->log_config[log_index]->ntab[k].num_count - 1] = j;
            } else
            {
              asprintf(&zz,"Section %s, invalid option: %s (%s)\n",tmp4,tmp2,str_ptr);
              print_log(MSG_ERROR,zz);
              nfree(zz);
              exit(0);
            }
          }
        }
      }

      /* Configuration input category 7 */
      for (k=0;k<config_error_highlight_enum_count;k++)
      {
        if (sscanf(str_ptr,config_error_highlight_fmt_strings[k],tmp2) == 1)
        {
          config->log_config[log_index]->error_highlight[k] = parse_eh_string(tmp2);
        }
      }

      /* Configuration input category 8 */

      char_replace(str_ptr,' ',255);
      if (sscanf(str_ptr,"ric-lookup=%s",tmp2) == 1)
      {
        config->log_config[log_index]->ric_strings = parse_ric_lookup(config->log_config[log_index]->ric_strings,tmp2);
      }
    }
  }
  qq:;
  if (fp)
  {
    fclose(fp);
    fp = NULL;
  }
  nfree(tmp);
  nfree(tmp2);
  nfree(tmp3);
  nfree(tmp4);
  nfree(buf);
  nfree(buf2);
}

int32_t get_int_index(int32_t *recipients_index, int32_t recipients_cnt, int32_t num, int32_t *i, int32_t *j)
{
  *j = -1;
  for (*i=0;*i<recipients_cnt;i[0]++)
  {
    if (recipients_index[*i] == num)
    {
      *j = *i;
    }
  }
  return *j;
}


struct whitelist_blacklist *parse_list(char *str,struct whitelist_blacklist *list_p, int32_t wb_type, int32_t id)
{
  char delimiter[] = ",";
  char *tmp,*tmp2,*tok;
  int32_t a,f,t,ch,num;
  struct whitelist_blacklist *list;
  struct whitelist_blacklist_item *list_item;

  tmp = NULL;
  list = NULL;
  list_item = NULL;

  t = 0;
  f = 0;

  if (strlen(str) > 0)
  {
    list = (list_p == NULL ? nmalloc(sizeof(struct whitelist_blacklist)) : list_p);
    if (list_p == NULL)
    {
      memset(list,0,sizeof(struct whitelist_blacklist));
    }

    tmp = nmalloc(strlen(str) + 1);
    memcpy(tmp,str,strlen(str) + 1);

    while ((tok = strtok(((t == 0) ? tmp:NULL),delimiter)) != NULL)
    {
      t++;
      tmp2 = nmalloc(strlen(str) + 1);
      memset(tmp2,0,strlen(str) + 1);
      list_item = nmalloc(sizeof(struct whitelist_blacklist_item));
      memset(list_item,0,sizeof(struct whitelist_blacklist_item));
      f = 0;
      if (sscanf(tok,"%u:%u",&ch,&num) == 2)
      {
        for (a=0;a<list->count;a++)
        {
          if ((list->item[a]->ch == ch) && (list->item[a]->num == num))
          {
            f = 1;
          }
        }

        if (!f)
        {
          list->item = (list->count == 0) ? nmalloc(sizeof(struct whitelist_blacklist_item *)) : nrealloc(list->item,(list->count + 1)*sizeof(struct whitelist_blacklist_item *));
          list->item[list->count++] = list_item;
          list_item->ch  = ch;
          list_item->num = num;
          list_item->kw  = 0;
          list_item->id  = id;
          list_item->wb_type = wb_type;
        }
        f = 0;
        nfree(tmp2);
      } else if (sscanf(tok,"%u:%s",&ch,tmp2) == 2)
      {
        for (a=0;a<list->count;a++)
        {
          if ((list->item[a]->ch == ch) && (list->item[a]->kw ? (strcmp(list->item[a]->keyword,tmp2) == 0) : 0))
          {
            f = 1;
          }
        }
        if (!f)
        {
          list->item = (list->count == 0) ? nmalloc(sizeof(struct whitelist_blacklist_item *)) : nrealloc(list->item,(list->count + 1)*sizeof(struct whitelist_blacklist_item *));
          list->item[list->count++] = list_item;
          list_item->ch  = ch;
          list_item->num = 0;
          list_item->kw  = 1;
          list_item->id  = id;
          list_item->wb_type = wb_type;
          list_item->keyword = tmp2;
        } else
        {
          nfree(tmp2);
        }
        f = 0;

      } else if (sscanf(tok,"%u",&num) == 1)
      {
        for (a=0;a<list->count;a++)
        {
          if ((list->item[a]->ch == -1) && (list->item[a]->num == num))
          {
            f = 1;
          }
        }

        if (!f)
        {
          list->item = (list->count == 0) ? nmalloc(sizeof(struct whitelist_blacklist_item *)) : nrealloc(list->item,(list->count + 1)*sizeof(struct whitelist_blacklist_item *));
          list->item[list->count++] = list_item;
          list_item->ch  = -1;
          list_item->num = num;
          list_item->kw  = 0;
          list_item->id  = id;
          asprintf(&list_item->keyword,"%u",num);
          list_item->wb_type = wb_type;
        }
        f = 0;
        nfree(tmp2);
      } else if (sscanf(tok,"%s",tmp2) == 1)
      {
        for (a=0;a<list->count;a++)
        {
          if ((list->item[a]->ch == -1) && (list->item[a]->kw ? (strcmp(list->item[a]->keyword,tmp2) == 0) : 0))
          {
            f = 1;
          }
        }
        if (!f)
        {
          list->item = (list->count == 0) ? nmalloc(sizeof(struct whitelist_blacklist_item *)) : nrealloc(list->item,(list->count + 1)*sizeof(struct whitelist_blacklist_item *));
          list->item[list->count++] = list_item;
          list_item->ch  = -1;
          list_item->kw  = 1;
          list_item->id  = id;
          list_item->wb_type = wb_type;
          list_item->keyword = tmp2;
        } else
        {
          nfree(tmp2);
        }
        f = 0;
      } else
      {
        nfree(tmp2);
      }
    }
  }
  nfree(tmp);
  return list;
}

struct integer_list *parse_int_csv(char *str,struct integer_list *lst, int32_t precedence)
{
  char *tmp,*tok;
  int32_t a,t,i,j,k,l;
  struct integer_list *list;
  a = 0;
  t = 0;

  if (lst == NULL)
  {
    list = nmalloc(sizeof(struct integer_list));
    memset(list,0,sizeof(struct integer_list));
  } else
  {
    list = lst;
  }

  if ((precedence > list->precedence) || (list->list == NULL))
  {
    list->precedence = precedence;
    tmp = nmalloc(strlen(str) + 1);
    memcpy(tmp,str,strlen(str) + 1);

    list->count = 0;
    if (list->list != NULL)
    {
      nfree(list->list);
      list->list = NULL;
    }

    while ((tok = strtok(((t == 0) ? tmp:NULL),",")) != NULL)
    {
      t++;
      if (sscanf(tok,"%u",&i) == 1)
      {
        list->list = (list->list == NULL ? nmalloc(++list->count*sizeof(int32_t)) : nrealloc(list->list,++list->count*sizeof(int32_t)));
        list->list[list->count - 1] = i;
      }
    }
    nfree(tmp);
  }

  a = 1;
  k = 0;

  while (a)
  {
    for (i=0;i<list->count;i++)
    {
      for (j=0;j<list->count;j++)
      {
        if (i != j)
        {
          if (list->list[i] == list->list[j])
          {
            k = 1;
            l = j;
            goto out;
          }
        }
      }
    }

    out:

    if (k)
    {
      if (j == (list->count - 1))
      {
        list->list = nrealloc(list->list,sizeof(int32_t)*(--list->count));
      } else
      {
        for (i=j;i<(list->count);i++)
        {
          list->list[i] = list->list[i+1];
        }
        list->count--;
        list->list = nrealloc(list->list,sizeof(int32_t)*list->count);
      }

      k = 0;
    } else
    {
      a = 0;
    }
  }
  int_sort(list->list,list->count);
  return list;
};


struct recipients_w parse_rcpt_list_new(struct recipients_w *rec, char *str,int32_t id, int32_t str_index)
{
  char *tok;
  char *tmp;
  int32_t t = 0;
  struct recipients rcpt_temp;
  struct recipients_w *rec_tmp,r_temp;

  rec_tmp = rec;
  memset(&rcpt_temp,0,sizeof(struct recipients));
  tmp = nmalloc(strlen(str) + 1);
  memcpy(tmp,str,strlen(str) + 1);

  while ((tok = strtok(((t == 0) ? tmp:NULL),",")) != NULL)
  {
    t++;
    rcpt_temp.num[_id] = id;
    rcpt_temp.str[str_index] = nmalloc(strlen(tok) + 1);
    memcpy(rcpt_temp.str[str_index],tok,strlen(tok) + 1);

    rec_tmp->count++;
    rec_tmp->rcpt = (rec_tmp->rcpt == NULL ? nmalloc(rec_tmp->count*sizeof(struct recipients *)) : nrealloc(rec_tmp->rcpt,rec_tmp->count*sizeof(struct recipients *)));
    rec_tmp->rcpt[rec_tmp->count - 1] = nmalloc(sizeof(struct recipients));
    memcpy(rec_tmp->rcpt[rec_tmp->count - 1],&rcpt_temp,sizeof(struct recipients));
    memset(&rcpt_temp,0,sizeof(struct recipients));
  }

  memcpy(&r_temp,rec_tmp,sizeof(struct recipients_w));
  nfree(tmp);
  return r_temp;
}

void char_replace(char *mystr, char from_char, char to_char)
{
  int32_t i;

  for (i=0;i<strlen(mystr);i++)
  {
    if (mystr[i] == from_char)
    {
      mystr[i] = to_char;
    }
  }
}

struct ric_lookup *parse_ric_lookup(struct ric_lookup *tbl, char *str)
{
  char *tmp,*tmp2,*tok;
  int32_t i,t,ric,ch,f,g;
  struct ric r_temp;
  struct ric_lookup *r;
  tmp = nmalloc(strlen(str) + 1);
  memcpy(tmp,str,strlen(str) + 1);
  t = 0;
  r = ((tbl == NULL) ? nmalloc(sizeof(struct ric_lookup)) : tbl);
  memset(r,0,sizeof(struct ric_lookup));

  while ((tok = strtok(((t == 0) ? tmp:NULL),",")) != NULL)
  {
    t++;
    tmp2 = nmalloc(strlen(str) + 1);
    memset(tmp2,0,strlen(str) + 1);
    f = 0;
    memset(&r_temp,0,sizeof(struct ric));
    char_replace(tok,' ',255); /* ugly workaround because %s ignores spaces */
    if (sscanf(tok,"%u:%u=%s",&ch,&ric,tmp2) == 3)
    {
      char_replace(tmp2,255,' ');
      f = 1;
      r_temp.ch = ch;
      r_temp.ric = ric;
      r_temp.str = nmalloc(strlen(tmp2) + 1);
      memcpy(r_temp.str,tmp2,strlen(tmp2) + 1);
    } else if (sscanf(tok,"%u=%s",&ric,tmp2) == 2)
    {
      char_replace(tmp2,255,' ');
      f = 1;
      r_temp.ch = -1;
      r_temp.ric = ric;
      r_temp.str = nmalloc(strlen(tmp2) + 1);
      memcpy(r_temp.str,tmp2,strlen(tmp2) + 1);
    }

    g = 1;
    if (f)
    {
      for (i=0;i<r->count;i++)
      {
        if ((r->table[i]->ric == ric) && ((r->table[i]->ch == -1) || (r->table[i]->ch == ch)))
        {
          g = 0;
        }
      }

      if (g)
      {
        r->count++;
        r->table = ((r->table == NULL) ? nmalloc(r->count*sizeof(struct ric *)): nrealloc(r->table,r->count*sizeof(struct ric *)));
        r->table[r->count - 1] = nmalloc(sizeof(struct ric));
        memcpy(r->table[r->count - 1],&r_temp,sizeof(struct ric));
        memset(&r_temp,0,sizeof(struct ric));
      }
    }
  }
  if (tmp != NULL)
  {
    nfree(tmp);
    tmp = NULL;
  }
  return r;
}

char *ric_lookup_str(struct ric_lookup *r, int32_t ric, int32_t ch)
{
  int32_t i;
  char *a;

  a = NULL;


  if (r)
  {
    for (i=0;i<r->count;i++)
    {
      if ((r->table[i]->ric == ric) && ((r->table[i]->ch == -1) || (r->table[i]->ch == ch)))
      {
        a = r->table[i]->str;
        goto end;
      }
    }
    end:
    return a;
  }
  return NULL;
}

void int_sort(int32_t *num, int32_t count)
{
  int32_t i = 1;
  int32_t a,b;

  while (i != 0)
  {
    i = 0;
    for (a=0;a<(count-1);a++)
    {
      if (num[a] > num[a+1])
      {
        i++;
        b = num[a];
        num[a] = num[a+1];
        num[a+1] = b;
      }
    }
  }
}

int32_t stricmp(char *s1, char *s2)
{
  char *p1,*p2;
  int32_t rv;

  p1 = nmalloc(strlen(s1) + 1);
  p2 = nmalloc(strlen(s2) + 1);

  memcpy(p1,s1,strlen(s1) + 1);
  memcpy(p2,s2,strlen(s2) + 1);

  str_tolower(p1);
  str_tolower(p2);

  rv = strcmp(p1,p2);
  nfree(p1);
  nfree(p2);
  return rv;
}

int32_t str_lookup(char *kw,char **str,int32_t count)
{
  int32_t i;
  int32_t f = -1;
  for (i=0;i<count;i++)
  {
    if (stricmp(kw,str[i]) == 0)
    {
      f = i;
    }
  }
  return f;
}

#ifdef USE_MYSQL
void log_mysql(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  int32_t cnt,i,qs_len,qs_offset,t1,t2,p = 0;

  char *qs,*tmp,*z;
  static char *t[5];
  char mysql_insert[] = "INSERT INTO `%s`.`%s` SET";
  int32_t o = 10;
  my_bool b;;
  int32_t mpi = 0;
  static int32_t port;

  if (rcpt->str[db])
  {
    if ((rcpt->num[is_parsed] == 0) && (strlen(rcpt->str[db])>0))
    {
      for (i=0;i<5;i++)
      {
        t[i] = nmalloc(strlen(rcpt->str[db]) + 1);
        memset(t[i],0,strlen(rcpt->str[db]) + 1);
      }

      if (sscanf(rcpt->str[db],"%[^:]:%[^:]:%[^:]:%[^:]:%[^:]:%u",t[0],t[1],t[2],t[3],t[4],&port) == 6)
      {
        for (i=0;i<5;i++)
        {
          t[i] = nrealloc(t[i],strlen(t[i]) + 1);
        }
        rcpt->num[is_parsed] = 1;
      } else
      {
        for (i=0;i<5;i++)
        {
          nfree(t[i]);
        }
      }
    }
  }

  if (rcpt->num[is_parsed])
  {
    if (rcpt->msql == NULL)
    {
      rcpt->msql = nmalloc(sizeof(struct mysql_w));
      memset(rcpt->msql,0,sizeof(struct mysql_w));
    }

    if (rcpt->msql->mysql_logging == NULL)
    {
      rcpt->msql->mysql_logging = nmalloc(sizeof(struct mysql_log_new));
      memset(rcpt->msql->mysql_logging,0,sizeof(struct mysql_log_new));
    }

    if (rcpt->v == NULL)
    {
      rcpt->v = nmalloc(sizeof(void *));
      memset(rcpt->v,0,sizeof(void *));
    }

    if (rcpt->queue == NULL)
    {
      rcpt->queue = nmalloc(sizeof(struct pointer_queue));
      memset(rcpt->queue,0,sizeof(struct pointer_queue));
    }

    #ifdef MYSQL_DATE
    rcpt->msql->mysql_logging->column[e_mysql_item_date]                          = 1;
    #endif

    #ifdef MYSQL_RIC
    rcpt->msql->mysql_logging->column[e_mysql_item_ric]                           = 1;
    #endif

    #ifdef MYSQL_BITS
    rcpt->msql->mysql_logging->column[e_mysql_item_bits]                          = 1;
    #endif

    #ifdef MYSQL_MSG
    rcpt->msql->mysql_logging->column[e_mysql_item_msg]                           = 1;
    #endif

    #ifdef MYSQL_MSG_FIXED
    rcpt->msql->mysql_logging->column[e_mysql_item_msg_fixed_no_crlf]             = 1;
    #endif

    #ifdef MYSQL_RIC_UNCORRECTED
    rcpt->msql->mysql_logging->column[e_mysql_item_ric_uncorrected]               = 1;
    #endif

    #ifdef MYSQL_MSG_UNCORRECTED
    rcpt->msql->mysql_logging->column[e_mysql_item_msg_uncorrected]               = 1;
    #endif

    #ifdef MYSQL_CODEWORD_STATUS
    rcpt->msql->mysql_logging->column[e_mysql_item_codeword_status]               = 1;
    #endif

    #ifdef MYSQL_CODEWORDS
    rcpt->msql->mysql_logging->column[e_mysql_item_codewords]                     = 1;
    #endif

    #ifdef MYSQL_CODEWORDS_UNCORRECTED
    rcpt->msql->mysql_logging->column[e_mysql_item_codewords_uncorrected]         = 1;
    #endif

    #ifdef MYSQL_ERRORED
    rcpt->msql->mysql_logging->column[e_mysql_item_errored]                       = 1;
    #endif

    #ifdef MYSQL_BITRATE
    rcpt->msql->mysql_logging->column[e_mysql_item_bitrate]                       = 1;
    #endif

    #ifdef MYSQL_CHANNEL
    rcpt->msql->mysql_logging->column[e_mysql_item_channel]                       = 1;
    #endif

    rcpt->msql->mysql_logging->db_name  = t[0];
    rcpt->msql->mysql_logging->db_table = t[1];
    rcpt->msql->mysql_logging->username = t[2];
    rcpt->msql->mysql_logging->password = t[3];
    rcpt->msql->mysql_logging->hostname = t[4];
    rcpt->msql->mysql_logging->port     = port;

    rcpt->msql->mysql_row_ptr_table.items = CONFIG_MYSQL_COLUMNS;

    if (rcpt->msql->mysql_row_ptr_table.mysql_item_ptr == NULL)
    {
      rcpt->msql->mysql_row_ptr_table.mysql_item_ptr = nmalloc(sizeof(struct mysql_item *)*rcpt->msql->mysql_row_ptr_table.items);
      memset(rcpt->msql->mysql_row_ptr_table.mysql_item_ptr,0,sizeof(struct mysql_item *)*rcpt->msql->mysql_row_ptr_table.items);
    }

    for (p=0;p<CONFIG_MYSQL_COLUMNS;p++)
    {
      rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[p]       =  &rcpt->msql->mysql_items[p];
      rcpt->msql->mysql_items[p].template                      = mysql_template_strings[p];
      rcpt->msql->mysql_items[p].is_num                        = (int32_t)mysql_is_numeric[p];
      rcpt->msql->mysql_items[p].include                       = rcpt->msql->mysql_logging->column[p];
    }

    if (rcpt->msql->mysql_items[e_mysql_item_date].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_date].length = strlen(msg->date);
      rcpt->msql->mysql_items[e_mysql_item_date].ptr    = (unsigned char *)msg->date;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_ric].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_ric].length = (int32_t)log10((double)(msg->ric ? 10*msg->ric : 10));
      rcpt->msql->mysql_items[e_mysql_item_ric].num    = msg->ric;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_bits].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_bits].length = (int32_t)log10((double)(msg->function_bits ? 10*msg->function_bits : 10));
      rcpt->msql->mysql_items[e_mysql_item_bits].num    = msg->function_bits;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_msg].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_msg].length = msg->msg_len;
      rcpt->msql->mysql_items[e_mysql_item_msg].ptr    = msg->msg;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_msg_fixed_no_crlf].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_msg_fixed_no_crlf].length = strlen((char *)msg->msg_fixed_no_crlf);
      rcpt->msql->mysql_items[e_mysql_item_msg_fixed_no_crlf].ptr    = msg->msg_fixed_no_crlf;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_ric_uncorrected].include)
    {
      if (msg->ric != msg->ric_uncorrected)
      {
        rcpt->msql->mysql_items[e_mysql_item_ric_uncorrected].length  = (int32_t)log10((double)(msg->ric_uncorrected ? 10*msg->ric_uncorrected : 10));
        rcpt->msql->mysql_items[e_mysql_item_ric_uncorrected].num     = msg->ric_uncorrected;
        rcpt->msql->mysql_items[e_mysql_item_ric_uncorrected].discard = 0;
      } else
      {
        rcpt->msql->mysql_items[e_mysql_item_ric_uncorrected].discard = 1;
      }
    }

    if (rcpt->msql->mysql_items[e_mysql_item_msg_uncorrected].include)
    {
      if (msg->errored)
      {
        rcpt->msql->mysql_items[e_mysql_item_msg_uncorrected].length = msg->msg_len;
        rcpt->msql->mysql_items[e_mysql_item_msg_uncorrected].ptr    = msg->msg_uncorrected;
        rcpt->msql->mysql_items[e_mysql_item_msg_uncorrected].discard = 0;
      } else
      {
        rcpt->msql->mysql_items[e_mysql_item_msg_uncorrected].discard = 1;
      }
    }

    if (rcpt->msql->mysql_items[e_mysql_item_codeword_status].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_codeword_status].length = msg->cw_len;
      rcpt->msql->mysql_items[e_mysql_item_codeword_status].ptr    = (unsigned char *)msg->codewords_ecc_flags_str;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_codewords].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_codewords].length = sizeof(uint32_t)*msg->idle_prefixed_codewords_len;
      rcpt->msql->mysql_items[e_mysql_item_codewords].ptr    = (unsigned char *)msg->idle_prefixed_codewords;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_codewords_uncorrected].include)
    {
      if (msg->errored)
      {
        rcpt->msql->mysql_items[e_mysql_item_codewords_uncorrected].length = sizeof(uint32_t)*msg->idle_prefixed_codewords_uncorrected_len;
        rcpt->msql->mysql_items[e_mysql_item_codewords_uncorrected].ptr    = (unsigned char *)msg->idle_prefixed_codewords_uncorrected;
        rcpt->msql->mysql_items[e_mysql_item_codewords_uncorrected].discard = 0;
      } else
      {
        rcpt->msql->mysql_items[e_mysql_item_codewords_uncorrected].discard = 1;
      }
    }

    if (rcpt->msql->mysql_items[e_mysql_item_errored].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_errored].length = (int32_t)log10((double)(msg->errored ? 10*msg->errored : 10));
      rcpt->msql->mysql_items[e_mysql_item_errored].num    = msg->errored;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_bitrate].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_bitrate].length = (int32_t)log10((double)(msg->bitrate ? 10*msg->bitrate : 10));
      rcpt->msql->mysql_items[e_mysql_item_bitrate].num    = msg->bitrate;
    }

    if (rcpt->msql->mysql_items[e_mysql_item_channel].include)
    {
      rcpt->msql->mysql_items[e_mysql_item_channel].length = (int32_t)log10((double)(msg->channel ? 10*msg->channel : 10));
      rcpt->msql->mysql_items[e_mysql_item_channel].num    = msg->channel;
    }

    qs_len = 0;
    qs_len += strlen(mysql_insert) + strlen(rcpt->msql->mysql_logging->db_name) + strlen(rcpt->msql->mysql_logging->db_table) - 2;
    asprintf(&qs,(char *)&mysql_insert,rcpt->msql->mysql_logging->db_name,rcpt->msql->mysql_logging->db_table);

    qs[qs_len - 2] = ' ';
    qs_offset = qs_len - 1;

    for (i=0;i<rcpt->msql->mysql_row_ptr_table.items;i++)
    {
      if ((rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->include) && (!rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->discard))
      {
        if (rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->is_num)
        {
          asprintf(&tmp,"%u",rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->num);
        } else
        {
          tmp = (char*)translate(rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->ptr,rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->length,char_table_mysql_escape);
        }
        qs_len += strlen((char *)tmp) + strlen(rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->template);
        qs = nrealloc(qs,sizeof(char)*qs_len);
        sprintf((char *)qs + qs_offset,rcpt->msql->mysql_row_ptr_table.mysql_item_ptr[i]->template,tmp);
        nfree(tmp);
        sprintf(&qs[qs_len - 3],"%s",", ");
        qs_offset = qs_len - 1;
      }
    }
    memset(&qs[qs_len - 3],0,3);
    if (rcpt->queue->count >= MYSQL_MESSAGE_QUEUE_LIMIT)
    {
      asprintf(&z,"MySQL message output queue limit (%d) reached. Message discarded.\n",MYSQL_MESSAGE_QUEUE_LIMIT);
      print_log(MSG_WARNING,z);
      nfree(qs);
      nfree(z);
    } else
    {
      nenqueue(rcpt->queue,qs);
    }

    if (!(rcpt->msql->mysql_row_ptr_table.connected))
    {
      rcpt->msql->mysql = NULL;
      rcpt->msql->mysql = mysql_init(rcpt->msql->mysql);
      if (rcpt->msql->mysql)
      {
        b = 1;
        mysql_options(rcpt->msql->mysql, MYSQL_OPT_RECONNECT, (void*)&b);
        mysql_options(rcpt->msql->mysql, MYSQL_OPT_READ_TIMEOUT, (void*)&o);
        mysql_options(rcpt->msql->mysql, MYSQL_OPT_WRITE_TIMEOUT, (void*)&o);

        if (mysql_real_connect(rcpt->msql->mysql,
                               rcpt->msql->mysql_logging->hostname,
                               rcpt->msql->mysql_logging->username,
                               rcpt->msql->mysql_logging->password,
                               rcpt->msql->mysql_logging->db_name,
                               rcpt->msql->mysql_logging->port,
                               NULL,
                               0))
        {
          rcpt->msql->mysql_row_ptr_table.connected = 1;
        } else
        {
          asprintf(&z,"Connection to MySQL server failed: %s\n",mysql_error(rcpt->msql->mysql));
          print_log(MSG_ERROR,z);
          nfree(z);
        }
      }
    }

    if (rcpt->msql->mysql_row_ptr_table.connected)
    {
      t1 = mysql_thread_id(rcpt->msql->mysql);
      mpi = mysql_ping(rcpt->msql->mysql);
      if (mpi)
      {
        rcpt->msql->mysql_row_ptr_table.connected = 0;
      } else
      {
        t2 = mysql_thread_id(rcpt->msql->mysql);
        if (t1 != t2)
        {
          mysql_options(rcpt->msql->mysql, MYSQL_OPT_RECONNECT, (void*)&b);
          mysql_options(rcpt->msql->mysql, MYSQL_OPT_READ_TIMEOUT,(void*) &o);
          mysql_options(rcpt->msql->mysql, MYSQL_OPT_WRITE_TIMEOUT, (void*)&o);
        }

        i = 0;
        cnt = rcpt->queue->count;

        while (i<cnt)
        {
          if (!mysql_real_query(rcpt->msql->mysql,(char *)rcpt->queue->ptr[0],(unsigned long)strlen((char *)rcpt->queue->ptr[0])))
          {
            i++;
            nfree(ndequeue(rcpt->queue, &rcpt->num[ii], rcpt->v));
          } else
          {
            asprintf(&z,"MySQL query failed: %s\n",mysql_error(rcpt->msql->mysql));
            print_log(MSG_ERROR,z);
            nfree(z);
            i++;
          }
        }
      }
    }
  } else
  {
    asprintf(&z,"Couldn't parse database string: %s\n",rcpt->str[db]);
    print_log(MSG_ERROR,z);
    nfree(z);
  }
}
#endif


void log_console (struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  unsigned char *msg_tmp1,*msg_tmp2,*msg_eh_tmp,*rl;
  char *msgstr,*ric_tmp;

  rl = (unsigned char*)ric_lookup_str(ric,msg->ric,msg->channel);

  if ((lf->str_num[eh_enable] == yes) && (lf->str_num_is_set[eh_enable]))
  {
    if (msg->ric_status == 0)
    {
      asprintf(&ric_tmp,"%u",msg->ric);
    } else if (msg->ric_status > 0)
    {
      asprintf(&ric_tmp,"%s%u%s",lf->error_highlight[eh_fully_corrected_prefix],msg->ric,lf->error_highlight[eh_fully_corrected_suffix]);
    } else if (msg->ric_status == -1)
    {
      asprintf(&ric_tmp,"%s%u%s",lf->error_highlight[eh_uncorrectable_prefix],msg->ric,lf->error_highlight[eh_uncorrectable_suffix]);
    }
    msg_eh_tmp = error_highlight_new(msg->msg,msg->codewords_ecc_flags,msg->cw_len,msg->function_bits,(unsigned char**)lf->error_highlight);
    msg_tmp1 = translate(msg_eh_tmp,strlen((char*)msg_eh_tmp),char_table_crlf_space);
    nfree(msg_eh_tmp);
    if (rl)
    {
      asprintf(&msgstr,"%s %s (%s):CH%u:%u:%u:%s",msg->date,ric_tmp,rl,msg->channel,msg->bitrate,msg->function_bits,msg_tmp1);
    } else
    {
      asprintf(&msgstr,"%s %s:CH%u:%u:%u:%s",msg->date,ric_tmp,msg->channel,msg->bitrate,msg->function_bits,msg_tmp1);
    }
    nfree(ric_tmp);
    nfree(msg_tmp1);
    msg_tmp2 = string_replace((unsigned char*)msgstr,(unsigned char*)"<<<ANSI_COLOR_PLACEHOLDER>>>",(unsigned char*)ansi_escape);
    printf("%s\n",msg_tmp2);
    nfree(msgstr);
    nfree(msg_tmp2);
  } else
  {
    msg_tmp1 = translate(msg->msg,strlen((char*)msg->msg),char_table_crlf_space);
    asprintf(&ric_tmp,"%u",msg->ric);
    if (rl)
    {
      printf("%s %s (%s):CH%u:%u:%u:%s\n",msg->date,ric_tmp,rl,msg->channel,msg->bitrate,msg->function_bits,msg_tmp1);
    } else
    {
      printf("%s %s:CH%u:%u:%u:%s\n",msg->date,ric_tmp,msg->channel,msg->bitrate,msg->function_bits,msg_tmp1);
    }
    nfree(ric_tmp);
    nfree(msg_tmp1);
  }
}

void log_file(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  unsigned char *msg_tmp1,*rl;
  uint32_t *p;
  char *msgstr,*z;
  char *filename;
  struct tm *now;
  FILE *file;
  time_t tss;

  rl = (unsigned char*)ric_lookup_str(ric,msg->ric,msg->channel);

  if (rcpt->str[logfile])
  {
    msg_tmp1 = translate(msg->msg,strlen((char*)msg->msg),char_table);
    if (rl)
    {
      asprintf(&msgstr,"%s %u (%s):CH%u:%u:%u:%s\n",msg->date,msg->ric,rl,msg->channel,msg->bitrate,msg->function_bits,msg_tmp1);
    } else
    {
      asprintf(&msgstr,"%s %u:CH%u:%u:%u:%s\n",msg->date,msg->ric,msg->channel,msg->bitrate,msg->function_bits,msg_tmp1);
    }
    filename = nmalloc(51);
    pthread_mutex_lock(&time_mutex);
    p = (uint32_t*)&msg->unixtime;
    tss = (time_t)(*p & 0xFFFFFFFF);
    now = localtime((time_t *)&tss);
    strftime(filename,50,rcpt->str[logfile],now);
    pthread_mutex_unlock(&time_mutex);
    if ((file = fopen(filename,"a+")) != NULL)
    {
      fwrite(msgstr,strlen(msgstr),1,file);
      fclose(file);
    } else
    {
      asprintf(&z,"Couldn't open %s\n",filename);
      print_log(MSG_ERROR,z);
      nfree(z);
    }
    nfree(filename);
    nfree(msgstr);
    nfree(msg_tmp1);
  }
}

void log_winmsg(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  FILE *fp;
  char *t,*tm,*rl,*buf,*tmp,*cmd_str,*z;
  int32_t found,i,a;
  char *rc;

  char smbclient_ip[]        = "%s --ip-address=%s --message=%s --user='%s' > /dev/null 2>&1";
  char smbclient_host[]      = "%s --message=%s --user='%s' > /dev/null 2>&1";
  char nmblookup_str[]       = "%s -A %s 2>&1";
  char smbclient_str_id[]    = "To: %u (%s)\n\n%s";
  char smbclient_str_no_id[] = "To: %u\n\n%s";
  const char msg_num[]  = "<03>";
  cmd_str = NULL;
  found = 0;
  buf = nmalloc(10000);
  memset(buf,0,10000);

  if (rcpt->str[host])
  {
    if (sscanf(rcpt->str[host],"%d.%d.%d.%d",&a,&a,&a,&a) == 4)
    {
      asprintf(&tmp,nmblookup_str,lf->str[nmblookup],rcpt->str[host]);
      if (!(fp = popen(tmp,"r")) == 0)
      {
        while ((!(rc = fgets(buf,500,fp)) == 0) && (!found))
        {
          if ((buf[strlen(buf)-1] == 0x0A) || (buf[strlen(buf)-1] == 0x0D))
          {
            buf[strlen(buf)-1] = 0;
          }
          for (i=0;i<strlen(buf);i++)
          {
            if ((buf[strlen(buf)-i-1] == 0x0A) || (buf[strlen(buf)-i-1] == 0x0D))
            {
              buf[strlen(buf)-i-1] = 0;
            }
          }

          if ((strstr((const char *)buf,msg_num) != NULL) && (!found))
          {
            found = 1;
            sscanf(buf,"%s",buf);
            asprintf(&cmd_str,smbclient_ip,lf->str[smbclient],rcpt->str[host],buf,program_string);
          }
        }
        if (found == 0)
        {
          asprintf(&z,"Host %s is either not running, or it's not configured to receive messages\n",rcpt->str[host]);
          print_log(MSG_ERROR,z);
          nfree(z);
        }
        pclose(fp);
      } else
      {
        asprintf(&z,"Unable to execute %s\n",tmp);
        print_log(MSG_ERROR,z);
        nfree(z);
      }
      nfree(tmp);
    } else
    {
      asprintf(&cmd_str,smbclient_host,lf->str[smbclient],rcpt->str[host],program_string);
    }

    if (cmd_str)
    {
      t = (char*)string_replace((unsigned char*)msg->msg_fixed_no_crlf,(unsigned char*)"\n\r",(unsigned char*)"\n");
      tm = (char*)string_replace((unsigned char*)t,(unsigned char*)"\r\n",(unsigned char*)"\n");
      nfree(t);

      if (!(fp = popen(cmd_str,"w")) == 0)
      {
        rl = ric_lookup_str(ric,msg->ric,msg->channel);
        if (rl)
        {
          fprintf(fp,smbclient_str_id,msg->ric,rl,tm);
        } else
        {
          fprintf(fp,smbclient_str_no_id,msg->ric,tm);
        }
        fclose(fp);
      }
      nfree(cmd_str);
      nfree(tm);
    }
  }
  nfree(buf);
}

void log_mail(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  FILE *fp;
  char *rl,*tmp1,*tmp2,*from_str,*ric_str,*z,*mta_str;

  ric_str = NULL;
  tmp1    = NULL;
  tmp2    = NULL;

  asprintf(&tmp1,"%s",msg->msg_fixed_no_crlf);

  tmp2 = (char*)string_replace((unsigned char*)tmp1,(unsigned char*)"\n\r",(unsigned char*)"\n");
  nfree(tmp1);
  tmp1 = (char*)string_replace((unsigned char*)tmp2,(unsigned char*)"\r\n",(unsigned char*)"\n");

  if (!(lf->str[from] == 0))
  {
    if (rcpt->str[to])
    {
      rl = ric_lookup_str(ric,msg->ric,msg->channel);
      if (rl)
      {
        asprintf(&ric_str,"(%s)",rl);
      } else
      {
        ric_str = nmalloc(2);
        memset(ric_str,0,2);
      }

      asprintf(&mta_str,"%s %s",lf->str[mta],rcpt->str[to]);

      if (!(fp = popen(mta_str,"w")) == 0)
      {
        asprintf(&from_str,"\"%s\" <%s>",program_string,lf->str[from]);
        fprintf(fp,"To: %s\nFrom: %s\nSubject: %d %s\n\nDate: %s\nTo:   %d %s\n\nMessage:\n--------\n\n%s\n",
                     rcpt->str[to], from_str, msg->ric, ric_str, msg->date, msg->ric, ric_str, tmp1);
        nfree(from_str);
        fclose(fp);
      } else
      {
        asprintf(&z,"Opening pipe to MTA (%s) failed.\n",lf->str[mta]);
        print_log(MSG_ERROR,z);
        nfree(z);
      }
      nfree(ric_str);
      nfree(mta_str);
    }
  } else
  {
    asprintf(&z,"No sender address specified (from=xxx)\n");
    print_log(MSG_ERROR,z);
    nfree(z);
  }
  nfree(tmp1);
  nfree(tmp2);
}

void log_irc(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  unsigned char *msg_tmp1,*msg_tmp2,*msg_eh_tmp;
  char *msgstr,*ric_tmp,*z;
  unsigned char *rl,*rlt;

  if (rcpt->str[host])
  {
    if (!rcpt->num[connected])
    {
      rcpt->str[_tmps] = nmalloc(strlen(rcpt->str[host]) + 1);
      rcpt->str[_channel] = nmalloc(strlen(rcpt->str[host]) + 1);
      rcpt->str[_ip_addr] = NULL;

      if (sscanf(rcpt->str[host],"%[0-9a-zA-Z.]:%d:%s",rcpt->str[_tmps],(int*)&rcpt->num[_port],rcpt->str[_channel]) == 3)
      {
        if (!((rcpt->str[_tmps][strlen(rcpt->str[_tmps]) - 1] >= '0') && rcpt->str[_tmps][strlen(rcpt->str[_tmps]) - 1] <= '9'))
        {
          rcpt->str[_ip_addr] = gethostname_rr(rcpt->str[_tmps]);
          if (!(rcpt->str[_ip_addr]))
          {
            asprintf(&z,"Unable to resolve %s\n",rcpt->str[_tmps]);
            print_log(MSG_ERROR,z);
            nfree(z);
          }
        } else
        {
          rcpt->str[_ip_addr] = nmalloc(strlen(rcpt->str[_tmps]) + 1);
          memcpy(rcpt->str[_ip_addr],rcpt->str[_tmps],strlen(rcpt->str[_tmps]) + 1);
        }
      } else
      {
        nfree(rcpt->str[_tmps]);
        rcpt->str[_tmps] = NULL;
        nfree(rcpt->str[_channel]);
        rcpt->str[_channel] = NULL;
        asprintf(&z,"Invalid address: %s.\n",rcpt->str[host]);
        print_log(MSG_ERROR,z);
        nfree(z);
      }

      if (rcpt->str[_tmps])
      {
        nfree(rcpt->str[_tmps]);
        rcpt->str[_tmps] = NULL;
      }

      if (rcpt->str[_ip_addr])
      {
        memset(&rcpt->srvr,0,sizeof(struct sockaddr_in));
        rcpt->srvr.sin_family = AF_INET;
        rcpt->srvr.sin_port   = htons(rcpt->num[_port]);
        rcpt->srvr.sin_addr.s_addr = inet_addr(rcpt->str[_ip_addr]);
        if (rcpt->srvr.sin_addr.s_addr == INADDR_NONE)
        {
          asprintf(&z,"Bad address: %s\n",rcpt->str[_ip_addr]);
          print_log(MSG_ERROR,z);
          nfree(z);
        }
        nfree(rcpt->str[_ip_addr]);
        rcpt->str[_ip_addr] = NULL;
      }

      rcpt->num[_socket] = socket(PF_INET,SOCK_STREAM,0);
      if (rcpt->num[_socket] == -1)
      {
        asprintf(&z,"socket: %s\n",strerror(errno));
        print_log(MSG_ERROR,z);
        nfree(z);
      } else
      {
        rcpt->num[fd] = connect(rcpt->num[_socket],(struct sockaddr *)&rcpt->srvr,sizeof(struct sockaddr_in));
        if (rcpt->num[fd] == -1)
        {
          asprintf(&z,"connect: %s\n",strerror(errno));
          print_log(MSG_ERROR,z);
          nfree(z);
          shutdown(rcpt->num[_socket],0);
          close(rcpt->num[_socket]);
        } else
        {
          rcpt->io = fdopen(rcpt->num[_socket],"w+");
          if (rcpt->io)
          {
            rcpt->num[connected] = 1;
          } else
          {
            rcpt->num[connected] = 0;
            asprintf(&z,"fdopen: %s\n",strerror(errno));
            print_log(MSG_ERROR,z);
            nfree(z);
            shutdown(rcpt->num[_socket],0);
            close(rcpt->num[_socket]);
          }
        }
      }
    }
  }

  if (rcpt->num[connected])
  {

    if (write(rcpt->num[_socket],&rcpt->num[_socket],(size_t)0) < 0)
    {
      fclose(rcpt->io);
      rcpt->num[connected] = 0;

      if (rcpt->str[_tmps])
      {
        nfree(rcpt->str[_tmps]);
        rcpt->str[_tmps] = NULL;
      }
      if (rcpt->str[_channel])
      {
        nfree(rcpt->str[_channel]);
        rcpt->str[_channel] = NULL;
      }
    } else
    {
      if ((lf->str_num[eh_enable] == yes) && (lf->str_num_is_set[eh_enable]))
      {
        if (msg->ric_status == 0)
        {
          asprintf(&ric_tmp,"%u",msg->ric);
        } else if (msg->ric_status > 0)
        {
          asprintf(&ric_tmp,"%s%u%s",lf->error_highlight[eh_fully_corrected_prefix],msg->ric,lf->error_highlight[eh_fully_corrected_suffix]);
        } else if (msg->ric_status == -1)
        {
          asprintf(&ric_tmp,"%s%u%s",lf->error_highlight[eh_uncorrectable_prefix],msg->ric,lf->error_highlight[eh_uncorrectable_suffix]);
        }

        msg_eh_tmp = error_highlight_new(msg->msg,msg->codewords_ecc_flags,msg->cw_len,msg->function_bits,(unsigned char**)lf->error_highlight);
        msg_tmp1 = translate(msg_eh_tmp,strlen((char*)msg_eh_tmp),char_table_crlf_space);
        nfree(msg_eh_tmp);
        msg_tmp2 = string_replace((unsigned char*)msg_tmp1,(unsigned char*)"%",(unsigned char*)"\%%");

        rl = (unsigned char*)ric_lookup_str(ric,msg->ric,msg->channel);
        if (rl)
        {
          rlt = string_replace((unsigned char*)rl,(unsigned char*)"%",(unsigned char*)"\%%");
          asprintf(&msgstr,"POCSAG %s (\002RIC\002) %s (\002%s\002) (\002MSG\002) %s\012",rcpt->str[_channel],ric_tmp,rlt,msg_tmp2);
          nfree(rlt);
        } else
        {
          asprintf(&msgstr,"POCSAG %s (\002RIC\002) %s (\002MSG\002) %s\012",rcpt->str[_channel],ric_tmp,msg_tmp2);
        }
        nfree(ric_tmp);
        nfree(msg_tmp2);
        msg_tmp2 = string_replace((unsigned char*)msgstr,(unsigned char*)"<<<COLOR_PLACEHOLDER>>>",(unsigned char*)irc_color);
        fprintf(rcpt->io,(char*)msg_tmp2);
        fflush(rcpt->io);
        nfree(msgstr);
        nfree(msg_tmp1);
        nfree(msg_tmp2);
      } else
      {
        msg_tmp1 = translate(msg->msg,strlen((char*)msg->msg),char_table_crlf_space);
        msg_tmp2 = string_replace((unsigned char*)msg_tmp1,(unsigned char*)"%",(unsigned char*)"\%%");

        rl = (unsigned char*)ric_lookup_str(ric,msg->ric,msg->channel);
        if (rl)
        {
          rlt = string_replace((unsigned char*)rl,(unsigned char*)"%",(unsigned char*)"\%%");
          asprintf(&msgstr,"POCSAG %s (\002RIC\002) %u (\002%s\002) (\002MSG\002) %s\012",rcpt->str[_channel],msg->ric,rlt,msg_tmp2);
          nfree(rlt);
        } else
        {
          asprintf(&msgstr,"POCSAG %s (\002RIC\002) %u (\002MSG\002) %s\012",rcpt->str[_channel],msg->ric,msg_tmp2);
        }
        fprintf(rcpt->io,msgstr);
        fflush(rcpt->io);
        nfree(msgstr);
        nfree(msg_tmp1);
        nfree(msg_tmp2);
      }
    }
  }
}

void log_telnet(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  unsigned char *msg_tmp,*rl;
  char *msgstr,*zz,*q;
  int32_t i,j,z,telnet_log_index;
  j = 0;

  rl = (unsigned char*)ric_lookup_str(ric,msg->ric,msg->channel);
  msg_tmp = translate(msg->msg,strlen((char*)msg->msg),char_table);

  if (rl)
  {
    asprintf(&msgstr,"%s %u (%s):CH%u:%u:%u:%s\n\r",msg->date,msg->ric,rl,msg->channel,msg->bitrate,msg->function_bits,msg_tmp);
  } else
  {
    asprintf(&msgstr,"%s %u:CH%u:%u:%u:%s\n\r",msg->date,msg->ric,msg->channel,msg->bitrate,msg->function_bits,msg_tmp);
  }

  nfree(msg_tmp);
  z = -1;

  pthread_mutex_lock(&telnet_server_mutex);
  pthread_mutex_lock(&int_add_remove);

  for (i=0;i<lf->nl_count;i++)
  {
    if (rcpt->num[_id] == lf->nl[i]->id)
    {
      for (j=0;j<lf->nl[i]->count;j++)
      {
        if (lf->nl[i]->msocket != lf->nl[i]->items[j].num)
        {
          z = write(lf->nl[i]->items[j].num,msgstr,strlen(msgstr));
          if (z!=strlen(msgstr))
          {
            q = sec2time(time(NULL) - lf->nl[i]->items[j].tconn);
            asprintf(&zz,"Telnet server: Connection to %s port %d lost. Time connected: %s.\n",lf->nl[i]->items[j].str,ntohs(lf->nl[i]->items[j].sin_port),q);
            print_log(MSG_NOTICE,zz);
            nfree(zz);
            telnet_connections--;
            if (lf->nl[i]->items[j].io_w)
            {
              fflush(lf->nl[i]->items[j].io_w);
              fclose(lf->nl[i]->items[j].io_w);
              lf->nl[i]->items[j].io_w = NULL;
            }

            if (lf->nl[i]->items[j].io_r)
            {
              fclose(lf->nl[i]->items[j].io_r);
              lf->nl[i]->items[j].io_r = NULL;
            }
            shutdown(lf->nl[i]->items[j].socket_dup,0);
            shutdown(lf->nl[i]->items[j].num,0);
            close(lf->nl[i]->items[j].socket_dup);
            close(lf->nl[i]->items[j].num);
            if (lf->nl[i]->items[j].str)
            {
              nfree(lf->nl[i]->items[j].str);
              lf->nl[i]->items[j].str = NULL;
            }
            nfree(q);
            telnet_log_index = get_log_function_index(config,"telnet");
            lf->nl[i] = int_remove_num(lf->nl[i],lf->nl[i]->items[j].num);
          }
        }
      }
    }
  }
  pthread_mutex_unlock(&int_add_remove);
  pthread_cond_signal(&telnet_server_cond);
  pthread_mutex_unlock(&telnet_server_mutex);
  nfree(msgstr);
}

void log_external(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf, struct ric_lookup *ric)
{
  int r;
  char *rl,*tmp1,*tmp2,*ric_str,*ric_str_temp,*z,*cmd_str;
  ric_str      = NULL;
  ric_str_temp = NULL;
  tmp1         = NULL;
  tmp2         = NULL;
  cmd_str      = NULL;

  asprintf(&tmp1,"%s",msg->msg_fixed_no_crlf);
  tmp2 = (char*)string_replace((unsigned char*)tmp1,(unsigned char*)"\n\r",(unsigned char*)"\n");
  nfree(tmp1);
  tmp1 = (char*)string_replace((unsigned char*)tmp2,(unsigned char*)"\r\n",(unsigned char*)"\n");
  nfree(tmp2);
  tmp2 = (char*)string_replace((unsigned char*)tmp1,(unsigned char*)"\"",(unsigned char*)"\\\"");
  nfree(tmp1);
  tmp1 = (char*)string_replace((unsigned char*)tmp2,(unsigned char*)"`",(unsigned char*)"\\\`");
  nfree(tmp2);
  tmp2 = tmp1;

  if (rcpt->str[external])
  {
    rl = ric_lookup_str(ric,msg->ric,msg->channel);
    if (rl)
    {
      asprintf(&ric_str,"%s",rl);
    } else
    {
      ric_str = nmalloc(2);
      memset(ric_str,0,2);
    }
    ric_str_temp = (char*)string_replace((unsigned char*)ric_str,(unsigned char*)"\"",(unsigned char*)"\\\"");
    nfree(ric_str);
    asprintf(&cmd_str,"%s \"%s\" \"%d\" \"%s\" \"%d\" \"%d\" \"%s\"",rcpt->str[external],msg->date,msg->ric,ric_str_temp,msg->channel,msg->function_bits,tmp2);
    nfree(ric_str_temp);
    r = system(cmd_str);
    nfree(cmd_str);
    if (r == -1)
    {
      asprintf(&z,"Couldn't execute %s\n",rcpt->str[external]);
      print_log(MSG_ERROR,z);
      nfree(z);
    }
  }
  nfree(tmp2);
}

int32_t check_message(struct message_block *msg, struct recipients *rcpt, struct log_functions_config *lf)
{
  int32_t p,y,z;
  int32_t rv = 0;

  p = get_int_index(lf->ntab[message_policy].num_lookup,lf->ntab[message_policy].num_count,rcpt->num[_id],&y,&z);
  p = (p >= 0 ? lf->ntab[message_policy].num[p] : allow);

  switch (p)
  {
    case allow:
    {
      rv =(!((wb_msg_match(msg,&lf->wb_wrap[ric_blacklist],lf,rcpt->num[_id],ric_blacklist))
          |
          (wb_msg_match(msg,&lf->wb_wrap[function_bits_blacklist],lf,rcpt->num[_id],function_bits_blacklist))
          |
          (wb_msg_match(msg,&lf->wb_wrap[keyword_blacklist],lf,rcpt->num[_id],keyword_blacklist))));
    }
    break;

    case drop:
    {
      rv =(((wb_msg_match(msg,&lf->wb_wrap[ric_whitelist],lf,rcpt->num[_id],ric_whitelist))
          |
          (wb_msg_match(msg,&lf->wb_wrap[function_bits_whitelist],lf,rcpt->num[_id],function_bits_whitelist))
          |
          (wb_msg_match(msg,&lf->wb_wrap[keyword_whitelist],lf,rcpt->num[_id],keyword_whitelist))));
    }
    break;
  }
  return rv;
}

int32_t wb_id_index(struct whitelist_blacklist_wrap *wbw, int32_t id)
{
  int32_t i;
  int32_t f = -1;

  for (i=0;i<wbw->id_count;i++)
  {
    if (wbw->id_lookup[i] == id)
    {
      f = i;
      goto end;
    }
  }
  end:;
  return f;
}

int32_t wb_msg_match(struct message_block *msg, struct whitelist_blacklist_wrap *wbw, struct log_functions_config *lf, int32_t id, int32_t wb_type)
{
  /* Checks if a message matches a whitelisting or blacklisting characteristic */

  int32_t a,i,n,y,z;
  int32_t rv = 0;
  char *tmp1,*tmp2;
  tmp1 = NULL;
  tmp2 = NULL;
  a = wb_id_index(wbw,id);

  if (a >= 0)
  {
    for (i=0;i<wbw->wb[a]->count;i++)
    {
      if ((wbw->wb[a]->item[i]->ch == -1) || (wbw->wb[a]->item[i]->ch == msg->channel))
      {
        if ((wbw->wb[a]->item[i]->wb_type == ric_whitelist) || (wbw->wb[a]->item[i]->wb_type == ric_blacklist))
        {
          if (wbw->wb[a]->item[i]->num == msg->ric)
          {
            rv = 1;
            goto end;
          }
        } else if ((wbw->wb[a]->item[i]->wb_type == function_bits_whitelist) || (wbw->wb[a]->item[i]->wb_type == function_bits_blacklist))
        {
          if (wbw->wb[a]->item[i]->num == msg->function_bits)
          {
            rv = 1;
            goto end;
          }
        } else if ((wbw->wb[a]->item[i]->wb_type == keyword_whitelist) || (wbw->wb[a]->item[i]->wb_type == keyword_blacklist))
        {
          n = get_int_index(lf->ntab[keyword_case_sensitive].num_lookup,lf->ntab[keyword_case_sensitive].num_count,id,&y,&z);
          n = (n >= 0 ? lf->ntab[keyword_case_sensitive].num[n] : ((lf->str_num_is_set[keyword_case_sensitive] ? lf->str_num[keyword_case_sensitive] : no)));

          switch (n)
          {
            case yes:
            {
              if ((strlen((char *)msg->msg_fixed) > 0) && (strstr((char *)msg->msg_fixed,wbw->wb[a]->item[i]->keyword) != NULL))
              {
                rv = 1;
                goto end;
              }
            }
            break;

            case no:
            {
              tmp1 = nmalloc(strlen((char *)msg->msg_fixed) + 1);
              tmp2 = nmalloc(strlen(wbw->wb[a]->item[i]->keyword) + 1);
              memcpy(tmp1,msg->msg_fixed,strlen((char *)msg->msg_fixed) + 1);
              memcpy(tmp2,wbw->wb[a]->item[i]->keyword,strlen(wbw->wb[a]->item[i]->keyword) + 1);
              str_tolower(tmp1);
              str_tolower(tmp2);
              if ((strlen(tmp1) > 0) && (strstr(tmp1,tmp2) != NULL))
              {
                rv = 1;
                goto end;
              }
            }
            break;
          }

        }
      }
    }
  } else
  {
    rv = 0;
  }
  end:;
  if (tmp1 != NULL)
  {
    nfree(tmp1);
    tmp1 = NULL;
  }
  if (tmp2 != NULL)
  {
    nfree(tmp2);
    tmp2 = NULL;
  }
  return rv;
}



void *n_malloc(int32_t a,int32_t line)
{
  void *tmp;
  tmp = malloc(a);
#ifdef DEBUG
  int32_t i,found;
  char *z;
  asprintf(&z,"Allocated %u bytes at address %p from a call on line %u\n",a,tmp,line);
  printf("%s",z);
  free(z);

  pthread_mutex_lock(&memory_mutex);
  found = 0;
  for (i=0;i<list.count;i++)
  {
    if (tmp == (void*)list.item[i]->address)
    {
      found = 1;
    }
  }

  if (!found)
  {

    if (list.item == NULL)
    {
      list.item = malloc(sizeof(struct mem_alloc_list)*++list.count);
    } else
    {
      list.item = realloc(list.item,sizeof(struct mem_alloc_list)*++list.count);
    }
    list.item[list.count - 1] = malloc(sizeof(struct mem_alloc_item));
    memset(list.item[list.count - 1],0,sizeof(struct mem_alloc_item));
    list.item[list.count - 1]->address = (long int)tmp;
    list.item[list.count - 1]->line_number = line;
    list.item[list.count - 1]->time = time(NULL);
    list.item[list.count - 1]->size = a;
    list.item[list.count - 1]->malloc = 1;
  }
  pthread_mutex_unlock(&memory_mutex);
  #endif
  return tmp;
}


void *n_realloc(void *a,int32_t b,int32_t line)
{
  void *tmp;
  tmp = realloc(a,b);
  #ifdef DEBUG
  int32_t i,found,pos;
  char *z;

  asprintf(&z,"Reallocating %p (Called from line %u)  New size: %u bytes\n",a,line,b);
  printf("%s",z);
  free(z);
  asprintf(&z,"Reallocated %p to %p (Called from line %u)  New size: %u bytes\n",a,tmp,line,b);
  printf("%s",z);
  free(z);
  pos = 0;
  pthread_mutex_lock(&memory_mutex);
  found = 0;
  for (i=0;i<list.count;i++)
  {
    if (a == (void*)list.item[i]->address)
    {
      found = 1;
      pos = i;
    }
  }

  if (found)
  {
    list.item[pos]->address = (long int)tmp;
    list.item[pos]->line_number = line;
    list.item[pos]->time = time(NULL);
    list.item[pos]->size = b;
    list.item[pos]->malloc = 0;
    list.item[pos]->realloc = 1;
  }
  pthread_mutex_unlock(&memory_mutex);
  #endif
  return tmp;
}

void n_free(void *tmp,int32_t line)
{
  #ifdef DEBUG
  char *z;
  asprintf(&z,"Freeing %p (Called from line %u)\n",tmp,line);
  printf("%s",z);
  free(z);
  #endif
  free(tmp);
}


void dump_mem_list(struct mem_alloc_list *mal)
{
  int32_t i;
  int32_t a;
  FILE *f;
  char *t;
  char *date_str,*date_str_item,*z;
  time_t *p;

  time_t tid_t;
  struct tm *now;
  struct mem_alloc_item *tmp;

  pthread_mutex_lock(&memory_mutex2);

  i = 1;
  while (i != 0)
  {
    i = 0;
    for (a=0;a<(mal->count-1);a++)
    {
      if (mal->item[a]->line_number > mal->item[a+1]->line_number)
      {
        i++;
        tmp = mal->item[a];
        mal->item[a] = mal->item[a+1];
        mal->item[a+1] = tmp;
      }
    }
  }

  date_str = malloc(50);
  time(&tid_t);
  pthread_mutex_lock(&time_mutex);
  now = localtime(&tid_t);
  strftime(date_str,49,date_format_debug_fn,now);
  pthread_mutex_unlock(&time_mutex);
  asprintf(&t,"%s-line.txt",date_str);
  if ((f = fopen(t,"a+")) != NULL)
  {
    date_str_item = malloc(100);
    for (i=0;i<mal->count;i++)
    {
      p = &mal->item[i]->time;
      tid_t = (time_t)(*p & 0xFFFFFFFF);
      pthread_mutex_lock(&time_mutex);
      now = localtime((time_t *)&tid_t);
      strftime(date_str_item,49,date_format_long,now);
      pthread_mutex_unlock(&time_mutex);
      fprintf(f,"Address: %p  Size: %d  Line: %d  Realloc: %d  Malloc: %d  Time: %s\n",(void*)mal->item[i]->address,
                                                                                              mal->item[i]->size,
                                                                                              mal->item[i]->line_number,
                                                                                              mal->item[i]->realloc,
                                                                                              mal->item[i]->malloc,
                                                                                              date_str_item);
    }
    fclose(f);
    asprintf(&z,"Wrote %d entries to %s\n",mal->count,t);
    print_log(MSG_NOTICE,z);
    free(z);
    free(date_str_item);
  } else
  {
    asprintf(&z,"Couldn't open %s\n",t);
    print_log(MSG_ERROR,z);
    free(z);
  }

  free(t);

  i = 1;
  while (i != 0)
  {
    i = 0;
    for (a=0;a<(mal->count-1);a++)
    {
      if (mal->item[a]->size > mal->item[a+1]->size)
      {
        i++;
        tmp = mal->item[a];
        mal->item[a] = mal->item[a+1];
        mal->item[a+1] = tmp;
      }
    }
  }

  time(&tid_t);

  pthread_mutex_lock(&time_mutex);
  now = localtime(&tid_t);
  strftime(date_str,49,date_format_debug_fn,now);
  pthread_mutex_unlock(&time_mutex);

  asprintf(&t,"%s-size.txt",date_str);
  if ((f = fopen(t,"a+")) != NULL)
  {
    date_str_item = malloc(50);
    p = &mal->item[i]->time;
    tid_t = (time_t)(*p & 0xFFFFFFFF);
    for (i=0;i<mal->count;i++)
    {
      pthread_mutex_lock(&time_mutex);
      now = localtime((time_t *)&tid_t);
      strftime(date_str_item,49,date_format_long,now);
      pthread_mutex_unlock(&time_mutex);
      fprintf(f,"Address: %p  Size: %d  Line: %d  Realloc: %d  Malloc: %d  Time: %s\n",(void*)mal->item[i]->address,
                                                                             mal->item[i]->size,
                                                                             mal->item[i]->line_number,
                                                                             mal->item[i]->realloc,
                                                                             mal->item[i]->malloc,
                                                                             date_str_item);
    }
    fclose(f);
    asprintf(&z,"Wrote %d entries to %s\n",mal->count,t);
    print_log(MSG_NOTICE,z);
    nfree(z);
    free(date_str_item);
  } else
  {
    asprintf(&z,"Couldn't open %s\n",t);
    print_log(MSG_ERROR,z);
    free(z);
  }


  free(t);
  free(date_str);
  pthread_cond_signal(&memory_cond2);
  pthread_mutex_unlock(&memory_mutex2);
}

void htonl_ptr(uint32_t *num)
{
  *num = htonl(*num);
}

void ntohl_ptr(uint32_t *num)
{
  *num = ntohl(*num);
}

uint32_t bitmatch(uint32_t num, char *str)
{
  uint32_t i,rv,strnum,tmp;
  rv = 0;
  strnum = 0;

  for (i=0;i<strlen(str);i++)
  {

    if (strlen(str) > 32)
    {
      goto end;
    }

    if ((str[i] != '0') && (str[i] != '1'))
    {
      goto end;
    }
    strnum = ((strnum << 1) | (str[i] - 0x30));
  }

  tmp = num;

  for (i=0;i<(33-strlen(str));i++)
  {
    if ((tmp & ((1 << strlen(str)) - 1)) == strnum)
    {
      rv = 1;
      goto end;
    }
    tmp = (tmp >> 1);
  }
  end:
  return rv;
}

char *parse_eh_string(char *str)
{
  char *tok,*tmp,*newstr;
  int32_t t = 0;
  int32_t len = 0;
  int32_t num;
  newstr = NULL;
  tmp = nmalloc(strlen(str) + 1);
  memcpy(tmp,str,strlen(str) + 1);

  if (sscanf(str,"\"%[^\"]\"",tmp) == 1)
  {
    tmp = nrealloc(tmp,strlen(tmp) + 1);
    return tmp;
  } else
  {
    while ((tok = strtok(((t == 0) ? tmp:NULL),",")) != NULL)
    {
      t++;
      if (sscanf(tok,"0x%02X",&num) == 1)
      {
        len++;
        newstr = (newstr == NULL ? nmalloc(len + 1) : nrealloc(newstr,len + 1));
        newstr[len - 1] = (char)num;
      }
    }
    if (newstr)
    {
      newstr[len + 1] = 0;
    }
    nfree(tmp);
    return newstr;
  }
}

int32_t pipe_handler(void)
{
  return 0;
}

int32_t stats_handler(void)
{
  struct poc_stats s;
  int32_t i;

  if ((config->log_config[main_config_index]->str_num[stats] == yes)
     &&
     (config->log_config[main_config_index]->str_num_is_set[stats] == 1))
  {
    memset(&s,0,sizeof(struct poc_stats));

    if (to_exit)
    {
      printf("Exiting...\n");
      exit(0);
    }

    pthread_mutex_lock(&stats_mutex);
    stats_load(statsfile,&s);
    stats_print(&s,&s_stats);

    for (i=0;i<s.ch_br_count;i++)
    {
      nfree(s.ch_br[i]);
    }
    nfree(s.ch_br);

    to_exit = 1;
    printf("Press CTRL-C again within one second to exit...");
    sleep (1);
    printf("timeout\n");
    to_exit = 0;
    pthread_mutex_unlock(&stats_mutex);
    return 0;
  } else
  {
    exit(0);
  }
}

void dump_config(struct configuration_new *config)
{
  int32_t i,j,k,l;

  for (i=0;i<config->count;i++)
  {
    printf("dump_config(): Configuration for %s (adress %p)\n",config->function_names[i],config->function[i]);
    if ((config->log_config[i]->str_num[enabled] == yes) && config->log_config[i]->str_num_is_set[enabled])
    {
      for (j=0;j<config->log_config[i]->rec.count;j++)
      {
        for (k=0;k<config_rcpt_str_enum_count;k++)
        {
          printf("config->log_config[%s]->rec.rcpt[%d]->str[%s] = %s\n",
                 config->function_names[i],
                 j,
                 config_rcpt_str[k],
                 config->log_config[i]->rec.rcpt[j]->str[k]);
        }

        for (k=0;k<recipients_int_enum_count;k++)
        {
          printf("config->log_config[%s]->rec.rcpt[%d]->num[%s] = %d\n",
                  config->function_names[i],
                  j,
                  config_rcpt_int_str[k],
                  config->log_config[i]->rec.rcpt[j]->num[k]);
        }
      }

      for (j=0;j<config_wb_enum_count;j++)
      {
        for (k=0;k<config->log_config[i]->wb_wrap[j].id_count;k++)
        {
          for (l=0;l<config->log_config[i]->wb_wrap[j].wb[k]->count;l++)
          {
            printf("config->log_config[%s]->wb_wrap[%s].wb[%d]->item[%d]->id = %d\n",
                    config->function_names[i],
                    config_wb_strings[j],
                    k,
                    l,
                    config->log_config[i]->wb_wrap[j].wb[k]->item[l]->id);

            printf("config->log_config[%s]->wb_wrap[%s].wb[%d]->item[%d]->ch = %d\n",
                    config->function_names[i],
                    config_wb_strings[j],
                    k,
                    l,
                    config->log_config[i]->wb_wrap[j].wb[k]->item[l]->ch);

            printf("config->log_config[%s]->wb_wrap[%s].wb[%d]->item[%d]->kw = %d\n",
                    config->function_names[i],
                    config_wb_strings[j],
                    k,
                    l,
                    config->log_config[i]->wb_wrap[j].wb[k]->item[l]->kw);

            printf("config->log_config[%s]->wb_wrap[%s].wb[%d]->item[%d]->num = %d\n",
                    config->function_names[i],
                    config_wb_strings[j],
                    k,
                    l,
                    config->log_config[i]->wb_wrap[j].wb[k]->item[l]->num);

            printf("config->log_config[%s]->wb_wrap[%s].wb[%d]->item[%d]->wb_type = %s\n",
                    config->function_names[i],
                    config_wb_strings[j],
                    k,
                    l,
                    config_wb_strings[config->log_config[i]->wb_wrap[j].wb[k]->item[l]->wb_type]);

            printf("config->log_config[%s]->wb_wrap[%s].wb[%d]->item[%d]->keyword = %s\n",
                    config->function_names[i],
                    config_wb_strings[j],
                    k,
                    l,
                    config->log_config[i]->wb_wrap[j].wb[k]->item[l]->keyword);
          }
        }
      }

      for (j=0;j<config_int_enum_count;j++)
      {
        if (config->log_config[i]->num_is_set[j])
        {
          printf("config->log_config[%s]->num[%s] = %d\n",
                  config->function_names[i],
                  config_int_strings[j],
                  config->log_config[i]->num[j]);
        }
      }

      for (j=0;j<config_str_enum_count;j++)
      {
        if (config->log_config[i]->str_num_is_set[j])
        {
          printf("config->log_config[%s]->str_num[%s] = %s\n",
                  config->function_names[i],
                  config_str_strings[j],
                  config_str_keyword_strings[config->log_config[i]->str_num[j]]
                  );
        }
      }

      for (j=0;j<config_int_array_enum_count;j++)
      {
        if (config->log_config[i]->num_list[j] != NULL)
        {
          if (config->log_config[i]->num_list[j]->count > 0)
          {
            printf("config->log_config[%s]->num_list[%s] =",
                    config->function_names[i],
                    config_int_array_str[j]);
          }

          for (k=0;k<config->log_config[i]->num_list[j]->count;k++)
          {
            printf(" %d",config->log_config[i]->num_list[j]->list[k]);
          }

          if (config->log_config[i]->num_list[j]->count > 0)
          {
            printf("\n");
          }
        }
      }

      for (j=0;j<config_str_enum_count;j++)
      {
        if (config->log_config[i]->str[j] != NULL)
        {
          printf("config->log_config[%s]->str[%s] = %s\n",
                  config->function_names[i],
                  config_str_strings[j],
                  config->log_config[i]->str[j]);
         }
      }

      for (j=0;j<config_mp_cs_enum_count;j++)
      {
        for (k=0;k<config->log_config[i]->ntab[j].num_count;k++)
        {
          printf("config->log_config[%s]->ntab[%s].num_lookup[%d] = %d\n",
                  config->function_names[i],
                  mp_cs_strings[j],
                  k,
                  config->log_config[i]->ntab[j].num_lookup[k]);
          printf("config->log_config[%s]->ntab[%s].num[%d] = %s\n",
                  config->function_names[i],
                  mp_cs_strings[j],
                  k,
                  config_str_keyword_strings[config->log_config[i]->ntab[j].num[k]]);
        }
      }

      for (j=0;j<config_error_highlight_enum_count;j++)
      {
        if (config->log_config[i]->error_highlight[j])
        {
          printf("config->log_config[%s]->error_highlight[%s] = %s\n",
                  config->function_names[i],
                  config_error_highlight_strings[j],
                  config->log_config[i]->error_highlight[j]);
        }
      }

      if (config->log_config[i]->ric_strings != NULL)
      {
        for (j=0;j<config->log_config[i]->ric_strings->count;j++)
        {
          printf("config->log_config[%s]->ric_strings->table[%d]->ch = %d\n",
                 config->function_names[i],
                 j,
                 config->log_config[i]->ric_strings->table[j]->ch);

          printf("config->log_config[%s]->ric_strings->table[%d]->ric = %d\n",
                 config->function_names[i],
                 j,
                 config->log_config[i]->ric_strings->table[j]->ric);

          printf("config->log_config[%s]->ric_strings->table[%d]->str = %s\n",
                 config->function_names[i],
                 j,
                 config->log_config[i]->ric_strings->table[j]->str);
        }
      }
    }
  }
}

void tcp_input(struct tcp_struct *tcp, int32_t length, int32_t (*match_parse)(char *, struct tcp_struct *))
{
  char *z;

  char tmp[100];
  char challenge[100];
  char *response;
  int32_t i;

  memset(&tmp,0,100);
  memset(&challenge,0,100);

  if (!tcp->connected)
  {
    asprintf(&z,"Attempting to connect to %s...\n",tcp->host);
    print_log(MSG_NOTICE,z);
    nfree(z);
    tcp->tmps = nmalloc(strlen(tcp->host) + 1);
    tcp->ip_addr = NULL;

    if (sscanf(tcp->host,"%[0-9a-zA-Z.]:%d",tcp->tmps,(int*)&tcp->port) == 2)
    {
      if (!((tcp->tmps[strlen(tcp->tmps) - 1] >= '0') && tcp->tmps[strlen(tcp->tmps) - 1] <= '9'))
      {
        tcp->ip_addr = gethostname_rr(tcp->tmps);
        if (!(tcp->ip_addr))
        {
          asprintf(&z,"Unable to resolve %s\n",tcp->tmps);
          print_log(MSG_ERROR,z);
          nfree(z);
        }
      } else
      {
        tcp->ip_addr = nmalloc(strlen(tcp->tmps) + 1);
        memcpy(tcp->ip_addr,tcp->tmps,strlen(tcp->tmps) + 1);
      }
    } else
    {
      asprintf(&z,"Invalid address: %s\n",tcp->host);
      print_log(MSG_ERROR,z);
      nfree(z);
      goto out;
    }

    nfree(tcp->tmps);
    if (tcp->ip_addr)
    {
      memset(&tcp->srvr,0,sizeof(struct sockaddr_in));
      tcp->srvr.sin_family = AF_INET;
      tcp->srvr.sin_port   = htons(tcp->port);
      tcp->srvr.sin_addr.s_addr = inet_addr(tcp->ip_addr);
      if (tcp->srvr.sin_addr.s_addr == INADDR_NONE)
      {
        asprintf(&z,"Bad address: %s\n",tcp->ip_addr);
        print_log(MSG_ERROR,z);
        nfree(z);
        goto out;
      }
      nfree(tcp->ip_addr);
      tcp->ip_addr = NULL;
    }

    tcp->socket = socket(PF_INET,SOCK_STREAM,0);
    if (tcp->socket == -1)
    {
      asprintf(&z,"socket: %s\n",strerror(errno));
      print_log(MSG_ERROR,z);
      nfree(z);
    } else
    {
      tcp->fd = connect(tcp->socket,(struct sockaddr *)&tcp->srvr,sizeof(struct sockaddr_in));
      if (tcp->fd == -1)
      {
        asprintf(&z,"connect: %s\n",strerror(errno));
        print_log(MSG_ERROR,z);
        nfree(z);
        shutdown(tcp->socket,0);
        close(tcp->socket);
      } else
      {
        tcp->socket_w = dup(tcp->socket);
        tcp->io_r = fdopen(tcp->socket,"r+");
        tcp->io_w = fdopen(tcp->socket_w,"w+");
        if ((tcp->io_r != NULL) && (tcp->io_w != NULL))
        {
          while (1)
          {
            fgets(tmp,99,tcp->io_r);

            for (i=0;i<strlen(tmp);i++)
            {
              if ((tmp[strlen(tmp)-1] == 0x0A) || (tmp[strlen(tmp)-1] == 0x0D))
              {
                tmp[strlen(tmp)-1] = 0;
              }
            }

            if ((match_parse)(tmp,tcp))
            {
              tcp->connected = 1;
              tcp->hdr_ok = 1;
              if (tcp->header_string)
              {
                nfree(tcp->header_string);
                tcp->header_string = NULL;
              }
              tcp->header_string = nmalloc(strlen(tmp)+1);
              memcpy(tcp->header_string,tmp,strlen(tmp)+1);
              goto q;
            } else if (sscanf(tmp,"AUTHREQ:%80s",challenge) == 1)
            {
              if (tcp->password)
              {
                response = build_sha1_response_string(challenge,tcp->password);
                fprintf(tcp->io_w,"%s\n",response);
                nfree(response);
                fflush(tcp->io_w);
              } else
              {
                asprintf(&z,"Authentication is required, but no password has been configured.\n");
                shutdown(tcp->socket,0);
                shutdown(tcp->socket_w,0);
                close(tcp->socket_w);
                close(tcp->socket);
                print_log(MSG_ERROR,z);
                nfree(z);
                goto q;
              }
            } else if (strcmp(tmp,"AUTH:OK") == 0)
            {
              asprintf(&z,"Authentication successful. Waiting for header string...\n");
              print_log(MSG_NOTICE,z);
              nfree(z);
            } else if (strcmp(tmp,"AUTH:ERR") == 0)
            {
              asprintf(&z,"Authentication failed. Bad password.\n");
              print_log(MSG_ERROR,z);
              nfree(z);
              goto q;
            } else if (strcmp(tmp,"BUSY") == 0)
            {
              asprintf(&z,"Connection to %s failed. Server is busy.\n",tcp->host);
              print_log(MSG_ERROR,z);
              nfree(z);
              tcp->connected = 0;
              tcp->hdr_ok = 0;
              goto q;
            } else if (strcmp(tmp,"TRYLATER") == 0)
            {
              asprintf(&z,"Connection to %s failed. Server asked us to try again later.\n",tcp->host);
              print_log(MSG_ERROR,z);
              nfree(z);
              tcp->connected = 0;
              tcp->hdr_ok = 0;
              goto q;
            } else
            {
              asprintf(&z,"Connection to %s failed. Bad handshake (%s).\n",tcp->host,tmp);
              print_log(MSG_ERROR,z);
              nfree(z);
              tcp->connected = 0;
              tcp->hdr_ok = 0;
              goto q;
            }
          }
        } else
        {
          asprintf(&z,"fdopen: %s\n",strerror(errno));
          print_log(MSG_ERROR,z);
          nfree(z);
          shutdown(tcp->socket,0);
          shutdown(tcp->socket_w,0);
          close(tcp->socket_w);
          close(tcp->socket);
        }
        q:;
      }
    }
  }

  if (tcp->connected)
  {
    tcp->buf = nmalloc(length + 1);
    tcp->length = fread(tcp->buf,length,1,tcp->io_r)*length;

    if ((tcp->length == 0) && (length>0))
    {
      nfree(tcp->buf);
      tcp->buf = NULL;
      tcp->connected = 0;
      tcp->hdr_ok = 0;
      if (tcp->bitrate_list)
      {
        if (tcp->bitrate_list->list)
        {
          nfree(tcp->bitrate_list->list);
          tcp->bitrate_list->list = NULL;
        }
        nfree(tcp->bitrate_list);
        tcp->bitrate_list = NULL;
      }

      if (tcp->io_w)
      {
        fflush(tcp->io_w);
        fclose(tcp->io_w);
        tcp->io_w = NULL;
      }

      if (tcp->io_r)
      {
        fclose(tcp->io_r);
        tcp->io_r = NULL;
      }

      shutdown(tcp->socket,0);
      shutdown(tcp->socket_w,0);
      close(tcp->socket);
      close(tcp->socket_w);
    }

  }
  out:;
}

void stats_add(struct poc_stats *s, uint32_t bitrate, uint32_t channel,
                  uint32_t msg_alpha_ok, uint32_t msg_alpha_err,
                  uint32_t msg_numeric_ok, uint32_t msg_numeric_err,
                  uint32_t cw_ok, uint32_t cw_err)
{
  int32_t i;

  for (i=0;i<s->ch_br_count;i++)
  {
    if ((s->ch_br[i]->bitrate == bitrate)  && (s->ch_br[i]->channel == channel))
    {
      s->ch_br[i]->msg_alpha_ok    += msg_alpha_ok;
      s->ch_br[i]->msg_alpha_err   += msg_alpha_err;
      s->ch_br[i]->msg_numeric_ok  += msg_numeric_ok;
      s->ch_br[i]->msg_numeric_err += msg_numeric_err;
      s->ch_br[i]->cw_ok           += cw_ok;
      s->ch_br[i]->cw_err          += cw_err;
    }
  }
}

void stats_add_to_file(char *stats_file, struct poc_stats *s)
{
  int32_t err,x,z,fd,fdo,i,i_pos,j,j_pos,k;
  struct poc_stats stat_tmp;
  struct ch_br_stats ch_br_tmp;
  char *fn,*zz;
  char t[128];
  FILE *tmp,*tmp_w;
  err = 0;
  tmp = NULL;
  tmp_w = NULL;

  j_pos = 0;
  i_pos = 0;

  fd = open(stats_file,O_RDONLY);
  if (fd>0)
  {
    asprintf(&fn,"%s.bak",stats_file);
    fdo = open(fn,O_WRONLY|O_CREAT,0644);
    if (fdo>0)
    {
      x = 1;
      while (x)
      {
        x = read(fd,&t,128);
        if (x>0)
        {
          z = write(fdo,&t,x);
          if (z != x)
          {
            err = 1;
          }
        }
      }

      close(fdo);
    } else
    {

      perror("open:");
      err = 1;
    }
    nfree(fn);
    close(fd);
  }

  memset(&stat_tmp,0,sizeof(struct poc_stats));
  memset(&ch_br_tmp,0,sizeof(struct ch_br_stats));

  if (!err)
  {
    if ((tmp = fopen(stats_file,"r")) != NULL)
    {
      if (fread((char*)&stat_tmp.s,sizeof(struct stats),1,tmp) > 0)
      {
        while (fread((char*)&ch_br_tmp,sizeof(struct ch_br_stats),1,tmp) == 1)
        {
          if (stat_tmp.ch_br == NULL)
          {
            stat_tmp.ch_br = nmalloc(sizeof(struct ch_br_stats*)*(++stat_tmp.ch_br_count));
            stat_tmp.ch_br[stat_tmp.ch_br_count - 1] = nmalloc(sizeof(struct ch_br_stats));
            memcpy(stat_tmp.ch_br[stat_tmp.ch_br_count - 1],&ch_br_tmp,sizeof(struct ch_br_stats));
          } else
          {
            stat_tmp.ch_br = nrealloc(stat_tmp.ch_br,sizeof(struct ch_br_stats*)*(++stat_tmp.ch_br_count));
            stat_tmp.ch_br[stat_tmp.ch_br_count - 1] = nmalloc(sizeof(struct ch_br_stats));
            memcpy(stat_tmp.ch_br[stat_tmp.ch_br_count - 1],&ch_br_tmp,sizeof(struct ch_br_stats));
          }
        }

        if (stat_tmp.ch_br_count > 0)
        {
          for (i=0;i<s->ch_br_count;i++)
          { k = 0;
            for (j=0;j<stat_tmp.ch_br_count;j++)
            {
              if ((stat_tmp.ch_br[j]->bitrate == s->ch_br[i]->bitrate) &&
                  (stat_tmp.ch_br[j]->channel == s->ch_br[i]->channel))
              {
                j_pos = j;
                i_pos = i;
                k++;
              }
            }

            if (k == 0)
            {
              if (stat_tmp.ch_br == NULL)
              {
                stat_tmp.ch_br = nmalloc(sizeof(struct ch_br_stats *)*(++stat_tmp.ch_br_count));
                stat_tmp.ch_br[stat_tmp.ch_br_count - 1] = nmalloc(sizeof(struct ch_br_stats));
                memcpy(stat_tmp.ch_br[stat_tmp.ch_br_count - 1],s->ch_br[i],sizeof(struct ch_br_stats));
              } else
              {
                stat_tmp.ch_br = nrealloc(stat_tmp.ch_br,sizeof(struct ch_br_stats *)*(++stat_tmp.ch_br_count));
                stat_tmp.ch_br[stat_tmp.ch_br_count - 1] = nmalloc(sizeof(struct ch_br_stats));
                memcpy(stat_tmp.ch_br[stat_tmp.ch_br_count - 1],s->ch_br[i],sizeof(struct ch_br_stats));
              }
            } else
            {
              stat_tmp.ch_br[j_pos]->msg_alpha_ok    += s->ch_br[i_pos]->msg_alpha_ok;
              s->ch_br[i_pos]->msg_alpha_ok          = 0;

              stat_tmp.ch_br[j_pos]->msg_alpha_err   += s->ch_br[i_pos]->msg_alpha_err;
              s->ch_br[i_pos]->msg_alpha_err         = 0;

              stat_tmp.ch_br[j_pos]->msg_numeric_ok  += s->ch_br[i_pos]->msg_numeric_ok;
              s->ch_br[i_pos]->msg_numeric_ok        = 0;

              stat_tmp.ch_br[j_pos]->msg_numeric_err += s->ch_br[i_pos]->msg_numeric_err;
              s->ch_br[i_pos]->msg_numeric_err       = 0;

              stat_tmp.ch_br[j_pos]->cw_ok           += s->ch_br[i_pos]->cw_ok;
              s->ch_br[i_pos]->cw_ok                 = 0;

              stat_tmp.ch_br[j_pos]->cw_err          += s->ch_br[i_pos]->cw_err;
              s->ch_br[i_pos]->cw_err                = 0;
            }
          }

          stat_tmp.s.time_running += (time(NULL) - s->s.last_updated);
          s->s.last_updated = time(NULL);
          s->s.first_run = stat_tmp.s.first_run;
          s->s.time_running = stat_tmp.s.time_running;

          stat_tmp.s.pcm_mbyte += s->s.pcm_mbyte;
          s->s.pcm_mbyte = 0;



          if ((tmp_w = fopen(stats_file,"w")) != NULL)
          {
            fwrite((char*)&stat_tmp.s,sizeof(struct stats),1,tmp_w);
            for (i=0;i<stat_tmp.ch_br_count;i++)
            {
              fwrite((char*)stat_tmp.ch_br[i],sizeof(struct ch_br_stats),1,tmp_w);
              nfree(stat_tmp.ch_br[i]);
              stat_tmp.ch_br[i] = NULL;
            }
            stat_tmp.ch_br_count = 0;
            if (stat_tmp.ch_br)
            {
              nfree(stat_tmp.ch_br);
              stat_tmp.ch_br = NULL;
            }

            fclose(tmp_w);
            tmp_w = NULL;
          } else
          {

            asprintf(&zz,"Couldn't open %s for writing\n",stats_file);
            print_log(MSG_ERROR,zz);
            nfree(zz);
          }
        }
      }
      fclose(tmp);
      tmp = NULL;
    } else
    {
      if ((tmp_w = fopen(stats_file,"w")) != NULL)
      {
        fwrite((char*)&s->s,sizeof(struct stats),1,tmp_w);
        for (i=0;i<s->ch_br_count;i++)
        {
          fwrite((char*)s->ch_br[i],sizeof(struct ch_br_stats),1,tmp_w);
          s->ch_br[i]->msg_alpha_ok    = 0;
          s->ch_br[i]->msg_alpha_err   = 0;
          s->ch_br[i]->msg_numeric_ok  = 0;
          s->ch_br[i]->msg_numeric_err = 0;
          s->ch_br[i]->cw_ok           = 0;
          s->ch_br[i]->cw_err          = 0;
        }
        fclose(tmp_w);
        tmp_w = NULL;
      } else
      {
        asprintf(&zz,"Couldn't open %s for writing\n",stats_file);
        print_log(MSG_ERROR,zz);
        nfree(zz);
      }
    }
  } else
  {
    asprintf(&fn,"%s.bak",stats_file);
    asprintf(&zz,"Couldn't create backup file %s. Will not attempt to update stats file %s\n",fn,stats_file);
    print_log(MSG_ERROR,zz);
    nfree(zz);
    nfree(fn);
  }
}

void stats_load(char *stats_file, struct poc_stats *s)
{
  int32_t i;
  char *zz;
  struct ch_br_stats ch_br_tmp;
  memset(&ch_br_tmp,0,sizeof(struct ch_br_stats));
  FILE *tmp1;
  for (i=0;i<s->ch_br_count;i++)
  {
    nfree(s->ch_br[i]);
    s->ch_br[i] = NULL;
  }

  memset(s,0,sizeof(struct poc_stats));

  if ((tmp1 = fopen(stats_file,"r")) != NULL)
  {
    if (fread((char*)&s->s,sizeof(struct stats),1,tmp1) > 0)
    {
      while (fread((char*)&ch_br_tmp,sizeof(struct ch_br_stats),1,tmp1) == 1)
      {
        if (s->ch_br == NULL)
        {
          s->ch_br = nmalloc(sizeof(struct ch_br_stats*)*(++s->ch_br_count));
          s->ch_br[s->ch_br_count - 1] = nmalloc(sizeof(struct ch_br_stats));
          memcpy(s->ch_br[s->ch_br_count - 1],&ch_br_tmp,sizeof(struct ch_br_stats));
        } else
        {
          s->ch_br = nrealloc(s->ch_br,sizeof(struct ch_br_stats*)*(++s->ch_br_count));
          s->ch_br[s->ch_br_count - 1] = nmalloc(sizeof(struct ch_br_stats));
          memcpy(s->ch_br[s->ch_br_count - 1],&ch_br_tmp,sizeof(struct ch_br_stats));
        }
      }
      fclose(tmp1);
    }
  } else
  {
    asprintf(&zz,"Couldn't open %s for reading\n",stats_file);
    print_log(MSG_ERROR,zz);
    nfree(zz);
  }
}

void stats_print(struct poc_stats *total_stats, struct poc_stats *session_stats)
{

  time_t *p,tss;
  int32_t *s_ch_int,ch_index,s_ch_cnt,ch_found,i,j,k;
  int32_t *s_br_int,br_index,s_br_cnt,br_found;
  int32_t *t_ch_int,*t_br_int;
  int32_t t_br_cnt,t_ch_cnt;

  struct ch_br_stats **s_br;
  struct ch_br_stats **s_ch;
  struct ch_br_stats s_all;

  struct ch_br_stats **t_br;
  struct ch_br_stats **t_ch;
  struct ch_br_stats t_all;

  struct tm *now;
  char *date_str;
  char *tmp;

  s_ch_int = NULL;
  s_ch_cnt = 0;
  s_br_int = NULL;
  s_br_cnt = 0;

  ch_index = 0;
  ch_found = 0;
  br_index = 0;
  br_found = 0;

  t_ch_int = NULL;
  t_br_int = NULL;
  t_br_cnt = 0;
  t_ch_cnt = 0;

  s_ch     = NULL;
  s_br     = NULL;
  t_ch     = NULL;
  t_br     = NULL;

  memset(&s_all,0,sizeof(struct ch_br_stats));
  memset(&t_all,0,sizeof(struct ch_br_stats));

  if (session_stats != NULL)
  {
    for (i=0;i<session_stats->ch_br_count;i++)
    {
      ch_found = 0;
      for (j=0;j<s_ch_cnt;j++)
      {
        if (s_ch_int[j] == session_stats->ch_br[i]->channel)
        {
          ch_found = 1;
        }
      }

      if (!ch_found)
      {
        if (s_ch_int == NULL)
        {
          s_ch_int = nmalloc(sizeof(int32_t)*++s_ch_cnt);
        } else
        {
          s_ch_int = nrealloc(s_ch_int,sizeof(int32_t)*++s_ch_cnt);
        }
        s_ch_int[s_ch_cnt - 1] = session_stats->ch_br[i]->channel;
      }

      br_found = 0;
      for (j=0;j<s_br_cnt;j++)
      {
        if (s_br_int[j] == session_stats->ch_br[i]->bitrate)
        {
          br_found = 1;
        }
      }

      if (!br_found)
      {
        if (s_br_int == NULL)
        {
          s_br_int = nmalloc(sizeof(int32_t)*++s_br_cnt);
        } else
        {
          s_br_int = nrealloc(s_br_int,sizeof(int32_t)*++s_br_cnt);
        }
        s_br_int[s_br_cnt - 1] = session_stats->ch_br[i]->bitrate;
      }
    }

    if (s_ch_cnt>0)
    {
      s_ch = nmalloc(sizeof(struct ch_br_stats *)*s_ch_cnt);
      for (i=0;i<s_ch_cnt;i++)
      {
        s_ch[i] = nmalloc(sizeof(struct ch_br_stats));
        memset(s_ch[i],0,sizeof(struct ch_br_stats));
        s_ch[i]->channel = s_ch_int[i];
      }
    }

    if (s_br_cnt>0)
    {
      s_br = nmalloc(sizeof(struct ch_br_stats *)*s_br_cnt);
      for (i=0;i<s_br_cnt;i++)
      {
        s_br[i] = nmalloc(sizeof(struct ch_br_stats));
        memset(s_br[i],0,sizeof(struct ch_br_stats));
        s_br[i]->bitrate = s_br_int[i];
      }
    }

    for (i=0;i<session_stats->ch_br_count;i++)
    {
      for (j=0;j<s_ch_cnt;j++)
      {
        if (s_ch[j]->channel == session_stats->ch_br[i]->channel)
        {
          s_ch[j]->msg_alpha_ok    += session_stats->ch_br[i]->msg_alpha_ok;
          s_ch[j]->msg_alpha_err   += session_stats->ch_br[i]->msg_alpha_err;
          s_ch[j]->msg_numeric_ok  += session_stats->ch_br[i]->msg_numeric_ok;
          s_ch[j]->msg_numeric_err += session_stats->ch_br[i]->msg_numeric_err;
          s_ch[j]->cw_ok           += session_stats->ch_br[i]->cw_ok;
          s_ch[j]->cw_err          += session_stats->ch_br[i]->cw_err;
        }
      }
    }

    for (i=0;i<session_stats->ch_br_count;i++)
    {
      for (j=0;j<s_br_cnt;j++)
      {
        if (s_br[j]->bitrate == session_stats->ch_br[i]->bitrate)
        {
          s_br[j]->msg_alpha_ok    += session_stats->ch_br[i]->msg_alpha_ok;
          s_br[j]->msg_alpha_err   += session_stats->ch_br[i]->msg_alpha_err;
          s_br[j]->msg_numeric_ok  += session_stats->ch_br[i]->msg_numeric_ok;
          s_br[j]->msg_numeric_err += session_stats->ch_br[i]->msg_numeric_err;
          s_br[j]->cw_ok           += session_stats->ch_br[i]->cw_ok;
          s_br[j]->cw_err          += session_stats->ch_br[i]->cw_err;
        }
      }
    }

    for (i=0;i<session_stats->ch_br_count;i++)
    {
      s_all.msg_alpha_ok    += session_stats->ch_br[i]->msg_alpha_ok;
      s_all.msg_alpha_err   += session_stats->ch_br[i]->msg_alpha_err;
      s_all.msg_numeric_ok  += session_stats->ch_br[i]->msg_numeric_ok;
      s_all.msg_numeric_err += session_stats->ch_br[i]->msg_numeric_err;
      s_all.cw_ok           += session_stats->ch_br[i]->cw_ok;
      s_all.cw_err          += session_stats->ch_br[i]->cw_err;
    }
  }

  if (total_stats != NULL)
  {
    for (i=0;i<total_stats->ch_br_count;i++)
    {
      ch_found = 0;
      for (j=0;j<t_ch_cnt;j++)
      {
        if (t_ch_int[j] == total_stats->ch_br[i]->channel)
        {
          ch_found = 1;
        }
      }

      if (!ch_found)
      {
        if (t_ch_int == NULL)
        {
          t_ch_int = nmalloc(sizeof(int32_t)*++t_ch_cnt);
        } else
        {
          t_ch_int = nrealloc(t_ch_int,sizeof(int32_t)*++t_ch_cnt);
        }
        t_ch_int[t_ch_cnt - 1] = total_stats->ch_br[i]->channel;
      }

      br_found = 0;
      for (j=0;j<t_br_cnt;j++)
      {
        if (t_br_int[j] == total_stats->ch_br[i]->bitrate)
        {
          br_found = 1;
        }
      }

      if (!br_found)
      {
        if (t_br_int == NULL)
        {
          t_br_int = nmalloc(sizeof(int32_t)*++t_br_cnt);
        } else
        {
          t_br_int = nrealloc(t_br_int,sizeof(int32_t)*++t_br_cnt);
        }
        t_br_int[t_br_cnt - 1] = total_stats->ch_br[i]->bitrate;
      }
    }

    if (t_ch_cnt>0)
    {
      t_ch = nmalloc(sizeof(struct ch_br_stats *)*t_ch_cnt);
      for (i=0;i<t_ch_cnt;i++)
      {
        t_ch[i] = nmalloc(sizeof(struct ch_br_stats));
        memset(t_ch[i],0,sizeof(struct ch_br_stats));
        t_ch[i]->channel = t_ch_int[i];
      }
    }

    if (t_br_cnt>0)
    {
      t_br = nmalloc(sizeof(struct ch_br_stats *)*t_br_cnt);
      for (i=0;i<t_br_cnt;i++)
      {
        t_br[i] = nmalloc(sizeof(struct ch_br_stats));
        memset(t_br[i],0,sizeof(struct ch_br_stats));
        t_br[i]->bitrate = t_br_int[i];
      }
    }

    for (i=0;i<total_stats->ch_br_count;i++)
    {
      for (j=0;j<t_ch_cnt;j++)
      {
        if (t_ch[j]->channel == total_stats->ch_br[i]->channel)
        {
          t_ch[j]->msg_alpha_ok    += total_stats->ch_br[i]->msg_alpha_ok;
          t_ch[j]->msg_alpha_err   += total_stats->ch_br[i]->msg_alpha_err;
          t_ch[j]->msg_numeric_ok  += total_stats->ch_br[i]->msg_numeric_ok;
          t_ch[j]->msg_numeric_err += total_stats->ch_br[i]->msg_numeric_err;
          t_ch[j]->cw_ok           += total_stats->ch_br[i]->cw_ok;
          t_ch[j]->cw_err          += total_stats->ch_br[i]->cw_err;
        }
      }
    }

    for (i=0;i<total_stats->ch_br_count;i++)
    {
      for (j=0;j<t_br_cnt;j++)
      {
        if (t_br[j]->bitrate == total_stats->ch_br[i]->bitrate)
        {
          t_br[j]->msg_alpha_ok    += total_stats->ch_br[i]->msg_alpha_ok;
          t_br[j]->msg_alpha_err   += total_stats->ch_br[i]->msg_alpha_err;
          t_br[j]->msg_numeric_ok  += total_stats->ch_br[i]->msg_numeric_ok;
          t_br[j]->msg_numeric_err += total_stats->ch_br[i]->msg_numeric_err;
          t_br[j]->cw_ok           += total_stats->ch_br[i]->cw_ok;
          t_br[j]->cw_err          += total_stats->ch_br[i]->cw_err;
        }
      }
    }

    for (i=0;i<total_stats->ch_br_count;i++)
    {
      t_all.msg_alpha_ok    += total_stats->ch_br[i]->msg_alpha_ok;
      t_all.msg_alpha_err   += total_stats->ch_br[i]->msg_alpha_err;
      t_all.msg_numeric_ok  += total_stats->ch_br[i]->msg_numeric_ok;
      t_all.msg_numeric_err += total_stats->ch_br[i]->msg_numeric_err;
      t_all.cw_ok           += total_stats->ch_br[i]->cw_ok;
      t_all.cw_err          += total_stats->ch_br[i]->cw_err;
    }
  }
  date_str = nmalloc(50);
  p = &session_stats->s.started;
  if (session_stats != NULL)
  {
    pthread_mutex_lock(&time_mutex);
    tss = (time_t)(*p & 0xFFFFFFFF);
    now = localtime((time_t *)&tss);
    strftime(date_str,49,date_format_long,now);
    pthread_mutex_unlock(&time_mutex);
    tmp = sec2time(time(NULL) - session_stats->s.started);
    printf("\n");
    printf("Session statistics\n");
    printf("------------------\n");
    printf("Launched:             %s\n",date_str);
    printf("Time running:         %s\n",tmp);
    nfree(tmp);
    tmp = NULL;
    for(i=0;i<s_ch_cnt;i++)
    {
      if (s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err + s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err)
      {
        printf("Channel:              %d\n",s_ch_int[i]);

        printf("Messages:             %d\n",(s_ch[i]->msg_alpha_ok   +
                                             s_ch[i]->msg_alpha_err  +
                                             s_ch[i]->msg_numeric_ok +
                                             s_ch[i]->msg_numeric_err));

        if (s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err)
        {
          printf("                      %d%*s  Alpha      (%.2f\%%)\n",(s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err),
                                                                       len_diff(s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err,s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err + s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err),
                                                                       "",
                                                                       ((double)(s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err)/(double)(s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err + s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err))*100);
        }

        if (s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err)
        {
          printf("                      %d%*s  Numeric    (%.2f\%%)\n",(s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err),
                                                                       len_diff(s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err,s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err + s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err),
                                                                       "",
                                                                       ((double)(s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err)/(double)(s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err + s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err))*100);
        }

        for (j=0;j<s_br_cnt;j++)
        {
          for (k=0;k<session_stats->ch_br_count;k++)
          {
            if ((session_stats->ch_br[k]->channel == s_ch_int[i]) &&  (session_stats->ch_br[k]->bitrate == s_br_int[j]))
            {
              if (session_stats->ch_br[k]->msg_alpha_ok + session_stats->ch_br[k]->msg_alpha_err + session_stats->ch_br[k]->msg_numeric_ok + session_stats->ch_br[k]->msg_numeric_err)
              {
                printf("                      %d%*s  @ %4d bps (%.2f\%%)\n",(session_stats->ch_br[k]->msg_alpha_ok   +
                                                                             session_stats->ch_br[k]->msg_alpha_err  +
                                                                             session_stats->ch_br[k]->msg_numeric_ok +
                                                                             session_stats->ch_br[k]->msg_numeric_err),
                                                                            len_diff(session_stats->ch_br[k]->msg_alpha_ok   +
                                                                                     session_stats->ch_br[k]->msg_alpha_err  +
                                                                                     session_stats->ch_br[k]->msg_numeric_ok +
                                                                                     session_stats->ch_br[k]->msg_numeric_err,
                                                                                     s_ch[i]->msg_alpha_ok   +
                                                                                     s_ch[i]->msg_alpha_err  +
                                                                                     s_ch[i]->msg_numeric_ok +
                                                                                     s_ch[i]->msg_numeric_err),
                                                                            "",
                                                                            s_br_int[j],
                                                                            (double)(session_stats->ch_br[k]->msg_alpha_ok + session_stats->ch_br[k]->msg_alpha_err  +  session_stats->ch_br[k]->msg_numeric_ok + session_stats->ch_br[k]->msg_numeric_err)/(double)(s_ch[i]->msg_alpha_ok + s_ch[i]->msg_alpha_err + s_ch[i]->msg_numeric_ok + s_ch[i]->msg_numeric_err)*100);
              }
            }
          }
        }
        printf("\n");
      }
    }
/******************/
    if (s_all.msg_alpha_ok + s_all.msg_alpha_err + s_all.msg_numeric_ok + s_all.msg_numeric_err)
    {

      printf("Total messages:       %d\n",(s_all.msg_alpha_ok  +
                                           s_all.msg_alpha_err  +
                                           s_all.msg_numeric_ok +
                                           s_all.msg_numeric_err));

      if (s_all.msg_alpha_ok + s_all.msg_alpha_err)
      {
        printf("                      %d%*s  Alpha      (%.2f\%%)\n",(s_all.msg_alpha_ok + s_all.msg_alpha_err),
                                                                     len_diff(s_all.msg_alpha_ok + s_all.msg_alpha_err,s_all.msg_alpha_ok + s_all.msg_alpha_err + s_all.msg_numeric_ok + s_all.msg_numeric_err),
                                                                     "",
                                                                     ((double)(s_all.msg_alpha_ok + s_all.msg_alpha_err)/(double)(s_all.msg_alpha_ok + s_all.msg_alpha_err + s_all.msg_numeric_ok + s_all.msg_numeric_err))*100);
      }

      if (s_all.msg_numeric_ok + s_all.msg_numeric_err)
      {
        printf("                      %d%*s  Numeric    (%.2f\%%)\n",(s_all.msg_numeric_ok + s_all.msg_numeric_err),
                                                                     len_diff(s_all.msg_numeric_ok + s_all.msg_numeric_err,s_all.msg_alpha_ok + s_all.msg_alpha_err + s_all.msg_numeric_ok + s_all.msg_numeric_err),
                                                                     "",
                                                                     ((double)(s_all.msg_numeric_ok + s_all.msg_numeric_err)/(double)(s_all.msg_alpha_ok + s_all.msg_alpha_err + s_all.msg_numeric_ok + s_all.msg_numeric_err))*100);
      }

      for (j=0;j<s_br_cnt;j++)
      {
        if (s_br[j]->msg_alpha_ok + s_br[j]->msg_alpha_err +  s_br[j]->msg_numeric_ok + s_br[j]->msg_numeric_err)
        {
          printf("                      %d%*s  @ %4d bps (%.2f\%%)\n",(s_br[j]->msg_alpha_ok   +
                                                                       s_br[j]->msg_alpha_err  +
                                                                       s_br[j]->msg_numeric_ok +
                                                                       s_br[j]->msg_numeric_err),
                                                                      len_diff(s_br[j]->msg_alpha_ok   +
                                                                               s_br[j]->msg_alpha_err  +
                                                                               s_br[j]->msg_numeric_ok +
                                                                               s_br[j]->msg_numeric_err,
                                                                               s_all.msg_alpha_ok   +
                                                                               s_all.msg_alpha_err  +
                                                                               s_all.msg_numeric_ok +
                                                                               s_all.msg_numeric_err),
                                                                       "",
                                                                      s_br_int[j],
                                                                     (double)(s_br[j]->msg_alpha_ok + s_br[j]->msg_alpha_err  +  s_br[j]->msg_numeric_ok + s_br[j]->msg_numeric_err)/(double)(s_all.msg_alpha_ok + s_all.msg_alpha_err + s_all.msg_numeric_ok + s_all.msg_numeric_err)*100);
        }
      }
      printf("\n");
    }



    if (s_all.cw_ok + s_all.cw_err)
    {
      printf("Decoder success rate: %.2f\%%\n",(double)s_all.cw_ok/(double)(s_all.cw_ok + s_all.cw_err)*100);
    }

    if (session_stats->s.pcm_mbyte)
    {
      printf("PCM data processed:   %d MB\n",session_stats->s.pcm_mbyte);
    }
    printf("\n");
  }

  /******************/

  if (total_stats != NULL)
  {
    p = &total_stats->s.first_run;
    pthread_mutex_lock(&time_mutex);
    tss = (time_t)(*p & 0xFFFFFFFF);
    now = localtime((time_t *)&tss);
    strftime(date_str,49,date_format_long,now);
    pthread_mutex_unlock(&time_mutex);
    tmp = sec2time(total_stats->s.time_running);

    printf("Cumulative statistics\n");
    printf("---------------------\n");
    printf("First launched:       %s\n",date_str);
    printf("Time running:         %s\n",tmp);
    nfree(tmp);
    tmp = NULL;
    for(i=0;i<t_ch_cnt;i++)
    {
      if (t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err + t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err)
      {
        printf("Channel:              %d\n",t_ch_int[i]);

        printf("Messages:             %d\n",(t_ch[i]->msg_alpha_ok   +
                                             t_ch[i]->msg_alpha_err  +
                                             t_ch[i]->msg_numeric_ok +
                                             t_ch[i]->msg_numeric_err));

        if (t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err)
        {
          printf("                      %d%*s  Alpha      (%.2f\%%)\n",(t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err),
                                                                       len_diff(t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err,t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err + t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err),
                                                                       "",
                                                                       ((double)(t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err)/(double)(t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err + t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err))*100);
        }

        if (t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err)
        {
          printf("                      %d%*s  Numeric    (%.2f\%%)\n",(t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err),
                                                                       len_diff(t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err,t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err + t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err),
                                                                       "",
                                                                       ((double)(t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err)/(double)(t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err + t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err))*100);
        }

        for (j=0;j<t_br_cnt;j++)
        {
          for (k=0;k<total_stats->ch_br_count;k++)
          {
            if ((total_stats->ch_br[k]->channel == t_ch_int[i]) &&  (total_stats->ch_br[k]->bitrate == t_br_int[j]))
            {
              if (total_stats->ch_br[k]->msg_alpha_ok + total_stats->ch_br[k]->msg_alpha_err + total_stats->ch_br[k]->msg_numeric_ok + total_stats->ch_br[k]->msg_numeric_err)
              {
                printf("                      %d%*s  @ %4d bps (%.2f\%%)\n",(total_stats->ch_br[k]->msg_alpha_ok   +
                                                                             total_stats->ch_br[k]->msg_alpha_err  +
                                                                             total_stats->ch_br[k]->msg_numeric_ok +
                                                                             total_stats->ch_br[k]->msg_numeric_err),
                                                                            len_diff(total_stats->ch_br[k]->msg_alpha_ok   +
                                                                                     total_stats->ch_br[k]->msg_alpha_err  +
                                                                                     total_stats->ch_br[k]->msg_numeric_ok +
                                                                                     total_stats->ch_br[k]->msg_numeric_err,
                                                                                     t_ch[i]->msg_alpha_ok   +
                                                                                     t_ch[i]->msg_alpha_err  +
                                                                                     t_ch[i]->msg_numeric_ok +
                                                                                     t_ch[i]->msg_numeric_err),
                                                                            "",
                                                                            t_br_int[j],
                                                                            (double)(total_stats->ch_br[k]->msg_alpha_ok + total_stats->ch_br[k]->msg_alpha_err  +  total_stats->ch_br[k]->msg_numeric_ok + total_stats->ch_br[k]->msg_numeric_err)/(double)(t_ch[i]->msg_alpha_ok + t_ch[i]->msg_alpha_err + t_ch[i]->msg_numeric_ok + t_ch[i]->msg_numeric_err)*100);
              }
            }
          }
        }
        printf("\n");
      }
    }

   /******************/

    if (t_all.msg_alpha_ok + t_all.msg_alpha_err + t_all.msg_numeric_ok + t_all.msg_numeric_err)
    {

      printf("Total messages:       %d\n",(t_all.msg_alpha_ok  +
                                           t_all.msg_alpha_err  +
                                           t_all.msg_numeric_ok +
                                           t_all.msg_numeric_err));

      if (t_all.msg_alpha_ok + t_all.msg_alpha_err)
      {
        printf("                      %d%*s  Alpha      (%.2f\%%)\n",(t_all.msg_alpha_ok + t_all.msg_alpha_err),
                                                                     len_diff(t_all.msg_alpha_ok + t_all.msg_alpha_err,t_all.msg_alpha_ok + t_all.msg_alpha_err + t_all.msg_numeric_ok + t_all.msg_numeric_err),
                                                                     "",
                                                                     ((double)(t_all.msg_alpha_ok + t_all.msg_alpha_err)/(double)(t_all.msg_alpha_ok + t_all.msg_alpha_err + t_all.msg_numeric_ok + t_all.msg_numeric_err))*100);
      }

      if (t_all.msg_numeric_ok + t_all.msg_numeric_err)
      {
        printf("                      %d%*s  Numeric    (%.2f\%%)\n",(t_all.msg_numeric_ok + t_all.msg_numeric_err),
                                                                     len_diff(t_all.msg_numeric_ok + t_all.msg_numeric_err,t_all.msg_alpha_ok + t_all.msg_alpha_err + t_all.msg_numeric_ok + t_all.msg_numeric_err),
                                                                     "",
                                                                     ((double)(t_all.msg_numeric_ok + t_all.msg_numeric_err)/(double)(t_all.msg_alpha_ok + t_all.msg_alpha_err + t_all.msg_numeric_ok + t_all.msg_numeric_err))*100);
      }

      for (j=0;j<t_br_cnt;j++)
      {
        if (t_br[j]->msg_alpha_ok + t_br[j]->msg_alpha_err +  t_br[j]->msg_numeric_ok + t_br[j]->msg_numeric_err)
        {
          printf("                      %d%*s  @ %4d bps (%.2f\%%)\n",(t_br[j]->msg_alpha_ok   +
                                                                       t_br[j]->msg_alpha_err  +
                                                                       t_br[j]->msg_numeric_ok +
                                                                       t_br[j]->msg_numeric_err),
                                                                      len_diff(t_br[j]->msg_alpha_ok   +
                                                                               t_br[j]->msg_alpha_err  +
                                                                               t_br[j]->msg_numeric_ok +
                                                                               t_br[j]->msg_numeric_err,
                                                                               t_all.msg_alpha_ok   +
                                                                               t_all.msg_alpha_err  +
                                                                               t_all.msg_numeric_ok +
                                                                               t_all.msg_numeric_err),
                                                                      "",
                                                                      t_br_int[j],
                                                                     (double)(t_br[j]->msg_alpha_ok + t_br[j]->msg_alpha_err  +  t_br[j]->msg_numeric_ok + t_br[j]->msg_numeric_err)/(double)(t_all.msg_alpha_ok + t_all.msg_alpha_err + t_all.msg_numeric_ok + t_all.msg_numeric_err)*100);
        }
      }
      printf("\n");
    }



    if (t_all.cw_ok + t_all.cw_err)
    {
      printf("Decoder success rate: %.2f\%%\n",(double)t_all.cw_ok/(double)(t_all.cw_ok + t_all.cw_err)*100);
    }

    if (total_stats->s.pcm_mbyte)
    {
      printf("PCM data processed:   %d MB\n",total_stats->s.pcm_mbyte);
    }
    printf("\n");
   /*****************/
  }

  if (s_ch_int != NULL)
  {
    nfree(s_ch_int);
    s_ch_int = NULL;
  }

  if (s_br_int != NULL)
  {
    nfree(s_br_int);
    s_br_int = NULL;
  }


  if (t_ch_int != NULL)
  {
    nfree(t_ch_int);
    t_ch_int = NULL;
  }

  if (t_br_int != NULL)
  {
    nfree(t_br_int);
    t_br_int = NULL;
  }

  for (i=0;i<s_ch_cnt;i++)
  {
    nfree(s_ch[i]);
    s_ch[i] = NULL;
  }

  if (s_ch != NULL)
  {
    nfree(s_ch);
    s_ch = NULL;
  }

  for (i=0;i<s_br_cnt;i++)
  {
    nfree(s_br[i]);
    s_br[i] = NULL;
  }

  if (s_br != NULL)
  {
    nfree(s_br);
    s_br = NULL;
  }

  for (i=0;i<t_ch_cnt;i++)
  {
    nfree(t_ch[i]);
    t_ch[i] = NULL;
  }

  if (t_ch != NULL)
  {
    nfree(t_ch);
    t_ch = NULL;
  }

  for (i=0;i<t_br_cnt;i++)
  {
    nfree(t_br[i]);
    t_br[i] = NULL;
  }

  if (t_br != NULL)
  {
    nfree(t_br);
    t_br = NULL;
  }

  #ifdef DEBUG
  dump_mem_list(&list);
  #endif

  nfree(date_str);
  date_str = NULL;
}


char *sec2time(uint32_t tid)
{
  int32_t len,years,days,hours,minutes,seconds;
  int32_t t;

  char *tmp,*years_str,*days_str,*hours_str,*minutes_str,*seconds_str;

  years_str   = NULL;
  days_str    = NULL;
  hours_str   = NULL;
  minutes_str = NULL;
  seconds_str = NULL;
  tmp         = NULL;
  len = 0;
  t = tid;

  years   = 0;
  days    = 0;
  hours   = 0;
  minutes = 0;
  seconds = 0;

  years = (t / (24*3600*365));
  t     = t - (years*(24*3600*365));

  days  = (t / (24*3600));
  t     = t -  (days*(24*3600));

  hours = (t / 3600);
  t     = t -  (hours*(3600));

  minutes = (t / 60);
  t     = t -  (minutes*(60));

  seconds = t;

  if (years == 1)
  {
    asprintf(&years_str,"%d year ",years);
  } else if (years > 1)
  {
    asprintf(&years_str,"%d years ",years);
  }

  if (years != 0)
  {
    len += strlen(years_str);
  }

  if (days == 1)
  {
    asprintf(&days_str,"%d day ",days);
  } else if (days > 1)
  {
    asprintf(&days_str,"%d days ",days);
  }

  if (days != 0)
  {
    len += strlen(days_str);
  }

  if (hours == 1)
  {
    asprintf(&hours_str,"%d hour ",hours);
  } else if (hours > 1)
  {
    asprintf(&hours_str,"%d hours ",hours);
  }

  if (hours != 0)
  {
    len += strlen(hours_str);
  }

  if (minutes == 1)
  {
    asprintf(&minutes_str,"%d minute ",minutes);
  } else if (minutes > 1)
  {
    asprintf(&minutes_str,"%d minutes ",minutes);
  }

  if (minutes != 0)
  {
    len += strlen(minutes_str);
  }

  if (seconds == 1)
  {
    asprintf(&seconds_str,"%d second",seconds);
  } else if ((seconds > 1) || (seconds == 0))
  {
    asprintf(&seconds_str,"%d seconds",seconds);
  }

  len += strlen(seconds_str);

  tmp = nmalloc(len + 1);
  memset(tmp,0,len+1);

  if (years_str)
  {
    strcat(tmp,years_str);
    nfree(years_str);
  }

  if (days_str)
  {
    strcat(tmp,days_str);
    nfree(days_str);
  }

  if (hours_str)
  {
    strcat(tmp,hours_str);
    nfree(hours_str);
  }

  if (minutes_str)
  {
    strcat(tmp,minutes_str);
    nfree(minutes_str);
  }

  if (seconds_str)
  {
    strcat(tmp,seconds_str);
    nfree(seconds_str);
  }

  if (tmp[strlen(tmp) - 1] == 0x20)
  {
    tmp[strlen(tmp) - 1] = 0;
  }

  return tmp;
}

int32_t len_diff(int32_t a,int32_t b)
{
  return abs((int32_t)log10((double)(a ? 10*a : 10)) - (int32_t)log10((double)(b ? 10*b : 10)));
}

void server(void *params) /* Handles connected PCM and RAW clients. */
{
  void **tmp_v;
  char *q,*zz,*header_string_new;
  int32_t tmp_i,i,z;

  struct server_parameters *parameters = (struct server_parameters *)params;

  struct pointer_queue queue;
  struct pcm_raw_chunk *ab;
  ab = NULL;
  memset(&queue,0,sizeof(struct pointer_queue));

  tmp_i = 0;
  tmp_v = nmalloc(sizeof(void *));

  while (1)
  {
    pthread_mutex_lock(parameters->server_mutex);
    pthread_cond_wait(parameters->server_cond,parameters->server_mutex);
    while (parameters->queue->count>0)
    {
      nenqueue(&queue,ndequeue(parameters->queue,&tmp_i,tmp_v));
    }
    pthread_mutex_unlock(parameters->server_mutex);
    while(queue.count>0)
    {
      ab = ndequeue(&queue,&tmp_i,tmp_v);
      if (ab)
      { pthread_mutex_lock(parameters->server_mutex2);
        header_string_new = (parameters->build_header_string)(ab);
        pthread_mutex_unlock(parameters->server_mutex2);
        pthread_mutex_lock(parameters->server_mutex2);
        if (*parameters->server_writable)
        {
          for (i=0;i<(*parameters->server_writable)->count;i++)
          {
            if (strcmp(header_string_new,(*parameters->server_writable)->items[i].header_string) != 0)
            {
              q = sec2time(time(NULL) - (*parameters->server_writable)->items[i].tconn);
              asprintf(&zz,"%s server: Client %s port %d has been disconnected because a new sample format was detected from the input. Time connected: %s,",parameters->id_string,(*parameters->server_writable)->items[i].str,ntohs((*parameters->server_writable)->items[i].sin_port),q);
              print_log(MSG_NOTICE,zz);
              if ((*parameters->server_writable)->items[i].io_r)
              {
                fclose((*parameters->server_writable)->items[i].io_r);
                (*parameters->server_writable)->items[i].io_r = NULL;
              }
              if ((*parameters->server_writable)->items[i].io_w)
              {
                fflush((*parameters->server_writable)->items[i].io_w);
                fclose((*parameters->server_writable)->items[i].io_w);
                (*parameters->server_writable)->items[i].io_w = NULL;
              }
              shutdown((*parameters->server_writable)->items[i].num,0);
              shutdown((*parameters->server_writable)->items[i].socket_dup,0);
              close((*parameters->server_writable)->items[i].num);
              close((*parameters->server_writable)->items[i].socket_dup);
              nfree(zz);
              nfree(q);
              if ((*parameters->server_writable)->items[i].str)
              {
                nfree((*parameters->server_writable)->items[i].str);
                (*parameters->server_writable)->items[i].str = NULL;
              }

              if ((*parameters->server_writable)->items[i].header_string)
              {
                nfree((*parameters->server_writable)->items[i].header_string);
                (*parameters->server_writable)->items[i].header_string = NULL;
              }
              (*parameters->connections)--;
              *parameters->server_writable = int_remove_num(*parameters->server_writable,(*parameters->server_writable)->items[i].num);
            }
          }
        }
        nfree(header_string_new);
        pthread_mutex_unlock(parameters->server_mutex2);
      }

      pthread_mutex_lock(parameters->server_mutex2);

      if (*parameters->server_writable)
      {
        for (i=0;i<(*parameters->server_writable)->count;i++)
        {
          z = write((*parameters->server_writable)->items[i].num,ab->ptr,ab->length);
          if (z>0)
          {
            int_add_bytes(*parameters->server_writable,(*parameters->server_writable)->items[i].num,z);
          }

          if (z!=ab->length)
          {
            q = sec2time(time(NULL) - (*parameters->server_writable)->items[i].tconn);
            asprintf(&zz,"%s server: Connection to %s port %d lost. Time connected: %s\n",parameters->id_string,(*parameters->server_writable)->items[i].str,ntohs((*parameters->server_writable)->items[i].sin_port),q);
            print_log(MSG_NOTICE,zz);
            if ((*parameters->server_writable)->items[i].io_r)
            {
              fclose((*parameters->server_writable)->items[i].io_r);
              (*parameters->server_writable)->items[i].io_r = NULL;
            }

            if ((*parameters->server_writable)->items[i].io_w)
            {
              fflush((*parameters->server_writable)->items[i].io_w);
              fclose((*parameters->server_writable)->items[i].io_w);
              (*parameters->server_writable)->items[i].io_w = NULL;
            }
            shutdown((*parameters->server_writable)->items[i].num,0);
            shutdown((*parameters->server_writable)->items[i].socket_dup,0);
            close((*parameters->server_writable)->items[i].num);
            close((*parameters->server_writable)->items[i].socket_dup);
            nfree(zz);
            nfree(q);
            if ((*parameters->server_writable)->items[i].str)
            {
              nfree((*parameters->server_writable)->items[i].str);
              (*parameters->server_writable)->items[i].str = NULL;
            }
            (*parameters->connections)--;
            *parameters->server_writable = int_remove_num(*parameters->server_writable,(*parameters->server_writable)->items[i].num);
          }
        }
      }
      pthread_mutex_unlock(parameters->server_mutex2);

      if (ab->ptr)
      {
        nfree(ab->ptr);
        ab->ptr = NULL;
      }
      if (ab)
      {
        nfree(ab);
        ab = NULL;
      }
    }
  }
}

void server_listen(void *params) /* Handles incoming connections for the RAW and PCM servers. */
{
  char *zz;
  int32_t len,i,s,z,j,ok,msocket,optval,optlen,connections;
  struct sockaddr_in srvr,clnt;
  memset(&srvr,0,sizeof(struct sockaddr_in));
  char *p,*client;
  p = NULL;
  FILE *io;
  fd_set r;
  struct timeval t_out;
  struct server_parameters *parameters = (struct server_parameters *)params;
  struct thread_slot server_authentication_thread_slots[parameters->server_limit];

  *(parameters->server_writable) = NULL;
  *(parameters->server_readable) = NULL;
  *(parameters->server_writable_new) = NULL;
  *(parameters->server_readable_new) = NULL;
  *(parameters->server_new_connections) = NULL;

  memset(&server_authentication_thread_slots, 0, parameters->server_limit*sizeof(struct thread_slot));

  FD_ZERO(&r);

  ok = 0;
  optval = 1;
  optlen = sizeof(int32_t);

  s = socket(PF_INET,SOCK_STREAM,0);
  z = setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&optval,optlen);
  pthread_mutex_lock(parameters->server_mutex2);
  if (s != -1)
  {
    srvr.sin_family      = AF_INET;
    srvr.sin_port        = htons(parameters->server_port);
    srvr.sin_addr.s_addr = INADDR_ANY;
    z = bind(s,(struct sockaddr *)&srvr,sizeof(struct sockaddr_in));
    if (z != -1)
    {
      z = listen(s,10);
      if (z != -1)
      {
        msocket = s;
        *(parameters->server_readable) = int_add(*(parameters->server_readable),s,0,&srvr.sin_addr,srvr.sin_port,0,NULL,NULL,NULL, NULL);
        FD_SET(s,&r);
        ok = 1;

      } else
      {
        asprintf(&zz,"listen: %s\n",strerror(errno));
        print_log(MSG_ERROR,zz);
        nfree(zz);
      }
    } else
    {
      asprintf(&zz,"bind: %s\n",strerror(errno));
      print_log(MSG_ERROR,zz);
      nfree(zz);
    }

  } else
  {
    asprintf(&zz,"socket: %s\n",strerror(errno));
    print_log(MSG_ERROR,zz);
    nfree(zz);
  }

  pthread_mutex_unlock(parameters->server_mutex2);

  if (ok)
  {
    while (1)
    {
      pthread_mutex_lock(parameters->server_auth);
      if (*(parameters->server_new_connections))
      {
        if ((*parameters->server_new_connections)->count>0)
        {
          for (i=0;i<(*parameters->server_new_connections)->count;i++)
          {
            *parameters->server_writable = int_add(*parameters->server_writable,(*parameters->server_new_connections)->items[i].num,(*parameters->server_new_connections)->items[i].socket_dup,&(*parameters->server_new_connections)->items[i].sin_addr,(*parameters->server_new_connections)->items[i].sin_port,(*parameters->server_new_connections)->items[i].tconn,(*parameters->server_new_connections)->items[i].io_r,(*parameters->server_new_connections)->items[i].io_w,(*parameters->server_new_connections)->items[i].str,(*parameters->server_new_connections)->items[i].header_string);
          }
          if ((*parameters->server_new_connections)->items)
          {
            nfree((*parameters->server_new_connections)->items);
            (*parameters->server_new_connections)->items = NULL;
          }
          nfree((*parameters->server_new_connections)->items);
          (*parameters->server_new_connections)->items = NULL;
          (*parameters->server_new_connections)->count = 0;
        }
      }
      pthread_mutex_unlock(parameters->server_auth);

      for (i=0;i<parameters->server_limit;i++)
      {
        if (((time(NULL) - server_authentication_thread_slots[i].started ) >= parameters->server_timeout) && server_authentication_thread_slots[i].inuse)
        {
          asprintf(&zz,"%s server: Client %s, port %d disconnected because authentication timeout has been reached.\n",parameters->id_string,server_authentication_thread_slots[i].client_str,ntohs(server_authentication_thread_slots[i].sin_port));
          pthread_cancel(server_authentication_thread_slots[i].thread);
          print_log(MSG_NOTICE,zz);
          nfree(zz);
        }
      }

      t_out.tv_sec  = 1;
      t_out.tv_usec = 0;
      z = select(FD_SETSIZE,&r,NULL,NULL,&t_out);
      pthread_mutex_lock(parameters->server_mutex2);

      if (z>0)
      {
        if (*parameters->server_readable_new)
        {
          if ((*parameters->server_readable_new)->items)
          {
            nfree((*parameters->server_readable_new)->items);
            (*parameters->server_readable_new)->items = NULL;
          }
          nfree(*parameters->server_readable_new);
          *parameters->server_readable_new = NULL;
        }

        if (*parameters->server_readable)
        {
          for (i=0;i<(*parameters->server_readable)->count;i++)
          {
            if (FD_ISSET((*parameters->server_readable)->items[i].num,&r))
            {
              *parameters->server_readable_new = int_add(*parameters->server_readable_new,(*parameters->server_readable)->items[i].num,0,&(*parameters->server_readable)->items[i].sin_addr,(*parameters->server_readable)->items[i].sin_port,(*parameters->server_readable)->items[i].tconn,NULL,NULL,NULL, NULL);
            }
          }
        }

        if (*parameters->server_readable_new)
        {
          for (i=0;i<(*parameters->server_readable_new)->count;i++)
          {
            if ((*parameters->server_readable_new)->items[i].num == msocket)
            {
              len = sizeof(struct sockaddr_in);
              memset(&clnt,0,sizeof(struct sockaddr_in));
              z = accept(msocket,(struct sockaddr *)&clnt,(uint32_t*)&len);
              if (z != -1)
              {
                client = gethostaddr_rr((struct in_addr *)&clnt.sin_addr.s_addr);
                if (client)
                {
                  asprintf(&p,"%s",client);
                } else
                {
                  asprintf(&p,"%s",inet_ntoa(clnt.sin_addr));
                }
                asprintf(&zz,"%s server: Connection from %s, port %d\n",parameters->id_string,p,ntohs(clnt.sin_port));
                print_log(MSG_NOTICE,zz);
                nfree(zz);
                nfree(client);

                if (*parameters->server_writable)
                {
                  connections = (*parameters->server_writable)->count;
                } else
                {
                  connections = 0;
                }

                if ((*parameters->header_string) == NULL)
                {
                  asprintf(&zz,"%s server: Client %s, port %d rejected because header string is unavailable.\n",parameters->id_string,p,ntohs(clnt.sin_port));
                  print_log(MSG_NOTICE,zz);
                  nfree(zz);
                  io = fdopen(z,"w+");
                  if (io)
                  {
                    fprintf(io,"TRYLATER\n");
                    fflush(io);
                    fclose(io);
                  } else
                  {
                    asprintf(&zz,"fdopen: %s\n",strerror(errno));
                    print_log(MSG_ERROR,zz);
                    nfree(zz);
                  }
                  nfree(p);
                  shutdown(z,0);
                  close(z);
                  goto rr;
                }

                if (*(parameters->connections) < parameters->server_limit)
                {
                  for (j=0;j<parameters->server_limit;j++)
                  {
                    if (server_authentication_thread_slots[j].inuse == 0)
                    {
                      server_authentication_thread_slots[j].socket_dup = dup(z);
                      server_authentication_thread_slots[j].io_r       = fdopen(z,"r+");
                      server_authentication_thread_slots[j].io_w       = fdopen(server_authentication_thread_slots[j].socket_dup,"w+");

                      if (server_authentication_thread_slots[j].socket_dup == -1)
                      {
                        asprintf(&zz,"%s server: Couldn't duplicate file descriptor (dup() == -1), can't handle incoming connection\n",parameters->id_string);
                        print_log(MSG_ERROR,zz);
                        nfree(zz);
                      }

                      if ((server_authentication_thread_slots[j].io_r != NULL) && (server_authentication_thread_slots[j].io_w != NULL))
                      {
                        server_authentication_thread_slots[j].samplerate          = pcm_server_samplerate;
                        server_authentication_thread_slots[j].bits                = pcm_server_bits;
                        server_authentication_thread_slots[j].channels            = pcm_server_channels;
                        server_authentication_thread_slots[j].msg_connected       = malloc(strlen(*parameters->header_string) + 1);
                        memcpy(server_authentication_thread_slots[j].msg_connected,*parameters->header_string,strlen(*parameters->header_string)+1);
                        server_authentication_thread_slots[j].client_disconnected = 0;
                        server_authentication_thread_slots[j].client_str          = p;
                        server_authentication_thread_slots[j].sin_addr            = clnt.sin_addr;
                        server_authentication_thread_slots[j].sin_port            = clnt.sin_port;
                        server_authentication_thread_slots[j].inuse               = 1;
                        server_authentication_thread_slots[j].started             = time(NULL);
                        server_authentication_thread_slots[j].socket              = z;
                        server_authentication_thread_slots[j].password            = parameters->password;
                        server_authentication_thread_slots[j].id_string            = parameters->id_string;
                        server_authentication_thread_slots[j].numlist_update_mutex = parameters->server_auth;;
                        server_authentication_thread_slots[j].connections          = parameters->connections;

                        pthread_mutex_lock(parameters->server_auth);
                        if ((*parameters->server_new_connections) == NULL)
                        {
                          (*parameters->server_new_connections) = nmalloc(sizeof(struct num_list));
                          memset((*parameters->server_new_connections),0,sizeof(struct num_list));
                        }
                        pthread_mutex_unlock(parameters->server_auth);
                        server_authentication_thread_slots[j].nl = parameters->server_new_connections;
                        pthread_attr_init(&server_authentication_thread_slots[j].thread_attr);
                        pthread_mutex_init(&server_authentication_thread_slots[j].thread_mutex,NULL);
                        pthread_attr_setdetachstate(&server_authentication_thread_slots[j].thread_attr,PTHREAD_CREATE_DETACHED);
                        pthread_create(&server_authentication_thread_slots[j].thread, &server_authentication_thread_slots[j].thread_attr, (void *)auth, (void *)&server_authentication_thread_slots[j]);
                        goto qq;
                      } else
                      {
                        asprintf(&zz,"fdopen: %s\n",strerror(errno));
                        print_log(MSG_ERROR,zz);
                        nfree(zz);
                        if (server_authentication_thread_slots[j].io_r)
                        {
                          fclose(server_authentication_thread_slots[j].io_r);
                          server_authentication_thread_slots[j].io_r = NULL;
                        }

                        if (server_authentication_thread_slots[j].io_w)
                        {
                          fclose(server_authentication_thread_slots[j].io_w);
                          server_authentication_thread_slots[j].io_w = NULL;
                        }
                        shutdown(server_authentication_thread_slots[j].socket_dup,0);
                        shutdown(z,0);
                        close(server_authentication_thread_slots[j].socket_dup);
                        close(z);
                        goto qq;
                      }
                    }
                  }
                  asprintf(&zz,"%s server: Client %s, port %d rejected because all authentication slots are busy.\n",parameters->id_string,p,ntohs(clnt.sin_port));
                  goto xxz;
                } else
                {
                  asprintf(&zz,"%s server: Client %s, port %d rejected because the maximum numbers of concurrent clients has been reached.\n",parameters->id_string,p,ntohs(clnt.sin_port));
                  xxz:;

                  print_log(MSG_NOTICE,zz);
                  nfree(zz);

                  io = fdopen(z,"w+");
                  if (io)
                  {
                    fprintf(io,"BUSY\n");
                    fflush(io);
                    fclose(io);
                  } else
                  {
                    asprintf(&zz,"fdopen: %s\n",strerror(errno));
                    print_log(MSG_ERROR,zz);
                    nfree(zz);
                  }
                  nfree(p);
                  shutdown(z,0);
                  close(z);
                }
                qq:;
              } else
              {
                asprintf(&zz,"accept: %s\n",strerror(errno));
                print_log(MSG_ERROR,zz);
                nfree(zz);
              }
            }
          }
        }
      }

      rr:;

      FD_ZERO(&r);
      if (*parameters->server_readable)
      {
        for (i=0;i<(*parameters->server_readable)->count;i++)
        {
          FD_SET((*parameters->server_readable)->items[i].num,&r);
        }
      }
      pthread_mutex_unlock(parameters->server_mutex2);
    }
  }
}


struct num_list *int_add(struct num_list *n, int32_t a, int32_t socket_dup, struct in_addr *sin_addr, unsigned short sin_port, int32_t tconn, FILE *io_r, FILE *io_w, char *str, char *header_string)
{
  struct num_list *tmp;
  if (n == NULL)
  {
    tmp = nmalloc(sizeof(struct num_list));
    memset(tmp,0,sizeof(struct num_list));
    tmp->items = nmalloc(sizeof(struct num_list_items));
    memset(tmp->items,0,sizeof(struct num_list_items));
    tmp->count++;
    tmp->items[tmp->count - 1].num = a;
    tmp->items[tmp->count - 1].socket_dup = socket_dup;
    tmp->items[tmp->count - 1].b  = 0;
    tmp->items[tmp->count - 1].mb = 0;
    tmp->items[tmp->count - 1].sin_addr      = *sin_addr;
    tmp->items[tmp->count - 1].sin_port      = sin_port;
    tmp->items[tmp->count - 1].tconn         = tconn;
    tmp->items[tmp->count - 1].io_r          = io_r;
    tmp->items[tmp->count - 1].io_w          = io_w;
    tmp->items[tmp->count - 1].str           = str;
    tmp->items[tmp->count - 1].header_string = header_string;
    return tmp;
  } else
  {
    tmp = n;
    if (tmp->items)
    {
      tmp->items = nrealloc(tmp->items,sizeof(struct num_list_items)*++tmp->count);
    } else
    {
      tmp->items = nmalloc(sizeof(struct num_list_items)*++tmp->count);
    }
    tmp->items[tmp->count - 1].num = a;
    tmp->items[tmp->count - 1].socket_dup    = socket_dup;
    tmp->items[tmp->count - 1].b  = 0;
    tmp->items[tmp->count - 1].mb = 0;
    tmp->items[tmp->count - 1].sin_addr      = *sin_addr;
    tmp->items[tmp->count - 1].sin_port      = sin_port;
    tmp->items[tmp->count - 1].tconn         = tconn;
    tmp->items[tmp->count - 1].io_r          = io_r;
    tmp->items[tmp->count - 1].io_w          = io_w;
    tmp->items[tmp->count - 1].str           = str;
    tmp->items[tmp->count - 1].header_string = header_string;
    return tmp;
  }
}

struct num_list *int_remove_num(struct num_list *n, int32_t num)
{
  struct num_list *tmp;
  tmp = NULL;
  int32_t i,j,pos;
  pos = 0;

  for (i=0;i<n->count;i++)
  {
    if (n->items[i].num == num)
    {
      pos = i;
    }
  }

  if (n)
  {
    if ((pos < n->count) && (n->count >= 1))
    {
      tmp = nmalloc(sizeof(struct num_list));
      memcpy(tmp,n,sizeof(struct num_list));

      tmp->count = (n->count - 1);

      if (tmp->count != 0)
      {
        tmp->items = nmalloc(sizeof(struct num_list_items)*tmp->count);
      } else
      {
        tmp->items = NULL;
      }

      j = 0;

      if (tmp->items)
      {
        for (i=0;i<n->count;i++)
        {
          if (i != pos)
          {
            tmp->items[j] = n->items[i];
            j++;
          }
        }
      }

      if (n)
      {
        if (n->items)
        {
          nfree(n->items);
          n->items = NULL;
        }
        nfree(n);
      }
      return tmp;
    } else
    {
      return n;
    }
  }
  return NULL;
}


void int_add_bytes(struct num_list *n, int32_t num,int32_t byte_count)
{
  int32_t i,pos;
  pos = 0;

  for (i=0;i<n->count;i++)
  {
    if (n->items[i].num == num)
    {
      pos = i;
    }
  }

  n->items[pos].b += byte_count;

  while (n->items[pos].b >= 1024*1024)
  {
    n->items[pos].b -= 1024*1024;
    n->items[pos].mb++;
  }
}


void telnet_listen(void *parameters) /* Handles incoming connections to the telnet server. */
{
  fd_set r,w;
  int32_t z,i,j,s,port,optval,optlen,len;
  char *t,*zz,*p,*client,*k;
  struct timeval t_out;
  struct sockaddr_in srvr,clnt;
  struct log_functions_config *lf;
  struct thread_slot telnet_server_authentication_thread_slots[TELNET_SERVER_LIMIT];
  int32_t    *id,id_cnt,found;

  id = NULL;
  id_cnt = 0;

  FILE *io;
  lf = (struct log_functions_config *)parameters;
  memset(&telnet_server_authentication_thread_slots,0,sizeof(struct thread_slot)*TELNET_SERVER_LIMIT);

  FD_ZERO(&r);
  FD_ZERO(&w);

  for (i=0;i<lf->rec.count;i++)
  {
    if (lf->rec.rcpt[i]->str[l_port])
    {
      t = nmalloc(strlen(lf->rec.rcpt[i]->str[l_port]) + 1);
      if (sscanf(lf->rec.rcpt[i]->str[l_port],"%d:%s",&port,t) == 2)
      {
        z = 0;
        found = 0;
        for (j=0;j<id_cnt;j++)
        {
          if (id[j] == lf->rec.rcpt[i]->num[_id])
          {
            found = 1;
          }
        }
        if (!found)
        {
          if (id == NULL)
          {
            id = nmalloc(sizeof(int32_t)*++id_cnt);
            id[id_cnt - 1] = lf->rec.rcpt[i]->num[_id];
          } else
          {
            id = nrealloc(id,sizeof(int32_t)*++id_cnt);
            id[id_cnt - 1] = lf->rec.rcpt[i]->num[_id];
          }
        }

        for (j=0;j<lf->nl_count;j++)
        { /* Make sure that each port is only used once. */
          if (lf->nl[j]->port == port)
          {
            asprintf(&k,"Telnet server: Port %d has already been allocated.\n",port);
            print_log(MSG_ERROR,k);
            nfree(k);
            z = 1;
          }
        }

        if (z==0)
        {
          optval = 1;
          optlen = sizeof(int32_t);
          s = socket(PF_INET,SOCK_STREAM,0);
          z = setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&optval,optlen);
          pthread_mutex_lock(&int_add_remove);
          if (s != -1)
          {
            srvr.sin_family      = AF_INET;
            srvr.sin_port        = htons(port);
            srvr.sin_addr.s_addr = INADDR_ANY;
            z = bind(s,(struct sockaddr *)&srvr,sizeof(struct sockaddr_in));
            if (z != -1)
            {
              z = listen(s,10);
              if (z != -1)
              {
                if (lf->nl)
                {
                  lf->nl = nrealloc(lf->nl,sizeof(struct num_list *)*++lf->nl_count);
                  lf->nl[lf->nl_count - 1] = nmalloc(sizeof(struct num_list));
                  memset(lf->nl[lf->nl_count - 1],0,sizeof(struct num_list));
                } else
                {
                  lf->nl = nmalloc(sizeof(struct num_list *)*++lf->nl_count);
                  lf->nl[lf->nl_count - 1] = nmalloc(sizeof(struct num_list));
                  memset(lf->nl[lf->nl_count - 1],0,sizeof(struct num_list));
                }
                lf->nl[lf->nl_count - 1]->password = t;
                lf->nl[lf->nl_count - 1]->id = lf->rec.rcpt[i]->num[_id];
                lf->nl[lf->nl_count - 1]->port = port;
                lf->nl[lf->nl_count - 1]->msocket = s;
                lf->nl[lf->nl_count - 1] = int_add(lf->nl[lf->nl_count - 1],s,0,&srvr.sin_addr,srvr.sin_port,0,NULL,NULL,NULL, NULL);
              } else
              {
                asprintf(&zz,"listen: %s\n",strerror(errno));
                print_log(MSG_ERROR,zz);
                nfree(zz);
              }
            } else
            {
              asprintf(&zz,"bind: %s\n",strerror(errno));
              print_log(MSG_ERROR,zz);
              nfree(zz);
            }

          } else
          {
            asprintf(&zz,"socket: %s\n",strerror(errno));
            print_log(MSG_ERROR,zz);
            nfree(zz);
          }
          pthread_mutex_unlock(&int_add_remove);
        }
      } else
      {
        asprintf(&zz,"Invalid port:password specification: %s\n",lf->rec.rcpt[i]->str[l_port]);
        print_log(MSG_ERROR,zz);
        nfree(zz);
      }
    }
  }

  for (i=0;i<lf->rec.count;i++)
  {
    if (lf->rec.rcpt[i]->str[l_port])
    {
      nfree(lf->rec.rcpt[i]->str[l_port]);
      lf->rec.rcpt[i]->str[l_port] = NULL;
    }
    nfree(lf->rec.rcpt[i]);
    lf->rec.rcpt[i] = NULL;
  }

  lf->rec.rcpt = nrealloc(lf->rec.rcpt,sizeof(struct recipients *)*id_cnt);
  for (i=0;i<id_cnt;i++)
  {
    lf->rec.rcpt[i] = nmalloc(sizeof(struct recipients));
    memset(lf->rec.rcpt[i],0,sizeof(struct recipients));
    lf->rec.rcpt[i]->num[_id] = id[i];
  }
  nfree(id);
  id = NULL;
  lf->rec.count = id_cnt;

  if (lf->nl_count>0)
  {
    while (1)
    {
      FD_ZERO(&r);
      for (i=0;i<lf->nl_count;i++)
      {
        FD_SET(lf->nl[i]->msocket, &r);
      }

      for (i=0;i<TELNET_SERVER_LIMIT;i++)
      {
        if (((time(NULL) - telnet_server_authentication_thread_slots[i].started ) >= TELNET_SERVER_TIMEOUT) && telnet_server_authentication_thread_slots[i].inuse)
        {
          asprintf(&zz,"Telnet server: Client %s, port %d disconnected because authentication timeout has been reached.\n",telnet_server_authentication_thread_slots[i].client_str,ntohs(telnet_server_authentication_thread_slots[i].sin_port));
          pthread_cancel(telnet_server_authentication_thread_slots[i].thread);
          print_log(MSG_NOTICE,zz);
          nfree(zz);
        }
      }

      t_out.tv_sec  = 1;
      t_out.tv_usec = 0;
      z = select(FD_SETSIZE,&r,NULL,NULL,&t_out);

      if (z>0)
      {
        for (i=0;i<lf->nl_count;i++)
        {
          if (FD_ISSET(lf->nl[i]->msocket,&r))
          {
            len = sizeof(struct sockaddr_in);
            memset(&clnt,0,sizeof(struct sockaddr_in));
            z = accept(lf->nl[i]->msocket,(struct sockaddr *)&clnt,(uint32_t*)&len);
            if (z != -1)
            {
              client = gethostaddr_rr((struct in_addr *)&clnt.sin_addr.s_addr);
              if (client)
              {
                asprintf(&p,"%s",client);
              } else
              {
                asprintf(&p,"%s",inet_ntoa(clnt.sin_addr));
              }
              asprintf(&zz,"Telnet server: Connection from %s, port %d.\n",p,ntohs(clnt.sin_port));
              print_log(MSG_NOTICE,zz);
              nfree(zz);
              nfree(client);

              if (telnet_connections<TELNET_SERVER_LIMIT)
              {
                for (j=0;j<TELNET_SERVER_LIMIT;j++)
                {
                  if (telnet_server_authentication_thread_slots[j].inuse == 0)
                  {
                    telnet_server_authentication_thread_slots[j].socket_dup = dup(z);
                    telnet_server_authentication_thread_slots[j].io_r       = fdopen(z,"r+");
                    telnet_server_authentication_thread_slots[j].io_w       = fdopen(telnet_server_authentication_thread_slots[j].socket_dup,"w+");

                    if (telnet_server_authentication_thread_slots[j].socket_dup == -1)
                    {
                      asprintf(&zz,"Telnet server: Couldn't duplicate file descriptor (dup() == -1), can't handle incoming connection\n");
                      print_log(MSG_ERROR,zz);
                      nfree(zz);
                    }

                    if ((telnet_server_authentication_thread_slots[j].io_r != NULL) && (telnet_server_authentication_thread_slots[j].io_w != NULL))
                    {
                      telnet_server_authentication_thread_slots[j].nl          = &lf->nl[i];
                      telnet_server_authentication_thread_slots[j].slot_no     = j;
                      telnet_server_authentication_thread_slots[j].password    = lf->nl[i]->password;
                      telnet_server_authentication_thread_slots[j].pw_temp     = nmalloc(50);
                      telnet_server_authentication_thread_slots[j].client_port = ntohs(clnt.sin_port);
                      telnet_server_authentication_thread_slots[j].client_str  = p;
                      telnet_server_authentication_thread_slots[j].sin_addr    = clnt.sin_addr;
                      telnet_server_authentication_thread_slots[j].sin_port    = clnt.sin_port;
                      telnet_server_authentication_thread_slots[j].inuse       = 1;
                      telnet_server_authentication_thread_slots[j].started     = time(NULL);
                      telnet_server_authentication_thread_slots[j].socket      = z;
                      pthread_attr_init(&telnet_server_authentication_thread_slots[j].thread_attr);
                      pthread_attr_setdetachstate(&telnet_server_authentication_thread_slots[j].thread_attr,PTHREAD_CREATE_DETACHED);
                      pthread_mutex_init(&telnet_server_authentication_thread_slots[j].thread_mutex,NULL);
                      pthread_create(&telnet_server_authentication_thread_slots[j].thread, &telnet_server_authentication_thread_slots[j].thread_attr, (void *)telnet_auth, (void *)&telnet_server_authentication_thread_slots[j]);
                      goto q;
                    } else
                    {
                      asprintf(&zz,"fdopen: %s\n",strerror(errno));
                      print_log(MSG_ERROR,zz);
                      nfree(zz);

                      if (telnet_server_authentication_thread_slots[j].io_r)
                      {
                        fclose(telnet_server_authentication_thread_slots[j].io_r);
                        telnet_server_authentication_thread_slots[j].io_r = NULL;
                      }

                      if (telnet_server_authentication_thread_slots[j].io_w)
                      {
                        fclose(telnet_server_authentication_thread_slots[j].io_w);
                        telnet_server_authentication_thread_slots[j].io_w = NULL;
                      }
                      shutdown(telnet_server_authentication_thread_slots[j].socket_dup,0);
                      shutdown(z,0);
                      close(telnet_server_authentication_thread_slots[j].socket_dup);
                      close(z);
                      goto q;
                    }
                  }
                }
                asprintf(&zz,"Telnet server: Client %s, port %d rejected because all authentication slots are busy.\n",p,ntohs(clnt.sin_port));
                goto xz;
              } else
              {
                asprintf(&zz,"Telnet server: Client %s, port %d rejected because the maximum numbers of concurrent clients has been reached.\n",p,ntohs(clnt.sin_port));
                xz:;

                print_log(MSG_NOTICE,zz);
                nfree(zz);
                io = fdopen(z,"w+");
                if (io)
                {
                  fprintf(io,"Telnet server is busy\r\n");
                  fflush(io);
                  fclose(io);
                  io = NULL;
                } else
                {
                  asprintf(&zz,"fdopen: %s\n",strerror(errno));
                  print_log(MSG_ERROR,zz);
                  nfree(zz);
                }
                nfree(p);
                shutdown(z,0);
                close(z);
                q:;
              }
            } else
            {
              asprintf(&zz,"accept: %s\n",strerror(errno));
              print_log(MSG_ERROR,zz);
              nfree(zz);
            }
          }
        }
      }
    }
  }
}

void telnet_auth(void *parameters)
{
  struct thread_slot *t_slot = (struct thread_slot *)parameters;

  char *z,*filter_id;
  int32_t cnt,auth,i;
  char *ret;
  char filter__id[20];

  pthread_cleanup_push(&telnet_cleanup, (void*)t_slot);
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

  memset(&filter__id,0,20);

  if ((*t_slot->nl)->id == -1)
  {
    asprintf(&filter_id,"%s","default");
  } else
  {
    asprintf(&filter_id,"%d",(*t_slot->nl)->id);
  }

  memcpy(&filter__id,filter_id,strlen(filter_id));
  nfree(filter_id);

  cnt = 0;
  auth = 0;
  fprintf(t_slot->io_w,"%s - Telnet Interface\r\n",program_string);
  fflush(t_slot->io_w);
  if (strcmp(t_slot->password,"*") != 0)
  {
    while (!auth)
    {
      fprintf(t_slot->io_w,"Password: ");
      fflush(t_slot->io_w);
      t_slot->cleanup_sockets = 1;
      pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
      ret = fgets(t_slot->pw_temp,48,t_slot->io_r);
      pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
      t_slot->cleanup_sockets = 0;
      if (ret == 0)
      {
        goto do_cleanup;
      }
      for (i=0;i<strlen(t_slot->pw_temp);i++)
      {
        if ((t_slot->pw_temp[i] == 10) || (t_slot->pw_temp[i] == 13))
        {
          t_slot->pw_temp[i] = 0;
        }
      }

      if (strcmp(t_slot->pw_temp,t_slot->password) == 0)
      {
        auth = 1;
      } else
      {
        cnt++;
        fprintf(t_slot->io_w,"\r\nIncorrect password\r\n");
        fflush(t_slot->io_w);
        asprintf(&z,"Telnet server (Filter ID: %s): Client %s, port %d failed authentication.\n",filter__id,t_slot->client_str,ntohs(t_slot->sin_port));
        print_log(MSG_NOTICE,z);
        nfree(z);
        if (cnt==3)
        {
          fprintf(t_slot->io_w,"Maximum number of tries exceeded\r\n");
          fflush(t_slot->io_w);
          goto do_cleanup;
        }
      }
    }
  } else
  {
    auth = 1;
  }

  if (auth)
  {
    if (strcmp(t_slot->password,"*") != 0)
    {
      asprintf(&z,"Telnet server (Filter ID: %s): Client %s, port %d successfully authenticated.\n",filter__id,t_slot->client_str,ntohs(t_slot->sin_port));
      print_log(MSG_NOTICE,z);
      nfree(z);
      fprintf(t_slot->io_w,"\r\nPassword accepted\r\nFilter ID: %s\r\n",filter__id);
      fflush(t_slot->io_w);
    } else
    {
      asprintf(&z,"Telnet server (Filter ID: %s): Client %s, port %d connected.\n",filter__id,t_slot->client_str,ntohs(t_slot->sin_port));
      print_log(MSG_NOTICE,z);
      nfree(z);
      fprintf(t_slot->io_w,"\r\nConnected - No Password Required\r\nFilter ID: %s\r\n",filter__id);
      fflush(t_slot->io_w);
    }
    fflush(t_slot->io_w);
    pthread_mutex_lock(&telnet_server_mutex);
    pthread_mutex_lock(&int_add_remove);
    *t_slot->nl = int_add(*t_slot->nl,t_slot->socket,t_slot->socket_dup,&t_slot->sin_addr,t_slot->sin_port,time(NULL),t_slot->io_r,t_slot->io_w,t_slot->client_str, NULL);
    t_slot->client_str = NULL;
    pthread_mutex_unlock(&int_add_remove);
    telnet_connections++;
    pthread_cond_signal(&telnet_server_cond);
    pthread_mutex_unlock(&telnet_server_mutex);
    goto no_cleanup;
  }
  no_cleanup:;
  t_slot->cleanup_sockets = 0;
  goto doit;
  do_cleanup:;
  t_slot->cleanup_sockets = 1;
  doit:;
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_exit((void*)0);
  pthread_cleanup_pop(0);
}

void telnet_cleanup(void *r)
{
  struct thread_slot *t_slot = (struct thread_slot *)r;

  if (t_slot->cleanup_sockets)
  {
    if (t_slot->client_str)
    {
      nfree(t_slot->client_str);
      t_slot->client_str = NULL;
    }

    if (t_slot->io_r)
    {
      fclose(t_slot->io_r);
      t_slot->io_r = NULL;
    }

    if (t_slot->io_w)
    {
      fclose(t_slot->io_w);
      t_slot->io_w = NULL;
    }

    if (t_slot->pw_temp)
    {
      nfree(t_slot->pw_temp);
      t_slot->pw_temp = NULL;
    }

    shutdown(t_slot->socket_dup,0);
    close(t_slot->socket_dup);
    shutdown(t_slot->socket,0);
    close(t_slot->socket);
  }
  t_slot->inuse = 0;
}

void auth(void *parameters) /* Handles authentication for the PCM and the RAW servers. */
{
  struct thread_slot *t_slot = (struct thread_slot *)parameters;

  int32_t cnt,auth,i,s;
  char *z,*tmp_str;
  char input[60];
  char tmp[60];
  char resp[100];
  char chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
  char *expected_response;

  pthread_cleanup_push(&auth_cleanup, (void*)t_slot);
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

  expected_response = NULL;
  memset(&resp,0,100);

  cnt = 0;
  auth = 0;

  s = time(NULL);
  srand(s);
  memset(&tmp,0,60);
  for (i=0;i<20;i++)
  {
    tmp[i] = chars[(rand() % strlen(chars))];
  }

  if (t_slot->password != NULL)
  {
    expected_response = build_sha1_response_string(tmp, t_slot->password);
    memcpy(resp,expected_response,strlen(expected_response));
    nfree(expected_response);
    expected_response = NULL;
  }

  memset(input,0,60);

  if (t_slot->password != NULL)
  {
    fprintf(t_slot->io_w,"AUTHREQ:%s\n",tmp);
    fflush(t_slot->io_w);
    asprintf(&z,"%s server: Sent challenge to %s, port %d\n",t_slot->id_string,t_slot->client_str,ntohs(t_slot->sin_port));
    print_log(MSG_NOTICE,z);
    nfree(z);
    t_slot->cleanup_sockets = 1;
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    fgets(input,58,t_slot->io_r);
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    t_slot->cleanup_sockets = 0;

    for (i=0;i<strlen(input);i++)
    {
      if ((input[strlen(input)-1] == 0x0A) || (input[strlen(input)-1] == 0x0D))
      {
        input[strlen(input)-1] = 0;
      }
    }

    if (strcmp(input,resp) == 0)
    {
      fprintf(t_slot->io_w,"AUTH:OK\n");
      (*t_slot->connections)++;
      fprintf(t_slot->io_w,"%s\n",t_slot->msg_connected);
      tmp_str = malloc(strlen(t_slot->msg_connected) + 1);
      memcpy(tmp_str,t_slot->msg_connected,strlen(t_slot->msg_connected) + 1);
      nfree(t_slot->msg_connected);
      t_slot->msg_connected = NULL;
      fflush(t_slot->io_w);
      asprintf(&z,"%s server: Client %s, port %d successfully authenticated.\n",t_slot->id_string,t_slot->client_str,ntohs(t_slot->sin_port));
      print_log(MSG_NOTICE,z);
      nfree(z);
      pthread_mutex_lock(t_slot->numlist_update_mutex);
      *t_slot->nl = int_add(*t_slot->nl,t_slot->socket,t_slot->socket_dup,&t_slot->sin_addr,t_slot->sin_port,time(NULL),t_slot->io_r,t_slot->io_w,t_slot->client_str, tmp_str);
      t_slot->client_str = NULL;
      pthread_mutex_unlock(t_slot->numlist_update_mutex);
      goto no_cleanup;
    }
    {
      fprintf(t_slot->io_w,"AUTH:ERR\n");
      fflush(t_slot->io_w);
      asprintf(&z,"%s server: Client %s, port %d failed authentication.\n",t_slot->id_string,t_slot->client_str,ntohs(t_slot->sin_port));
      print_log(MSG_NOTICE,z);
      nfree(z);
      goto do_cleanup;
    }
  } else
  {
    (*t_slot->connections)++;
    fprintf(t_slot->io_w,"%s\n",t_slot->msg_connected);
    tmp_str = malloc(strlen(t_slot->msg_connected) + 1);
    memcpy(tmp_str,t_slot->msg_connected,strlen(t_slot->msg_connected) + 1);
    nfree(t_slot->msg_connected);
    t_slot->msg_connected = NULL;
    fflush(t_slot->io_w);
    pthread_mutex_lock(t_slot->numlist_update_mutex);
    *t_slot->nl = int_add(*t_slot->nl,t_slot->socket,t_slot->socket_dup,&t_slot->sin_addr,t_slot->sin_port,time(NULL),t_slot->io_r,t_slot->io_w,t_slot->client_str,tmp_str);
    pthread_mutex_unlock(t_slot->numlist_update_mutex);
    asprintf(&z,"%s server: Client %s, port %d connected.\n",t_slot->id_string,t_slot->client_str,ntohs(t_slot->sin_port));
    t_slot->client_str = NULL;
    print_log(MSG_NOTICE,z);
    nfree(z);
    goto no_cleanup;
  }

  no_cleanup:;
  t_slot->cleanup_sockets = 0;
  goto doit;
  do_cleanup:;
  t_slot->cleanup_sockets = 1;
  doit:;
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_exit((void*)0);
  pthread_cleanup_pop(0);
}

void auth_cleanup(void *r)
{
  struct thread_slot *t_slot = (struct thread_slot *)r;

  if (t_slot->cleanup_sockets)
  {
    if (t_slot->client_str)
    {
      nfree(t_slot->client_str);
      t_slot->client_str = NULL;
    }

    if (t_slot->io_r)
    {
      fclose(t_slot->io_r);
      t_slot->io_r = NULL;
    }

    if (t_slot->io_w)
    {
      fclose(t_slot->io_w);
      t_slot->io_w = NULL;
    }

    if (t_slot->msg_connected)
    {
      nfree(t_slot->msg_connected);
      t_slot->msg_connected = NULL;
    }

    shutdown(t_slot->socket_dup,0);
    close(t_slot->socket_dup);
    shutdown(t_slot->socket,0);
    close(t_slot->socket);
  }
  t_slot->inuse = 0;
}

char *gethostname_rr (char *host)
{
    struct hostent hostbuf, *hp;
    size_t hstbuflen;
    char *tmphstbuf;
    int32_t res;
    int32_t herr;
    hstbuflen = 1024;
    char *str;
    str = NULL;
    tmphstbuf = nmalloc (hstbuflen);

    while ((res = gethostbyname_r (host, &hostbuf, tmphstbuf, hstbuflen, &hp, &herr)) == ERANGE)
    {
        hstbuflen *= 2;
        tmphstbuf = nrealloc (tmphstbuf, hstbuflen);
    }

    if (hp)
    {
        if (hp->h_addr)
        {
            str = nmalloc(16);
            inet_ntop(AF_INET,(void*)hp->h_addr,str,16);
        }
    }
    nfree(tmphstbuf);
    return str;
}

char *gethostaddr_rr (struct in_addr *addr)
{
    struct hostent hostbuf, *hp;
    size_t hstbuflen;
    char *tmphstbuf;
    int32_t herr;
    hstbuflen = 1024;
    char *str;
    str = NULL;
    tmphstbuf = nmalloc (hstbuflen);

    gethostbyaddr_r(addr,sizeof(struct in_addr),AF_INET,&hostbuf, tmphstbuf, hstbuflen, &hp, &herr);

    if (hp)
    {
        if (hp->h_name)
        {
            str = nmalloc(strlen(hp->h_name)+1);
            memset(str,0,strlen(hp->h_name)+1);
            memcpy(str,hp->h_name,strlen(hp->h_name));
        }
    }
    nfree(tmphstbuf);
    return str;
}

void print_log(int32_t msg_type,char *msg)
{
  int32_t mi;
  int32_t ret;
  char *date;
  time_t tid_t;
  struct tm *now;
  char *log_msg;
  FILE *file;
  date = malloc(50);
  time(&tid_t);

  pthread_mutex_lock(&time_mutex);
  now = localtime(&tid_t);
  strftime(date,49,date_format_long,now);
  pthread_mutex_unlock(&time_mutex);

  if (strlen(msg_type_prefix[msg_type]) != 0)
  {
    asprintf(&log_msg,"%s %s: %s",date,msg_type_prefix[msg_type],msg);
  } else
  {
    asprintf(&log_msg,"%s %s",date,msg);
  }

  printf("%s",log_msg);
  pthread_mutex_lock(&print_log_mutex);
  mi = get_log_function_index(config,"main");
  if (mi>=0)
  {
    if (config->log_config[mi]->str_num_is_set[program_log] && (config->log_config[mi]->str_num[program_log] == yes))
    {
      if (config->log_config[mi]->str[program_logfile])
      {
        if ((file = fopen(config->log_config[mi]->str[program_logfile],"a+")) != NULL)
        {
          ret = fwrite(log_msg,strlen(log_msg),1,file);
          if (ret != 1)
          {
            printf("Error: Writing log entry to %s failed\n",config->log_config[mi]->str[program_logfile]);
          }
          fclose(file);
        } else
        {
          printf("Error: Couldn't open %s for writing\n",config->log_config[mi]->str[program_logfile]);
        }
      }
    }
  }
  free(date);
  free(log_msg);
  pthread_mutex_unlock(&print_log_mutex);
}

char *build_sha1_response_string(char *challenge, char *password)
{
  sha1_context sha;
  char tmp[20];
  int32_t i;
  char *temp;

  if ((challenge != NULL) && (password != NULL))
  {
    for (i=0;i<strlen(challenge);i++)
    {
      if ((challenge[strlen(challenge)-1] == 0x0A) || (challenge[strlen(challenge)-1] == 0x0D))
      {
        challenge[strlen(challenge)-1] = 0;
      }
    }

    for (i=0;i<strlen(password);i++)
    {
      if ((password[strlen(password)-1] == 0x0A) || (password[strlen(password)-1] == 0x0D))
      {
        password[strlen(password)-1] = 0;
      }
    }

    asprintf(&temp,"%s%s",challenge,password);
    sha1_starts(&sha);
    sha1_update(&sha, (unsigned char*)temp, strlen(temp));
    sha1_finish(&sha,(unsigned char*)&tmp);

    free(temp);

    asprintf(&temp,"AUTHRESP:%08X%08X%08X%08X%08X",(unsigned int)sha.state[0],
                                                   (unsigned int)sha.state[1],
                                                   (unsigned int)sha.state[2],
                                                   (unsigned int)sha.state[3],
                                                   (unsigned int)sha.state[4]);
    return temp;
  }
  return NULL;
}

char *pcm_build_header_string(struct pcm_raw_chunk *ab)
{
  char *tmp;
  if (ab)
  {
    asprintf(&tmp,"%d:%d:%d",ab->samplerate,ab->bits,ab->channels);
    return tmp;
  }
  return NULL;
}

char *raw_build_header_string(struct pcm_raw_chunk *ab)
{
  int32_t i;
  char *t;
  t = NULL;
  char tmp[1000];
  memset(&tmp,0,1000);

  if (ab)
  {
    for (i=0;i<ab->bitrate_list->count;i++)
    {
      asprintf(&t,"%d",ab->bitrate_list->list[i]);
      strcat(tmp,t);
      nfree(t);
      if (i != (ab->bitrate_list->count - 1))
      {
        strcat(tmp,",");
      }
    }
    asprintf(&t,"%d:%s",ab->channels,tmp);
    return t;
  }
  return NULL;
}

int32_t pcm_header_string_match_parse(char *str, struct tcp_struct *tcp_struct)
{
  if (sscanf(str,"%d:%d:%d",&tcp_struct->samplerate,&tcp_struct->bits,&tcp_struct->channels) == 3)
  {
    return 1;
  } else
  {
    return 0;
  }
}

int32_t raw_header_string_match_parse(char *str, struct tcp_struct *tcp_struct)
{
  struct integer_list *list;
  list = NULL;
  char *tmp;
  int32_t channels;

  tmp = nmalloc(strlen(str) + 1);
  memset(tmp,0,strlen(str)+1);

  if (sscanf(str,"%d:%s",&channels,tmp) == 2)
  {
    list = parse_int_csv(tmp,NULL,1);
  }

  nfree(tmp);

  if (list)
  {
    if ((list->count > 0) && (channels > 0))
    {
      tcp_struct->bitrate_list = list;
      tcp_struct->channels = channels;
      return 1;
    }
  } else
  {
    tcp_struct->bitrate_list = NULL;
    tcp_struct->channels = channels;
    nfree(list);
  }
  return 0;
}

void update_stats_structs(void)
{
  int32_t i,j,k;
  struct sigaction sv;

  pthread_mutex_lock(&stats_mutex);

  for (i=0;i<s_stats.ch_br_count;i++)
  {
    nfree(t_stats.ch_br[i]);
    nfree(s_stats.ch_br[i]);
  }

  if (t_stats.ch_br)
  {
    nfree(t_stats.ch_br);
    t_stats.ch_br = NULL;
  }

  if (s_stats.ch_br)
  {
    nfree(s_stats.ch_br);
    s_stats.ch_br = NULL;
  }

  s_stats.ch_br_count = config->log_config[main_config_index]->num[dsp_decode_channels]*config->log_config[main_config_index]->num_list[bitrates]->count;
  s_stats.ch_br = nmalloc(s_stats.ch_br_count*sizeof(struct ch_br_stats *));

  t_stats.ch_br_count = config->log_config[main_config_index]->num[dsp_decode_channels]*config->log_config[main_config_index]->num_list[bitrates]->count;
  t_stats.ch_br = nmalloc(t_stats.ch_br_count*sizeof(struct ch_br_stats *));

  for (i=0;i<s_stats.ch_br_count;i++)
  {
    s_stats.ch_br[i] = nmalloc(sizeof(struct ch_br_stats));
    t_stats.ch_br[i] = nmalloc(sizeof(struct ch_br_stats));
    memset(s_stats.ch_br[i],0,sizeof(struct ch_br_stats));
    memset(t_stats.ch_br[i],0,sizeof(struct ch_br_stats));
  }

  sv.sa_handler = (void*)stats_handler;
  sv.sa_flags   = SA_RESTART;
  sigaction(SIGINT,&sv,NULL);

  k = 0;

  for (i=0;i<config->log_config[main_config_index]->num[dsp_decode_channels];i++)
  {
    for (j=0;j<config->log_config[main_config_index]->num_list[bitrates]->count;j++)
    {
      s_stats.ch_br[k]->channel = i;
      s_stats.ch_br[k]->bitrate = config->log_config[main_config_index]->num_list[bitrates]->list[j];

      t_stats.ch_br[k]->channel = i;
      t_stats.ch_br[k]->bitrate = config->log_config[main_config_index]->num_list[bitrates]->list[j];
      k++;
    }
  }
  s_stats.s.last_updated = time(NULL);
  t_stats.s.last_updated = time(NULL);
  pthread_mutex_unlock(&stats_mutex);
}


struct string_list *glob_files(char *str)
{
  char *tmp,*s;
  glob_t g;
  struct string_list *list;
  int32_t t,i;
  t = 0;
  memset(&g,0,sizeof(glob_t));
  list = nmalloc(sizeof(struct string_list));
  memset(list,0,sizeof(struct string_list));
  tmp = nmalloc(strlen(str) + 1);
  memcpy(tmp,str,strlen(str)+1);

  while ((s = strtok(((t == 0) ? tmp:NULL)," ")) != NULL)
  {
    t++;
    glob(s,0,NULL,&g);
    for (i=0;i<g.gl_pathc;i++)
    {
      list->count++;
      list->str = (list->count == 1) ? nmalloc(sizeof(char *)*list->count) : nrealloc(list->str,sizeof(char *)*list->count);
      list->str[list->count - 1] = nmalloc(strlen(g.gl_pathv[i]) + 1);
      memcpy(list->str[list->count - 1],g.gl_pathv[i],strlen(g.gl_pathv[i]) + 1);
    }
    globfree(&g);
  }
  nfree(tmp);
  return list;
}
